;
; OpenSSL SSL/TLS library module
; Copyright (C) Mikael More 2008, 2011, 2012, 2013
; MIT attribution license.
;
; For console interaction with an SSL server, you can use gnutls-cli --insecure -p [port] --crlf [host] .
;
; The choice of license is because making this module spanned several years and took several
; man-weeks, and was thus an extraordinary effort considering the amount of functionality provided.
;
; ## Note on using several instances with this module serving one and the same domain
; It appears that SSL server setups where one and the same server IP is served by several concurrently
; running SSL servers, use a clustered 'session cache'. Currently this module does not support such a
; feature.
;
; ## OpenSSL version requirement
; It appears that a version of at least approx 1.0.1 is required, older ones fail with OpenSSL internal errors.
;
; Version 1.0.1e tested and found to work.
;
; ## Platform support
; Known to work on Linux and OpenBSD, on 32bit and 64bit platforms.
; 
; An earlier version worked on Windows (Microsoft Visual C), which supposedly remains true also today.
;
; Due to the proven support above, this module supposedly works in all mainstream OS:es such as FreeBSD.
:
; ## Description
; This library provides a complete set of routines for performing client- and server-side SSL channel
; communication using OpenSSL.
;
; It provides interfaces both in the form of
;
;  * An application level u8vector port implementation for Gambit's IO system, suitable for any code where convenience
;    is of priority (the way integration with Gambit's IO system is made implies at least one extra round of buffering),
;    and
;
;  * Routines for directly establishing and accessing an SSL channel using specific routines provided by this module,
;    suitable for applications optimized for performance, such as web servers. (These routines require no object allocation
;    for communicated data.)
;
; OpenSSL can communicate with the other side two ways: By directly itself reading from the endpoint socket, or
; by getting the input data in the form of an u8vector. The current version of this module supports primarily
; the first variant, but can easily be extended to the second variant too.
;
; ## Export
; # Application-level client port
; 
;
; ## Note on our model of use of Gambit's IO system and the...
; # Closed socket handling 
; Quite incredibly, there's no way in the standard Unix API:s to check if a socket is closed or not. The only way to
; check for this is by select():ing on the regarded socket and when it gives a read signal, making a read() operation
; yourself, and if you don't get data then, then you know it's closed. Perhaps theoretically there could be some
; hypothethical cornercase issue even with this model, though I wouldn't believe so.
;
; Experience at least until now shows that OpenSSL correctly interprets a connection as closed, and we get
; 'syscall/eof-made-in-protocol-violating-way on those, which the routines in this module interpret correctly
; and respond with a proper shutdown.
;
; # Timeout handling
; Interestingly, use of Gambit's |input-port-timeout-set!| and |output-port-timeout-set!| interferes with our IO work,
; in the respects that (a) when a timeout happens, we do not get a callback to the timeout thunk, BUT, (b) subsequent
; to that the Gambit timeout happened, all SSL_read() operations fail and also doing the OS' read() eternally gives
; a -1 error and errno is set to EAGAIN.
;
; Because of this, we use TCP ports with no port-global timeout setting, but specify timeouts on a per-invocation
; basis to the |port-wait-for-input| and |port-wait-for-output| procedures. This approach works perfectly.
;
; ## Note on the ambiguity of OpenSSL documentation and use
; Please bear in mind that practical use of OpenSSL always is an at-the-very-very-least mildly tricky business,
; as the OpenSSL library documentation is ambiguous on how to respond to certain status codes and in what order
; to do some things. For these cases, a study of how other extremely commonly used libraries such as libCurl use
; OpenSSL has been done, and their behavior has been mimicked.
;      Obviously it may be that there is some detail yet to mimick correctly, so in case of any unexpected
; behavior, you may further this comparative study.
;      Just for you to get the width of this: How to run OpenSSL operations as for OpenSSL to perform a handshake
; is not completely clear, what to do in response to an OpenSSL read operation that didn't return any data is
; unclear, and how to differentiate between an SSL channel that failed and one that was closed the right way,
; is unclear (well, internally an error code is produced for the way *every* web browser closes the channel).
;
; ## Handling of OpenSSL's error queue
; OpenSSL maintains an 'error queue' per OS thread that uses OpenSSL.
;
; OpenSSL adds elements to this error queue during OpenSSL procedure calls, anytime it pleases.
; It appears to me for now that it only does that when an error has occurred, with the exception of
; SSL_load_client_CA_file (on OpenSSL 0.9.7h and 0.9.8 that always leaves an error, see
; https://github.com/git-mirror/nginx/blob/master/src/event/ngx_event_openssl.c ngx_ssl_client_certificate
; row 360-365).
;
; The error queue handling becomes a delicate question as the SSL_get_error routine that is invoked to
; get the error code for the last OpenSSL call done, relies not only on the return value from that
; call *but also on the error queue*!
;
; E.g.:
;
;      SSL_accept(...); // - fails, and feeds some error to the error queue
;      int ret = SSL_write(ssl,...); // - wants to read data i.e. would produce an SSL_WANTS_READ
;      int err = SSL_error_get(ssl,ret); // - returns a COMPLETELY UNRELATED error such as SSL_ERROR_SSL ,
;                                        //   based on peeking the earliest error queue element on the queue.
;
; Thus, due to this, if error queue handling is not done completely properly, a perfectly working SSL
; connection could be reported as closed because of error, and so on.
;
; Because of this, code using OpenSSL tends to constantly clear the error queue just as to keep safe.
; In particular, SSL_read and SSL_write calls are preceded by a clearing of the error queue by for
; instance Nginx and Cherokee.
;
; ## Debugging notes
; The first step in debugging is to (set! openssl-verbose #t) .
;
; The most primary error to track down would probably be eternal loops because of 'want-read or 'want-write
; by some reason happening indefinitely.
;
; ## Development background
; Started as a Gambit port of the OpenSSL Egg for Chicken Scheme v 1.1.7 by Thomas Chust,
; which in turn is a port of the OpenSSL functionality found in PLT Scheme.
; Today, none of either of those two projects is in use in this module.
;
; Various sources have been meditated upon in order to get the non-blocking functionality
; going. While there are numerous who question its doability, most admit that it can really
; be done.
;
; One source fully taken into account in the development is lib/ssluse.c of libCurl 7.19.0.
;
; Another one is https://github.com/engineyard/loudmouth/blob/master/loudmouth/lm-ssl-openssl.c .
;
; A revealing example of how this library should be used can be found in the OpenSSL
; source distribution's demmos/ssl/cli.cpp .
;
; An OpenSSL interface that's similar in functionality to this one is Python's.
;
; Instructive for how to use the OpenSSL API:s for performing SSL links based on u8vectors as input and output
; in both directions, is the example given by the Futhark web server for Gambit, by Francesco Bracchi.
; Its repository is at http://git.berlios.de/cgi-bin/gitweb.cgi?p=futhark;a=tree;hb=HEAD and the main files
; are in src/ehwas/ and src/encode/openssl*.scm .
;
; Production OpenSSL server use examples are
;  * src/event/ngx_event_openssl.c of Nginx' sourcecode distribution available at
;    https://github.com/git-mirror/nginx/blob/master/src/event/ngx_event_openssl.c
;  * Cherokee/cryptor_libssl.c of Cherokee's sourcecode distribution
;
; ## Further references on OpenSSL use
; "SSL Programming Tutorial"
; http://h71000.www7.hp.com/doc/83final/ba554_90007/ch04s03.html
;
; ## Instructions to build OpenSSL from scratch
; ./config --prefix=/usr/local --openssldir=/usr/lib shared
; make; make install; ldconfig
;
; ## To build on Windows:
; http://www.devside.net/guides/windows/openssl instructs. Essentially:
;  - Download and install ActivePerl. Add it to PATH environment variable.
;  - Download and uncompress OpenSSL source files.
;  - Open a command prompt, go to OpenSSL's root dir.
;  - Run perl Configure VC-WIN32 then nmake -f ms\ntdll.mak then nmake -f ms\ntdll.mak test
;  - .lib files are put in OpenSSL's out32 dir.
;
; ## History
; 2012-11-14: Made the application-layer not allocate new u8vectors on each read op but allocate and reuse
;             an u8vector read buffer for all work.
; 2012-11-16: The routines for direct channel operation are now implemented and working.
; 2012-11-20: Commented out the SSLv2 support as to support some secure-by-default OpenSSL distributions,
;             such as the one bundled with OpenBSD, do not include the SSLv2 support and thus give
;             errors when compiling this module, if attempting to use that.
;                  There are documents that say that SSLv2 is fundamentally flawed, so it ought not be
;             a loss that we don't support its explicit use. Now do note that we support the 'sslv2-or-v3
;             mode and that that is the default for client connections.
; 2012-11-23: Now implemented channel routines that give a beneficiary direct access to an SSL channel.
;             These routines are zero-copying and optimized for performance.
;
;             Both channel routines and Gambit SSL port now work both in client and server modes.
;
;             The work with making the channel routines work has required a great effort with meticulous
;             study and precision in order to get it right.
;
;             Based on experience with web browsers, the default SSL version has been set to SSLv3, TLS
;             might cause errors.
; 2013-03-10: Found a concurrency bug thanks to a stress test - Between a SSL_write/SSL_read invocation and
;             invocation of SSL_get_error , other SSL_write/SSL_read/etc. calls could be made so that the
;             SSL_get_error call would return error code result for another SSL call than the one regarded.
;                  This was not obvious at all, as SSL_get_error takes both a SSL and int ret argument
;             making it appear reasonable that it's OK if other SSL calls are done in between. This was,
;             until reading the following at https://www.openssl.org/docs/ssl/SSL_get_error.html :
; 
;                  "In addition to ssl and ret, SSL_get_error() inspects the current thread's OpenSSL error
;                  queue. Thus, SSL_get_error() must be used in the same thread that performed the TLS/SSL
;                  I/O operation, and no other OpenSSL function calls should appear in between."
;
;              and furthermore that
;
;                   "The current thread's error queue must be empty before the TLS/SSL I/O operation is
;                   attempted, or SSL_get_error() will not work reliably."
;
;              Now that was a realization. (Please note that OpenSSL's error queue is local to the current
;              OS thread.)
;
;              So basically we need to know that any SSL_get_error call we do follows the regarded SSL
;              operation directly, with no other OpenSSL call done in between.
;                   Because I'm not completely clear about the scope of (declare (not interrupts-enabled))
;              and because doing start-stops of Gambit's thread scheduler would probably go at a higher CPU
;              expense than the following method, then, we from now do any SSL_get_error and OpenSSL error
;              queue access relating to a particular OpenSSL call, in the same c-lambda as that OpenSSL call
;              was performed.
;                   We store the error results into a special C structure we devised for this purpose.
;
;              See the "Handling of OpenSSL's error queue" section in the comments in this header and
;              "Our internal error result vector C type: ..." section for more information.
;
; ## TODO
;  - Support the "NULL" authentication mode i.e. the server encryption mode that generates a new key
;    for each connection and does not use any certificate files.
;  - When the TLS connection goes down, the software port doesn't go down so the beneficiary notices it,
;    to resolve.
;  - When getting a 'syscall return value from a read or write operation, the right thing would be to make a
;    ERR_get_error() call to see what the error is, and then handle it and if zero, the return value number from
;    the call that provided the 'syscall return value. Currently we don't see this happen, therefore I did not make
;    a priority of bringing order to this.
;         Such a routine needs to be invoked before any more call to OpenSSL API:s are done, as they could
;    overwrite the error code.
;  - Introduce IO timeouts
;  - Unilateral port close for both port and channel routines implementations.
;  - See the note about session cache at the top of the header here.
;  - Note, if to start supporting multiple OS threads, at thread termination ERR_remove_state should be called,
;    see https://www.openssl.org/docs/crypto/ERR_remove_state.html# .
;
; ## USAGE EXAMPLES
#||

    (import openssl)

    ; CLIENT EXAMPLE 1: Simple HTTP GET using application-layer OpenSSL channel, via |call-with-ssl-tcp-client|.
    (call-with-ssl-tcp-client
     '(server-address: "www.ft.com" port-number: 443)
     (lambda (port)
       (display "Sending request to server.\n")
       (thread-sleep! 5)
       (display "GET / HTTP/1.1\nHost: www.ft.com\n\n" port)
       (force-output port 1)
       (display "Reading response from server.\n")
       (thread-sleep! 5)
       (display " *** Have read:\n ** ")

       (let loop ()
         (let* ;((c (read-u8 port)))
             ((v (make-u8vector 1)))
           (read-subu8vector v 0 1 port)
           (let ((c (u8vector-ref v 0)))
             (if (not (eq? #!eof c))
                 (begin
                   (display (make-string 1 (integer->char c)))
                   (force-output)
                   (loop))))))
       (display " (end)\n")
       (thread-sleep! 5)))

    ; CLIENT EXAMPLE 2: Simple HTTP GET using application-layer OpenSSL channel, via |open-ssl-tcp-client|.
    (define port (open-ssl-tcp-client '(server-address: "www.ft.com" port-number: 443)))

    (display "GET / HTTP/1.1\r\nHost: www.ft.com\r\n\r\n" port) (force-output port)

    (let loop ()
      (let ;((c (read-u8 port)))
          ((v (make-u8vector 1)))
        (read-subu8vector v 0 1 port)
        (let ((c (u8vector-ref v 0)))
          (if (not (eq? #!eof c))
              (begin
                (display (integer->char c))
                (loop))))))
    (display "Out.")
    
    ; CLIENT EXAMPLE 3: Simple HTTP get using direct channel routines.
    (import openssl_latest) (set! openssl-verbose #t) ; For getting stream of debug output to the console.
    (import openssl_latest)
    (define port (open-tcp-client '(server-address: "www.ft.com" port-number: 443)))
    (define channel (tcp-port->ssl-client-channel port))
    
    ((channel 'display) "GET / HTTP/1.1\r\nHost: www.ft.com\r\n\r\n")
    ((channel 'close-output-port))
    (let ((read-u8 (channel 'read-u8)))
      (let loop ()
        (let ((b (read-u8)))
          (if (not (eq? #!eof b))
              (begin
                (display (make-string 1 (integer->char b)))
                (loop))))))

    ; SERVER EXAMPLE 1 & 2 & 3 & 4
    (import openssl_latest) (set! openssl-verbose #t) ; For getting stream of debug output to the console.
    (import openssl_latest)
    (define server-ssl-ctx (make-server-ssl-ctx))

    ; Load the certificate chain file into the CTX
    ; (OpenSSL does *not* easily provide any possibility to feed the chain file from a string or alike.
    ; For an example how that would be done, see the reference in the comments in this file of how Nginx' 
    ; does something related.)
    (ssl-ctx-use-certificate-chain-file server-ssl-ctx "[filename].ca-bundle")

    ; Load the certificate file into the CTX
    (ssl-ctx-use-certificate-file server-ssl-ctx "[filename].crt" OpenSSL-FileType-PEM)

    ; Load the private key file into the CTX
    (ssl-ctx-use-private-key-file server-ssl-ctx "[filename].key" OpenSSL-FileType-PEM)

    (define (run-server port-number make-connection/channel-proc handle-connection)
      (let ((server (open-tcp-server `(server-address: "*" port-number: ,port-number backlog: 128))))
        (let loop ((at 0))
          (let ((connection-raw (read server)))
            (thread-start! (make-thread
                            (lambda ()
                              (print "\n\n\n\n ## Got a connection, " at ".\n")
                              (let ((connection/channel (make-connection/channel-proc connection-raw server-ssl-ctx)))
                                (handle-connection connection/channel)))))
            (loop (+ at 1))))))

    ; SERVER EXAMPLE 1: SSL server using the application-layer OpenSSL channel, that receives incoming connections and dumps their output to the console.
    (run-server 8080 wrap-tcp-port-into-ssl-server-port
                (lambda (connection)
                  (display "Reading data.\n")
                  (let loop ()
                    (let ;((c (read-u8 connection)))
                        ((v (make-u8vector 1)))
                      (read-subu8vector v 0 1 connection)
                      (let ((c (u8vector-ref v 0)))
                        (if (not (eq? #!eof c))
                            (begin
                              (display (integer->char c))
                              (loop))))))
                  (display "Connection closed.\n")))

    ; SERVER EXAMPLE 2: SSL server using the application-layer OpenSSL channel, that receive incoming connections, and provide them with a HTTP response.
    ;                   (Presumes all HTTPS access is done to the / URL path.)
    (run-server 8081 wrap-tcp-port-into-ssl-server-port
                (lambda (connection)
                  (display "Responding.\n")
                  (display "HTTP/1.1 200 OK\r\nContent-Type: text/plain\r\nContent-Length: 12\r\n\r\nHello World!\r\n" connection)
                  (force-output connection 1)
                  (close-port connection)))

    ; SERVER EXAMPLE 4: SSL server using the direct channel routines, that receives incoming connections, and provides them with a HTTP response.
    ;                   (Presumes all HTTPS access is done to the / URL path.)
    ;
    ; Please note that as not to make web browsers believe a connection can be re-used for two HTTP requests, we signal Connection: close
    ; and for safety also we signal HTTP 1.0 rather than 1.1, which has Connection: close by default afaik.
    (run-server 8081 tcp-port->ssl-server-channel
                (lambda (channel)
                  (display "Responding.\n")
                  (let* ((string->u8vector (lambda (str) (apply u8vector (map char->integer (string->list str)))))
                         (data (string->u8vector "Hello World!"))
                         (data-length (u8vector-length data))
                         (display (channel 'display))
                         (write-subu8vector (channel 'write-subu8vector))
                         (close-port (channel 'close-port)))
                    (display "HTTP/1.0 200 OK\r\nContent-Type: text/plain\r\nConnection: close\r\nContent-Length: ")
                    (display (number->string data-length))
                    (display "\r\n\r\n")
                    (write-subu8vector data 0 data-length)
                    (close-port))))

||#

; At least with the OpenSSL bundled with OpenBSD -current as of Nov 2012, Firefox 10.0.0.7 would give ssl_error_no_cypher_overlap
; error with 'tls . So let's go with SSLv3 for now.
(define default-server-ssl-version 'sslv3)

(define default-client-ssl-version 'sslv2-or-v3)

; The verbose output code is *always* compiled into the module so you can always switch on the verbosity anytime.
; This is of value due to the arbitrariness of the OpenSSL docs and conventions, at some unexpected point you might
; very well want to know what's going on under the hood.
(define openssl-verbose #f)
(set! openssl-verbose #f) ; To make compiler get it's mutable (otherwise due to (block) #f would be inlined)

(define openssl-port-beneficiary-to-host-buffer-size   4096)
(define openssl-port-beneficiary-from-host-buffer-size 4096)

; OpenSSL can make the SSL handshake either explicitly during the SSL_connect/SSL_accept call, or implicitly as part of
; the SSL_write/SSL_read calls for the connection. I'd guess the latter is much more suitable for us as it should
; mean it's a nonblocking behavior from OpenSSL whereas I'd guess the further one could be blocking, but I don't know.
; In all cases we leave it configurable here.
(define ssl-perform-explicit-connect/accept-on-open? #f)

(define empty-u8v '#u8())

(compile-options ld-options-prelude: "-lssl -lcrypto"
                 ; cc-options: "-I/usr/local/ssl/include" ; If there's a custom install of SSL, /usr/local/ssl is its default
                 ;                                        ; installation directory, so that's where we look first.
                 force-compile: #t)

(declare (block)
         (standard-bindings)
         (extended-bindings)
         (mostly-fixnum)
         (not optimize-dead-local-variables) ; For the ctx argument to |ssl-make-bidirectional-port| and |ssl-channel| not to be GC:ed.
         )

(import (std string/util net/tcpip)
        (only: (std misc/u8v) subu8vector->hex-string) ; for debug only
        ../util/util)
(syntax-begin (import (std string/util)))

; ## Debug output routines
(define-macro (if-verbose . code)
  `(if openssl-verbose
       (begin
         ,@code)))

(define-macro (if-verbose-dbg . a) `(if-verbose (dbg "ssl-channel for #" (object->serial-number tcp-socket-port) ": " ,@a)))

; ## Utils
(define (apply-string-append string-list)
  (call-with-output-string
   '()
   (lambda (output-port)
     (let loop ((rest string-list))
       (if (not (null? rest))
           (let ((v (car rest)))
             (if (not (string? v)) (error "apply-string-append got non-string argument" v))
             (display v output-port)
             (loop (cdr rest))))))))

(define (seconds-since-epoch)
  (time->seconds (current-time)))

(define-macro (fx1+ v) `(fx+ ,v 1))

; Routine to make a test OS read() to see what return and errno values it comes back with, for possible
; further analysis of an IO problem.
(define dbg:os-read-test
  (c-lambda (int) void #<<c-lambda-end
char b[1];
int r = read(___arg1,(void*) &b,1);
fprintf(stdout,"Got %i. errno is %i. (EAGAIN=%i, ECONNRESET=%i, ENOTCONN=%i, ETIMEDOUT=%i.)\n",r,errno,EAGAIN,ECONNRESET,ENOTCONN,ETIMEDOUT); 

c-lambda-end
            ))

; ## OpenSSL library code
(c-declare #<<c-declare-end

// For Microsoft Windows support: Library dependency instructions to Microsoft Visual C++ 
#ifdef _WIN32
#pragma comment(lib,"ws2_32")
#pragma comment(lib,"libeay32")
#pragma comment(lib,"ssleay32")
#endif

// C includes
#include <errno.h>
#include <unistd.h>

#include <openssl/err.h>
#include <openssl/ssl.h>
#include <openssl/bio.h>
#include <openssl/conf.h> // For OPENSSL_config()

// C includes for Microsoft Windows support
#ifdef _WIN32
  #ifdef _MSC_VER
    #include <winsock2.h>
  #else
    #include <ws2tcpip.h>
  #endif

  #include <openssl/rand.h>
#else
  #define closesocket     close
#endif

// Not clear in this moment what this is about.
#ifdef ECOS
  #include <sys/sockio.h>
#else
  #include <unistd.h>
#endif

int create_u8vector_from_buffer(const void* buf, size_t bytes) {
     ___SCMOBJ r = ___EXT(___alloc_scmobj)(___sU8VECTOR,bytes,___MOVABLE0);
     ___EXT(___release_scmobj) (r);

     // Determine where in RAM the u8vector is (0 is the offset into u8vector)
     void *u8vectorptr = ___CAST(void*,&___FETCH_U8(___BODY(r),___INT(0)));

     // Copy gotten data to it
     memcpy(u8vectorptr,buf,(int) bytes);

     return r;
}

c-declare-end
           )

; # OpenSSL library initialization
(c-initialize #<<c-initialize-end

// Inspired by Nginx' ngx_ssl_init(), https://github.com/git-mirror/nginx/blob/master/src/event/ngx_event_openssl.c row 73
// Reads the OpenSSL configuration file from its default location.
// https://www.openssl.org/docs/crypto/OPENSSL_config.html
// OPENSSL_config(NULL);

// (Nginx does the following three in another order,
// https://github.com/git-mirror/nginx/blob/master/src/event/ngx_event_openssl.c ngx_ssl_init row 75-76)
ERR_load_crypto_strings();
SSL_load_error_strings(); // Order of these two calls inspired by
SSL_library_init();       // http://www.openssl.org/docs/ssl/SSL_library_init.html

#ifdef _WIN32
RAND_screen();
#endif


// Inspired by Nginx, https://github.com/git-mirror/nginx/blob/master/src/event/ngx_event_openssl.c ngx_ssl_init row 78
OpenSSL_add_all_algorithms();

// (Nginx turns off compression for <v1.0.0 here, in row ngx_ssl_init row 80-98)

// (Nginx checks SSL connection/server conf/session cache/certificate/stapling indexes here, for some purpose. Row 100-138.)

c-initialize-end
              )

; # FFI type declarations for OpenSSL types
(c-define-type SSL* (pointer "SSL" SSL* ; "release_SSL_star"
                             ))
(c-define-type SSL-CTX* (pointer "SSL_CTX" SSL-CTX* ; "release_SSL_CTX_star"
                                 ))
(c-define-type CTX* (pointer "CTX" CTX* ; "release_CTX_star"
                             ))
(c-define-type SSL-METHOD* (pointer "SSL_METHOD" SSL-METHOD* ; "release_SSL_CTX_star"
                                    ))

; # Our internal error result vector C type:
; # Definition, C routine for filling it out, Scheme accessors
;
; The error vector datatype does *not* include the value returned directly by the respective
; OpenSSL procedure - this value we let the c-lambda for the respective procedure return;
; that provides for the most efficient and clean design.
;
; The error vector datatype's job is to convey error information, for the case that we already
; know that an error happened, as OpenSSL procedure returned negative return value or false etc..
(c-declare #<<c-declare-end

#define OpenSSL_error_vector_queue_errors_length 10

typedef struct OpenSSL_error_vector {
     // Result of SSL_get_error as done for the passed const SSL* instance and int return value from
     // the regarded OpenSSL operation. This slot is int as SSL_get_error returns int.
     // For reference see https://www.openssl.org/docs/ssl/SSL_get_error.html .
     int error;

     // Dump of the error queue through use of ERR_get_error that pops the earliest error code from
     // the current OS thread's local OpenSSL error queue.
     //
     // We allow retrieval of more than one error queue element, as when
     // https://www.openssl.org/docs/ssl/SSL_get_error.html speaks about checking the error queue
     // for clarifications to an error state as reported by SSL_get_error (which it does for
     // SSL_ERROR_SYSCALL and SSL_ERROR_SSL), it speaks about the queue in general and makes no
     // mention that it regards the earliest error code on the queue only, so we impose no such
     // presumption.
     //
     // We limit the retrieval to 10 elements in total, really the first 1-2 should be more than enough.
     //
     // Please note that these error queue values are complete in themselves. They are completely
     // sufficient in themselves, for serialization to string using ERR_error_string_n documented at
     // https://www.openssl.org/docs/crypto/ERR_error_string.html .
     //
     // For reference, in addition to the error queue values, we also store a copy of the OpenSSL
     // source code file and location that it happened at.
     //
     // For reference see https://www.openssl.org/docs/crypto/ERR_get_error.html .
     unsigned long queue_errors     [OpenSSL_error_vector_queue_errors_length + 1];
     const char*   queue_errors_file[OpenSSL_error_vector_queue_errors_length + 1]; // Thanks to #c on Freenode for clarifying this def needs to look exactly like this.
     int           queue_errors_line[OpenSSL_error_vector_queue_errors_length + 1];
} OpenSSL_error_vector_t;

// This procedure is separate from OpenSSL_error_capture , for failed operations that regard
// a SSL_CTX* and not an SSL* to be able to fill out the error vector.
void OpenSSL_error_capture_queue(OpenSSL_error_vector* v, bool skip) {
     // While the ERR_get_error page indicates that we should check the error queue in case of
     // SSL_ERROR_SYSCALL or SSL_ERROR_SSL, let's do it for all errors except SSL_ERROR_WANT_READ and
     // SSL_ERROR_WANT_WRITE - OpenSSL is clearly tricky to operate so let's be generous about the 
     // error info retrieval we do.
     //
     // For more info see the comments above the definition of this slot in the struct type definition
     // above.
     unsigned long* queue = v->queue_errors;
     if (!skip) {
          unsigned long* queue_end = queue + sizeof(v->queue_errors)
                                     - sizeof(v->queue_errors[0]); // Reserve a slot for 
          unsigned long c;
          const char**  file = v->queue_errors_file;
          int*          line = v->queue_errors_line;
          *file = NULL; *line = -1; // To reset the slot just in case ERR_get_error_line won't set it.
          while ((queue < queue_end) && // Thanks to this way of checking we don't need a separate counter
                 ((c = ERR_get_error_line(file,line)) != 0)) {
               *queue = c;
               // fprintf(stderr,"OpenSSL_error_capture_queue set element to file %s line %i c %i",(*file != NULL) ? *file : "NULL",*line,c); fflush(stderr);

               queue++; // not: = sizeof(v->queue_errors     [0]);
               file ++; // not:  += sizeof(v->queue_errors_file[0]);
               line ++; // not:  += sizeof(v->queue_errors_line[0]);
               *file = NULL; *line = -1; // To reset the slot just in case ERR_get_error_line won't set it.
          }
     }
     *queue = 0;
}

void OpenSSL_store_result_and_error(OpenSSL_error_vector* v, const SSL* ssl, int ret) {

     // Load the SSL_get_error result into the error slot.
     int error = SSL_get_error(ssl,ret);
     v->error = error;

     OpenSSL_error_capture_queue(v,(error == SSL_ERROR_WANT_READ) || (error == SSL_ERROR_WANT_WRITE));

}

void OpenSSL_CTX_error_capture(OpenSSL_error_vector* v) {
     OpenSSL_error_capture_queue(v,false);
}

// For the purpose of this procedure, please see the "Handling of OpenSSL's error queue" section in the header comments.
//
// This procedure takes the opportunity to dump any queued errors to the console, as to give us the opportunity to
// understand on what call made previously it was retrieved and handled.
void OpenSSL_clear_error_queue(char* caller_name) {
     bool first = true;
     unsigned long error;
     const char*   file;
     const char*   data;
     int           line;
     int           flags;
     char          error_s[10240];
     while ((error = ERR_get_error_line_data(&file,&line,&data,&flags)) != 0) {
          if (first) {
               fprintf(stderr,"The OpenSSL module found OpenSSL queue error(s) that were present from a call prior to this %s call, namely:\n",caller_name);
               first = false;
          }
          ERR_error_string_n(error,error_s,sizeof(error_s));
          fprintf(stderr,"   * OpenSSL sourcecode file %s row %i reported error: %s, flags %i, data %s .\n",
                  (file != NULL) ? file : "NULL",line,error_s,flags,(data != NULL) ? data : "NULL");
     }
     if (!first) fprintf(stderr,"You are free to pass on this to the administrator or developer for proper determination and handling of the source, though generally you can just ignore it.\n");
}

c-declare-end
           )

(c-define-type OpenSSL-error-vector* (pointer "OpenSSL_error_vector_t" OpenSSL-error-vector*)) ; The OS' free() for freeing is fine.

(define alloc-OpenSSL-error-vector (c-lambda () OpenSSL-error-vector* "___result = (OpenSSL_error_vector*) malloc(sizeof(OpenSSL_error_vector_t));"))

(define OpenSSL-error-vector-error (c-lambda (OpenSSL-error-vector*) int "___result = ___arg1->error;"))

(define OpenSSL-error-vector-queue-errors-error      (c-lambda (OpenSSL-error-vector* int) unsigned-long "___result =         ___arg1->queue_errors     [___arg2];"))
(define OpenSSL-error-vector-queue-errors-error-file (c-lambda (OpenSSL-error-vector* int) char-string   "___result = (char*) ___arg1->queue_errors_file[___arg2];"))
(define OpenSSL-error-vector-queue-errors-error-line (c-lambda (OpenSSL-error-vector* int) int           "___result =         ___arg1->queue_errors_line[___arg2];"))

(define ERR-error-string (c-lambda (unsigned-long) char-string "char c[10240]; ERR_error_string_n(___arg1,c,sizeof(c)); ___result = c;"))

(define (OpenSSL-error-vector-dump-queue OpenSSL-error-vector*)
  (let loop ((at 0) (r '()))
    (let ((code (OpenSSL-error-vector-queue-errors-error OpenSSL-error-vector* at)))
      ; (dbg "Element " at " has code " code) (thread-sleep! 3)
      (if (zero? code)
          (reverse r)
          
          (let* ((error (ERR-error-string code))
                 (file (OpenSSL-error-vector-queue-errors-error-file OpenSSL-error-vector* at))
                 (line (OpenSSL-error-vector-queue-errors-error-line OpenSSL-error-vector* at))
                 (c `(,file ,line ,code ,error)))

            ; (dbg "OpenSSL-error-vector-dump-queue: Read " OpenSSL-error-vector* " element " at ": " c) (thread-sleep!  3)

            (loop (fx+ at 1)
                  (cons c r)))))))

; We do never run this one separately from an individual OpenSSL operation, as described in the comments
; above regarding the OpenSSL_error_vector / OpenSSL-error-vector* structure and in the 2013-03-10
; History comment.
; (define SSL_get_error (c-lambda (SSL* int) int "SSL_get_error"))

; # Import of OpenSSL constants
(define-macro (define-constants prefix c-prefix type . names)
  (let ((r `(begin
              ,@(map
                 (lambda (sym)
                   `(define ,(symbol-append prefix sym)
                      ((c-lambda
                        ()
                        ,type
                        ,(string-append
                          "___result = "
                          c-prefix
                          (string-replace-char (symbol->string sym) #\- #\_)
                          ";")))))
                 names))))
    ; (pp r)
    r))

(define-constants OpenSSL-Error- "SSL_ERROR_" long
                  NONE SSL WANT-READ WANT-WRITE WANT-X509-LOOKUP
                  SYSCALL ; look at error stack/return value/errno
                  ZERO-RETURN WANT-CONNECT WANT-ACCEPT)

(define-constants OpenSSL-OP- "SSL_OP_" long
                  ALL

                  NO-SSLv2 NO-SSLv3 NO-TLSv1
                  NO-TICKET ; Not avail in 0.9.8c. I think it exists in 0.9.8j and 0.9.9, and for sure in 1.0.0e .

                  MICROSOFT-SESS-ID-BUG
                  NETSCAPE-CHALLENGE-BUG
                  SSLREF2-REUSE-CERT-TYPE-BUG
                  MICROSOFT-BIG-SSLV3-BUFFER
                  MSIE-SSLV2-RSA-PADDING
                  SSLEAY-080-CLIENT-DH-BUG
                  TLS-D5-BUG
                  TLS-BLOCK-PADDING-BUG
                  DONT-INSERT-EMPTY-FRAGMENTS
                  SINGLE-DH-USE

                  )

(define-constants OpenSSL-BIO- "BIO_C_" long
                  SET-NBIO)

(define-constants OpenSSL-FileType- "X509_FILETYPE_" long
                  PEM ASN1)

(define ssl-method:sslv23        ((c-lambda () SSL-METHOD* "___result_voidstar = (void*) SSLv23_method       ();"))) ; "for combined client and server" say the docs.
(define ssl-method:sslv23-server ((c-lambda () SSL-METHOD* "___result_voidstar = (void*) SSLv23_server_method();")))
(define ssl-method:sslv23-client ((c-lambda () SSL-METHOD* "___result_voidstar = (void*) SSLv23_client_method();")))

; The SSLv2 method is removed in certain safe OpenSSL distributions such as that bundled with OpenBSD.
; Let's just keep it commented out.
; (define ssl-method:sslv2         ((c-lambda () SSL-METHOD* "___result_voidstar = (void*) SSLv2_method       ();"))) ; "for combined client and server" say the docs.
; (define ssl-method:sslv2-server  ((c-lambda () SSL-METHOD* "___result_voidstar = (void*) SSLv2_server_method();")))
; (define ssl-method:sslv2-client  ((c-lambda () SSL-METHOD* "___result_voidstar = (void*) SSLv2_client_method();")))

(define ssl-method:sslv3         ((c-lambda () SSL-METHOD* "___result_voidstar = (void*) SSLv3_method       ();"))) ; "for combined client and server" say the docs.
(define ssl-method:sslv3-server  ((c-lambda () SSL-METHOD* "___result_voidstar = (void*) SSLv3_server_method();")))
(define ssl-method:sslv3-client  ((c-lambda () SSL-METHOD* "___result_voidstar = (void*) SSLv3_client_method();")))

(define ssl-method:tls           ((c-lambda () SSL-METHOD* "___result_voidstar = (void*) TLSv1_method       ();"))) ; "for combined client and server" say the docs.
(define ssl-method:tls-server    ((c-lambda () SSL-METHOD* "___result_voidstar = (void*) TLSv1_server_method();")))
(define ssl-method:tls-client    ((c-lambda () SSL-METHOD* "___result_voidstar = (void*) TLSv1_client_method();")))

; server? = #t = This SSL-CTX is for a server to use to accept clients
;           #f = This SSL-CTX is for a client to use to make connections
(define (ssl-ctx-new protocol server?)
  (let ((ctx
         ((c-lambda (SSL-METHOD*) SSL-CTX* #<<c-lambda-end
                    SSL_CTX *ctx;
                    if ((ctx = SSL_CTX_new(___arg1)))
                         SSL_CTX_set_mode(ctx,SSL_MODE_ENABLE_PARTIAL_WRITE       | // Partial writes are OK, i.e. if n bytes were commanded to be written to SSL_write
                                                                                    // and number of written bytes is 0 < r < n then it's OK.
                                              SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER | // If retrying SSL_write (I guess as in, calling SSL_write again after that it
                                                                                    // returned WANT_READ or alike) then it's OK that the buffer to be written is now
                                                                                    // at another memory location. The contents need to be the same though.
                                                                                    // This is of value as Gambit's GC may move the u8vector we're writing between
                                                                                    // SSL_write invocations.
                                              SSL_MODE_RELEASE_BUFFERS              // Release unused buffers on idle connections. Seems to be quite bit popular to use
                                                                                    // in other open source projects, so let's use it here.
                                              // The other available options are:
                                              // SSL_MODE_AUTO_RETRY - Not relevant to us as this one regards blocked mode only, and we go with nonblocking always.
                                          );

                    // (No checking of the error queue done currently as it remains for us to see that queue elements may be added by this code.)

                    ___result_voidstar = ctx;
c-lambda-end
                    )
          (case protocol
            ((sslv2-or-v3) (if server? ssl-method:sslv23-server ssl-method:sslv23-client))
            ; ((sslv2      ) (if server? ssl-method:sslv2-server  ssl-method:sslv2-client ))
            ((sslv3      ) (if server? ssl-method:sslv3-server  ssl-method:sslv3-client ))
            ((tls        ) (if server? ssl-method:tls-server    ssl-method:tls-client   ))
            (else          (error "invalid SSL/TLS connection protocol" protocol))))))
    (if (not ctx) (error "Failed to create SSL_CTX." protocol server?))
    ctx))

; Note that we do a SSL_CTX_set_mode call in ssl-ctx-new and that that is all we need, so no need
; for making a special define - c-lambda for it.

; (No checking of the error queue done currently as it remains for us to see that queue elements may be added by this code.)
(define ctx-set-options (c-lambda (SSL-CTX* long) long "SSL_CTX_set_options"))

; (No checking of the error queue done currently as it remains for us to see that queue elements may be added by this code.)
(define ctx-set-read-ahead (c-lambda (SSL-CTX* int) int "SSL_CTX_set_read_ahead"))

; (No checking of the error queue done currently as it remains for us to see that queue elements may be added by this code.)
(define ctx-ctrl (c-lambda (SSL-CTX* int long) long "___result = SSL_CTX_ctrl(___arg1,___arg2,___arg3,NULL);"))

; 2012-11-16: Switched to int return value type as to pick up the result
; (No checking of the error queue done currently as it remains for us to see that queue elements may be added by this code.)
(define ctx-set-default-verify-paths (c-lambda (SSL-CTX*) int "SSL_CTX_set_default_verify_paths"))

; (No checking of the error queue done currently as it remains for us to see that queue elements may be added by this code.)
(define ssl-new (c-lambda (SSL-CTX*) SSL* "SSL_new"))

; (No checking of the error queue done currently as it remains for us to see that queue elements may be added by this code.)
(define ssl-free (c-lambda (SSL*) void "SSL_free"))

(define (ssl-result OpenSSL-error-vector*)
  (let* ((v (OpenSSL-error-vector-error OpenSSL-error-vector*)))
    (cond
     ((eq? v OpenSSL-Error-NONE)
      ret)
     ((eq? v OpenSSL-Error-ZERO-RETURN)
      'zero-return)
     ((eq? v OpenSSL-Error-WANT-READ)
      'want-read)
     ((eq? v OpenSSL-Error-WANT-WRITE)
      'want-write)
     ((eq? v OpenSSL-Error-WANT-CONNECT)
      'want-connect)
     ((eq? v OpenSSL-Error-WANT-ACCEPT)
      'want-accept)
     ((eq? v OpenSSL-Error-WANT-X509-LOOKUP)
      'want-X509-lookup)

     ; SSL_ERROR_SYSCALL error: "Some I/O error occurred. The OpenSSL error queue may contain more information on the error."
     ((eq? v OpenSSL-Error-SYSCALL)
      ; http://www.openssl.org/docs/ssl/SSL_get_error.html advises us to invoke ERR_get_error() to get an idea of the underlying
      ; error. Because we're in Gambit and lots of things are going on though, let's just skip any such error detection followup
      ; for now, because we know that anyhow it will not lead to a situation where we can rescue this SSL channel. (At least, to
      ; date I've not seen a situation where we'd get this error if the link somehow could be made to work again.)
      ;
      ; Thus, (at least given current knowledge,) we can conclude that this means we're at EOF.
      
      ; http://www.openssl.org/docs/ssl/SSL_get_error.html says that if ERR_get_error() provides no instruction, then
      ; based on the ret value we can get a further clue, it actually is subdivided into three categories. So let's differentiate
      ; the result value we provide based on these, and this way we have done our duties regarding digging into the causes of
      ; the SYSCALL error.
      ;
      ; * Note that we do have the SSL_get_error results in OpenSSL-error-vector* now. We could make any use of this and also
      ; print out the queue contents for nicety, if this would be found relevant.

      (case ret
        ; If ret == 0, an EOF was observed that violates the protocol.
        ((0)  'syscall/eof-made-in-protocol-violating-way)
        ; If ret == -1, the underlying BIO reported an I/O error (for socket I/O on Unix systems, consult errno for details).
        ((1)  'syscall/underlying-bio-reported-io-error)
        (else 'syscall/other-io-error)))

     ((eq? v OpenSSL-Error-SSL)
      'ssl)
     (else
      #f))))

; Note that possibly an element could be written to OpenSSL's OS-thread local error queue here, in case of failure.
; If/when we see this happen, we'll add handling.
(define set-fd!-internal (c-lambda (SSL* int) int "SSL_set_fd"))

; Used by |wrap-tcp-port-into-ssl-port-or-channel/internal| to wrap a Gambit TCP port into an SSL port or channel.
; The point here is that the procedure assigns the Gambit TCP port's socket to the OpenSSL SSL channel instance,
; for OpenSSL to do socket reads and writes directly on the OS socket, instead of proxying data via Gambit as
; for Gambit to do the reads and writes in turn. This is more performant than the latter described option due to
; need of less buffering and logics.
;
; It still remains for us to see this procedure fail even once, so therefore we do not do additional error
; reporting using OpenSSL-error-vector* and OpenSSL_error_capture .
;
; => #t = success, otherwise error.
;
; https://www.openssl.org/docs/ssl/SSL_set_fd.html
(define (set-fd! ssl fd)
  (eq? 1 (set-fd!-internal ssl fd)))

(define ssl-connect-internal (c-lambda (SSL* OpenSSL-error-vector*) int #<<c-lambda-end

SSL* ssl = ___arg1;
OpenSSL_error_vector_t* error_vector = ___arg2;
OpenSSL_clear_error_queue("SSL_connect");
___result = SSL_connect(ssl);

// https://www.openssl.org/docs/ssl/SSL_connect.html says the return values are either 0, 1 or -1.
/* if ((___result != 0) && (___result != 1)) */
OpenSSL_store_result_and_error(error_vector,ssl,___result);

c-lambda-end
))

; => #t = success, or the return value of |ssl-result|, which may be 'want-read or 'want-write which
; means that whether establishing the SSL channel was successful or not, not yet has been
; established.
(define (ssl-connect ssl OpenSSL-error-vector*)
  (let ((result (ssl-connect-internal ssl OpenSSL-error-vector*)))
    (if (eq? result 1)
        #t
        (ssl-result OpenSSL-error-vector*))))

(define ssl-accept-internal (c-lambda (SSL* OpenSSL-error-vector*) int #<<c-lambda-end

SSL* ssl = ___arg1;
OpenSSL_error_vector_t* error_vector = ___arg2;
OpenSSL_clear_error_queue("SSL_accept");
___result = SSL_accept(ssl);

// https://www.openssl.org/docs/ssl/SSL_accept.html says the return values are either 0, 1 or -1.
/* if ((___result != 0) && (___result != 1)) */
OpenSSL_store_result_and_error(error_vector,ssl,___result);

c-lambda-end
))

; => #t = success, or the return value of |ssl-result|, which may be 'want-read or 'want-write which
; means that whether establishing the SSL channel was successful or not, not yet has been
; established.
(define (ssl-accept ssl OpenSSL-error-vector*)
  (let ((result (ssl-accept-internal ssl OpenSSL-error-vector*)))
    (if (eq? result 1)
        #t
        (ssl-result OpenSSL-error-vector*))))

; => #t = SSL channel has been shut down in both directions, or
;    'unodirectionally = SSL channel has been shut down in one direction (the outbound direction I guess?), or
;    a |ssl-result| result value.
(define (shutdown ssl OpenSSL-error-vector*)
  (let ((result ((c-lambda (SSL* OpenSSL-error-vector*) int #<<c-lambda-end

SSL* ssl = ___arg1;
OpenSSL_error_vector_t* error_vector = ___arg2;
OpenSSL_clear_error_queue("SSL_shutdown");
___result = SSL_shutdown(ssl);

// https://www.openssl.org/docs/ssl/SSL_shutdown.html says the return values are either 0, 1 or -1.
/* if ((___result != 0) && (___result != 1)) */
OpenSSL_store_result_and_error(error_vector,ssl,___result);

c-lambda-end
                           ) ssl OpenSSL-error-vector*)))
    
    ; http://www.openssl.org/docs/ssl/SSL_shutdown.html says:
    (cond ((eq? result 0) 'unidirectionally) ; 0: "The shutdown is not yet finished. Call SSL_shutdown() for a second time, if a bidirectional shutdown
                                             ; shall be performed. The output of SSL_get_error(3) may be misleading, as an erroneous SSL_ERROR_SYSCALL
                                             ; may be flagged even though no error occurred."

          ((eq? result 1) #t) ; 1: "The shutdown was successfully completed. The ``close notify'' alert was sent and the peer's
                              ; ``close notify'' alert was received."

          (else               ; -1: "The shutdown was not successful because a fatal error occurred either at the protocol level or a
                              ; connection failure occurred. It can also occur if action is need to continue the operation for non-blocking
                              ; BIOs. Call SSL_get_error(3) with the return value ret to find out the reason."
           
           (ssl-result OpenSSL-error-vector*)))))

; Obsolete:
; Since we use asynchronous sockets always, the SSL_read call could return complex error codes and require
; complex logics for properly reading and otherwise handling the SSL channel in relation with the read
; operation. Therefore we do not do this just as a read procedure call, but only as part of our SSL
; channel wrappers.
;
; (define (ssl-get-char ssl)
;   (let ((ret
;          ((c-lambda (SSL*) int #<<c-lambda-end
;                      char ch;
;                      int ret;
;                      ret = SSL_read((SSL *)___arg1, &ch, 1); // Reintroduce some removed code here to get working. 
; c-lambda-end
;                     )
;           ssl)))
;     (if (fixnum? ret)
;         (result-or-abort 'get-char ssl ret #t)
;         ret)))

; Obsolete for the same reason as ssl-get-char above has been obsoleted.
; ; => u8vector or error code
; (define (ssl-read ssl bufsize)
;   ((c-lambda (SSL* int) scheme-object #<<c-lambda-end
; 
; // fprintf(stderr,"SSL_read: SSL_pending() gives %i for this SSL.\n",SSL_pending(___arg1));
; 
; SSL* ssl = ___arg1;
; int bufsize = ___arg2;
; char buf[bufsize]; // Since this is allocated on the stack only, it's fine.
; // Another way of implementing this reading would be to do it to a pre-allocated u8vector,
; // this would decrease object turnover = less load put on GC.
; int readbytes = SSL_read(ssl,buf,sizeof(buf));
; 
; if // Up to 2011-11-28: ((readbytes > 0) && (readbytes != -1) /* why is this one even necessary. */) {
;    // Replacement, inspired by https://github.com/engineyard/loudmouth/blob/master/loudmouth/lm-ssl-openssl.c row 398 & 243+ :
;    (readbytes > 0) {
;    // fprintf(stderr,"SSL_read: Returned success %i.\n",readbytes);
;    ___result = create_u8vector_from_buffer(buf,readbytes);
; } else {
;    // fprintf(stderr,"SSL_read: Returned failure %i.\n",readbytes);
;    ___result = ___FIX(readbytes);
; }
; 
; // fprintf(stderr,"SSL_read: Returning %i.\n",___result);
; 
; c-lambda-end
;              ) ssl bufsize))

; => non-negative fixnum = Bytes read
;    (negative fixnums never returned)
;    other = Status code for what beneficiary needs to do next
; This routine is for use only by code internal to this module, as it may return a command to perform a particular
; SSL operation.
(define (ssl-read-subu8vector-private u8vector start end ssl OpenSSL-error-vector*)
 (let ((result (ssl-read-subu8vector-private/internal u8vector start end ssl OpenSSL-error-vector*)))

   ; http://www.openssl.org/docs/ssl/SSL_read.html could not be more clear that only a positive response value
   ; communicated success - a zero response value indicates failure:
   ;
   ;      "The read operation was not successful. The reason may either be a clean shutdown due to a ``close notify'' alert sent
   ;      by the peer (in which case the SSL_RECEIVED_SHUTDOWN flag in the ssl shutdown state is set (see SSL_shutdown(3),
   ;      SSL_set_shutdown(3)). It is also possible, that the peer simply shut down the underlying transport and the shutdown is
   ;      incomplete. Call SSL_get_error() with the return value ret to find out, whether an error occurred or the connection was
   ;      shut down cleanly (SSL_ERROR_ZERO_RETURN).
   ;
   ;      SSLv2 (deprecated) does not support a shutdown alert protocol, so it can only be detected, whether the underlying
   ;      connection was closed. It cannot be checked, whether the closure was initiated by the peer or by something else."
   ; (note though this module never uses SSLv2)
   (if (fx< 0 result) ; (Supposedly result is always fixnum; it remains for us to see a result that's > or < 500M.)

       result

       (ssl-result OpenSSL-error-vector*))))

; Used by |ssl-read-subu8vectoru8vector/internal|.
(define (ssl-read-subu8vector-private/internal u8vector start end ssl OpenSSL-error-vector*)
  ((c-lambda (scheme-object unsigned-int unsigned-int SSL* OpenSSL-error-vector*) int #<<c-lambda-end

  void* u8v = ___CAST(void*,&___FETCH_U8(___BODY(___arg1),___INT(0)));
  int start = ___arg2;
  int end = ___arg3;
  SSL* ssl = ___arg4;
  OpenSSL_error_vector_t* error_vector = ___arg5;

  OpenSSL_clear_error_queue("SSL_read");

  ___result = SSL_read(ssl,u8v,end - start);

  // For more details about how the Unix read operation signals end of file on non-blocking sockets, see
  // http://linux.die.net/man/2/read .
  // fprintf(stdout,"SSL_read came back with OS errno %i. (EAGAIN=%i, ECONNRESET=%i, ENOTCONN=%i, ETIMEDOUT=%i.)\n",errno,EAGAIN,ECONNRESET,ENOTCONN,ETIMEDOUT); 

  // If the returned value was negative, an error happened. If so, capture it to the error vector.
  /* if (___result < 0) */
  OpenSSL_store_result_and_error(error_vector,ssl,___result);

c-lambda-end
   ) u8vector start end ssl OpenSSL-error-vector*))

; (ssl-write-subu8vector-private u8vector start end SSL* OpenSSL-error-vector*)
; => fixnum equal to end - start = success
;    fixnum, positive but less than end - start = ??? success ???
;    fixnum, zero = ??? success ???
;    (negative fixnum neever returned.)
;    other = status code for what the beneficiary needs to do next
(define (ssl-write-subu8vector-private u8vector start end ssl OpenSSL-error-vector*)
  ; (if-verbose (dbg "(ssl-write-subu8vector-private): Writing " u8vector " (length " (u8vector-length u8vector) ") start " start
  ;                  " end " end " to ssl " ssl "."))
  (let ((result (ssl-write-subu8vector-private/internal u8vector start end ssl OpenSSL-error-vector*)))
    ; (if-verbose (dbg "(ssl-write-subu8vector-private): Returned from write, got result " result "."))

    ; http://www.openssl.org/docs/ssl/SSL_write.html is extraordinarily clear that a positive result value
    ; is the only one that indicates success - a zero result indicates failure:
    ;
    ;      "The write operation was not successful. Probably the underlying connection was closed. Call SSL_get_error()
    ;      with the return value ret to find out, whether an error occurred or the connection was shut down cleanly
    ;      (SSL_ERROR_ZERO_RETURN).
    ;
    ;      SSLv2 (deprecated) does not support a shutdown alert protocol, so it can only be detected, whether the underlying
    ;      connection was closed. It cannot be checked, why the closure happened."
    ; (note though this module never uses SSLv2)
    (if (fx< 0 result) ; (Supposedly result is always fixnum; it remains for us to see a result that's > or < 500M.)

        result

        (ssl-result OpenSSL-error-vector*))))

(define ssl-write-subu8vector-private/internal (c-lambda (scheme-object int int SSL* OpenSSL-error-vector*) int #<<c-lambda-end

  ___SCMOBJ u8v = ___arg1;
  int start = ___arg2;
  int end = ___arg3;
  SSL* ssl = ___arg4;
  OpenSSL_error_vector_t* error_vector = ___arg5;

  OpenSSL_clear_error_queue("SSL_write");

  int offset = start;
  int length = end - start;
  // fprintf(stderr,"ssl#write: Initiated.\n");
  void* buf = ___CAST(void*,&___FETCH_U8(___BODY(u8v),___INT(0)));
  // fprintf(stderr,"ssl#write: Writing to SSL* ptr %i from ptr %i offset %i %i bytes.\n",(int) ___arg1,(int) ___arg2,___arg3,___arg4);
  ___result = SSL_write(ssl,(char*) buf + offset,length);

  // If the returned value was negative, an error happened. If so, capture it to the error vector.
  /* if (___result < 0) */
  OpenSSL_store_result_and_error(error_vector,ssl,___result);

c-lambda-end
                                      ))

; Commented out just because of the string->utf8-u8vector dependency - we don't use this procedure anyhow.
; (define (write-string ssl s)
;   (let ((v (string->utf8-u8vector s)))
;     (write ssl v 0 (u8vector-length v))))

; (No checking of the error queue done currently as it remains for us to see that queue elements may be added by this code.)
(define get-cipher (c-lambda (SSL*) char-string "___result = (char*) SSL_get_cipher(___arg1);"))

; Prepare SSL object to work in client mode
; (No checking of the error queue done currently as it remains for us to see that queue elements may be added by this code.)
(define SSL-set-connect-state (c-lambda (SSL*) void "SSL_set_connect_state"))

; Prepare SSL object to work in server mode
; (No checking of the error queue done currently as it remains for us to see that queue elements may be added by this code.)
(define SSL-set-accept-state (c-lambda (SSL*) void "SSL_set_accept_state"))

; The custom OpenSSL client port implementation.
; Gives the caller an u8vector port using which to access OpenSSL.
;
; ssl = The SSL* object that has been fed with the FD:s of the TCP socket port and otherwise is the one to use for this SSL channel.
;       This one is passed as the 'object' used with OpenSSL for it to perform the SSL channel logics and actual socket IO on.
; tcp-socket-port = The TCP socket port which the channel is performed over.
;                   This one is passed in order for this port mechanism to use to make Gambit's IO system wait for input/output-ability.
; fd = The File Descriptor of the tcp-socket-port's socket. Currently not used actually.
; ctx = The SSL-CTX* object which the ssl value (an SSL* object) was made within. This is passed on here as it'd be an error to
;       GC it while there's till an SSL* object that uses it.
(define (ssl-make-bidirectional-port ctx tcp-socket-port fd ssl)
  ;; note that the ctx parameter is never used but it is passed in order
  ;; to be present in the closure data of the various port functions
  ;; so it isn't garbage collected before the ports are all gone
  (let* ((to-beneficiary-thread #f)
         (from-beneficiary-thread #f)
         (read-mutex (make-mutex))
         (shutdown-mutex (make-mutex))

         (has-read? #f) (has-written? #f) ; We record these for use as a possible basis for differentiated error handling.

         (read-buffer-length openssl-port-beneficiary-from-host-buffer-size)
         (read-buffer (make-u8vector read-buffer-length))
         )
    (call-with-values
     (lambda () (open-u8vector-pipe '() ; '(char-encoding: UTF-8) - Gambit's current UTF-8 decoder is not reliable anyhow.
                                    '(char-encoding: UTF-8)))
     (lambda (our-port beneficiary-port)

       (define (shutdown-ssl)
         (mutex-lock! shutdown-mutex) ; If this procedure would be active in parallell, we would have a mess.
         (if-verbose (dbg "make-i/o-ports: Shutting down."))
         (if (and from-beneficiary-thread to-beneficiary-thread)
             (thread-terminate! (if (equal? (current-thread) to-beneficiary-thread) from-beneficiary-thread to-beneficiary-thread)))
         (set! ctx #f) ;; ensure that this reference is lost
         (dynamic-wind
          void
          (lambda ()
            (let loop ()
              (let ((result (shutdown ssl OpenSSL-error-vector*)))
                (case result
                  ((want-read)
                   (if-verbose (dbg "(shutdown ssl) gave 'want-read , waiting for more input data."))
                   (port-wait-for-input tcp-socket-port)
                   (if-verbose (dbg "Got more input data, continuing with shutdown."))
                   (thread-yield!)
                   (loop))
                  ((want-write)
                   (if-verbose (dbg "(shutdown ssl) gave 'want-write , waiting for output-ability."))
                   (port-wait-for-output tcp-socket-port)
                   (if-verbose (dbg "Got output-ability, continuing with shutdown."))
                   (thread-yield!)
                   (loop))
                  (else (if-verbose-dbg "(shutdown-ssl): SSL shutdown gave us unhandled result: " result
                                        " OpenSSL error queue dump: " (OpenSSL-error-vector-dump-queue OpenSSL-error-vector*)))
                  )))
            )
          (lambda ()
            (ssl-free ssl)
            (close-port tcp-socket-port)
            (if-verbose (dbg "make-i/o-ports: Shut down. (Terminating port's last thread.)"))
            (thread-terminate! (current-thread))
            )))

       (define (handle-syscall-response result caller)
         (cond
          ((eq? result 'syscall/eof-made-in-protocol-violating-way)
           
           ; http://www.openssl.org/docs/ssl/SSL_get_error.html says that "If ret == 0, an EOF was observed that violates the protocol".
           ;
           ; According to http://groups.google.com/group/mailing.openssl.users/browse_thread/thread/1e652cb1a03be017/4f6e295c6789f901?lnk=st&q=SSL_ERROR_SYSCALL#4f6e295c6789f901 ,
           ; there are instances when we get this error due to broken SSL implementations.
           ; Let's deal with it this way: If we have had successful SSL sending AND reading previously,
           ; let's consider this an OK way to close the connection. This is the occasion when we very much
           ; primarily tend to get this event - that the other side (in the instances I tested: the server)
           ; closed the connection.

           ;(if (and has-read? has-written?)
           ;(shutdown-ssl)
           (if-verbose (dbg "An EOF was observed that violates the protocol (upon " caller "). Shutting down connection."))
           ; (error (dumps "An EOF was observed that violates the protocol (upon " caller ")."))
           (shutdown-ssl)
           )

          ; http://www.openssl.org/docs/ssl/SSL_get_error.html says that
          ; "If ret == -1, the underlying BIO reported an I/O error (for socket I/O on Unix systems, consult errno for details)".
          ; I.e., there was an IO error.
          ;
          ; Let's treat this as an EOF - even if actually there's a more sophisticated error status dug down somewhere in the system,
          ; essentially that error's only meaning (other than for the possible purpose of presentation to the human user) is that
          ; it means we're at the EOF.
          ((eq? result 'syscall/underlying-bio-reported-io-error)
           (if-verbose (dumps "The underlying BIO reported an I/O error. (upon " caller ". the posix errno might give advice, "
                              "be careful to pull it before any other system call is made.) Shutting down connection."))
           ; (error (dumps "The underlying BIO reported an I/O error. (upon " caller ". the posix errno might give advice, "
           ;               "be careful to pull it before any other system call is made.)"))
           (shutdown-ssl)
           )
          (else
           ; result = 'syscall/other-io-error .
           (if-verbose (dumps "Syscall error reflected other IO error. (upon " caller ". result " result ".) Shutting down connection"))
           ; (error (dumps "Syscall error reflected other IO error. (upon " caller ". result is " result ".)"))
           (shutdown-ssl))))

       (define from-beneficiary-thread-thunk
         (lambda ()
           (run-dump-exceptions
            (let* ((buffer-length openssl-port-beneficiary-to-host-buffer-size)
                   (buffer (make-u8vector buffer-length)))
              (lambda ()
                (let loop ()
                  (if-verbose "OpenSSL writer-to-ssl-connection-thread thread started.")
                  ; Wait for there to be data to feed OpenSSL with on the beneficiary side.
                  ;(if-verbose (dbg "make-i/o-ports: Waiting for input from beneficiary."))
                  ;(port-wait-for-input our-port)
                  ;(if-verbose (dbg "make-i/o-ports: Had input from beneficiary."))

                  ; Read it
                  ; We make the read operation from the beneficiary block through requiring at least 1 byte to be read
                  ; on each operation, thus the read operation blocks until at least one byte has been passed to us.
                  ; Of course if the beneficiary has written more than one byte at once, we get more here directly also.
                  ; So this is an effective way of doing it.
                  (let* ((tmpx (if-verbose (dbg "make-i/o-ports: Going into read from beneficiary.")))
                         (bytes (read-subu8vector buffer 0 buffer-length our-port 1))) ; Let's go for at least one byte.

                    (if (eq? 0 bytes)
                        ; We got zero bytes from the read-subu8vector even while we sayd we require at least one, this means
                        ; that the beneficiary made at least |close-output-port|.
                        ;
                        ; We just end this thread here. Since the SSL link may still need writing (it may renegotiate anytime)
                        ; as part of receiving data and we do not know here that that part of the port is closed, we do not
                        ; close the output port part of the TCP port.
                        (begin
                          (if-verbose (dbg "from-beneficiary-port: Got zero bytes from read operation. This means the beneficiary "
                                           "closed at least the output part of the port. Therefore we know that this from-beneficiary-thread "
                                           "will never get any more work, so we close it now."))
                          (thread-terminate! (current-thread))
                          ))

                    (if-verbose (dbg "make-i/o-ports: Read from beneficiary. Going into write to ssl " ssl
                                     " data " (subu8vector buffer 0 bytes) " bytes " bytes "."))

                    ; Write it
                    (let write-again ((offset 0) (bytes bytes))
                      (let ((result (ssl-write-subu8vector-private buffer offset bytes ssl)))
                        (force-output tcp-socket-port 1) ; Level 1 means that the OS actually writes it, but does not wait for it to have been written.
                                                         ; Level 2 would be to wait for it to be written too, if supported. We don't want that.
                        (if-verbose (dbg "make-i/o-ports: Write of " bytes " bytes to TLS TCP port returned: " result))
                        (if (fixnum? result) ; (Note that if r is a fixnum here, it's never negative.)

                            ; We're running with the SSL_MODE_ENABLE_PARTIAL_WRITE option turned on, so fewer bytes may have been written
                            ; than commanded - this we need to handle here.
                            (let ((bytes (fx- bytes result))) ; Calculate how many bytes are left to write now.
                              (set! has-written? #t) ; We record this for using as a possible basis in differentiated error handling.
                              (if (eq? 0 bytes)
                                  ; No bytes left. Just leave this loop.
                                  (loop)
                                  ; At least one byte left.
                                  (begin
                                    (if-verbose
                                     (dbg "from-beneficiary-thread: Write succeeded but with fewer bytes than commanded (succeeded with "
                                          result ", " bytes " left). Ensuring that the TCP socket is write-able and continuing."))
                                    (port-wait-for-output tcp-socket-port)
                                    (if-verbose
                                     (dbg "from-beneficiary-thread: TCP socket said it's writeable, continuing with write."))
                                    (write-again (fx+ offset result) bytes))))
                            (begin
                              (if-verbose (dbg "make-i/o-ports: Looks like the write didn't go through on this try, as its result is " result ". Let's see what."))
                              (let ((meaning result))
                                ; (if-verbose (dbg "ssl#make-i/o-ports: The error is " meaning "."))
                                (cond

                                 ; 'zero-return :
                                 ; The TLS/SSL connection has been closed. [..] does not necessarily indicate that the underlying transport has been closed. 
                                 ((eq? meaning 'zero-return)
                                  (if-verbose
                                   (dbg "from-beneficiary-thread: had zero-return status, means connection is closed."))
                                  (shutdown-ssl))

                                 ; 'want-read :
                                 ; If I understand libCurl's lib/ssluse.c right here, it interprets both 'want-read and 'want-write
                                 ; as blocking signals, only to lead to select().
                                 ((eq? meaning 'want-read)
                                  (if-verbose
                                   (dbg "from-beneficiary-thread: Got want-read. Making a round of reading."))
                                  (do-read #f ; The mutexes in it make this safe.
                                           #t) ; called-by-write-thread?
                                  (if-verbose
                                   (dbg "from-beneficiary-thread: Round of reading done, retrying with the writing."))
                                  (thread-yield!)
                                  (thread-sleep! 0.1)
                                  (write-again offset bytes))

                                 ; 'want-write :
                                 ; If I understand libCurl's lib/ssluse.c right here, it interprets both 'want-read and 'want-write
                                 ; as blocking signals, only to lead to select().
                                 ((eq? meaning 'want-write)
                                  (if-verbose
                                   (dbg "from-beneficiary-thread: got want-write. Waiting for output deblock."))
                                  (port-wait-for-output tcp-socket-port)
                                  (if-verbose
                                   (dbg "from-beneficiary-thread: got output deblock. Trying to write again."))
                                  (write-again offset bytes))

                                 ; Syscall error:
                                 ((memq meaning '(syscall/eof-made-in-protocol-violating-way syscall/underlying-bio-reported-io-error syscall/other-io-error))
                                  (handle-syscall-response meaning "SSL_write")
                                  ; (Always leads either to shutting down the connection or to an (error).)
                                  )

                                 (else
                                  (error (dumps "Unknown return value to SSL_write " result))))))))))))))))

       ; This procedure is not threadsafe; it's for invocation by |to-beneficiary-thread| only.
       (define (ensure-from-beneficiary-thread)
         ; If write thread isn't started yet, start.
         (if (not from-beneficiary-thread)
             (begin
               (set! from-beneficiary-thread
                     (thread-start!
                      (make-thread from-beneficiary-thread-thunk
                                   '(openssl writer-to-ssl-connection-thread)))))))

       (define (do-read last-read-gave-zero-return? called-by-write-thread?)
         (dynamic-wind
          (lambda () #t)
          (lambda ()
            (mutex-lock! read-mutex)
            (let read-again ((at-reiteration 0))
              (if-verbose (dbg "make-i/o-ports: Reading from TLS TCP port " tcp-socket-port "."))
              (let ((result (ssl-read-subu8vector-private read-buffer 0 read-buffer-length ssl)))
                (if (fixnum? result) ; Got data? The only case in which result is a fixnum is if it's also a positive fixnum.

                    (let ((read-bytes result))
                      (set! has-read? #t) ; We record this for use as a possible basis for differentiated error handling.

                      (if-verbose
                       (trc-nobr "make-i/o-ports: Read from TLS SSL port. GOT " read-bytes " bytes: ")
                       (write-subu8vector read-buffer 0 read-bytes console-output-port)
                       (trc-nobr " / hex: " (subu8vector->hex-string read-buffer 0 read-bytes))
                       (trc-nobr " /GOT - writing to beneficiary\n"))

                      (if-verbose (dbg "make-i/o-ports: Read from TLS SSL port. Writing results to beneficiary (" read-bytes " bytes)."))
                      ; Yep, write to beneficiary. There's no need parameter to write-subu8vector.
                      (write-subu8vector read-buffer 0 read-bytes our-port)
                      ; (if-verbose (dbg "make-i/o-ports: Written the " v-length " bytes to beneficiary."))

                      )
                    ;(if (zero? v)
                    ; Had zero-length result.
                    ;#t

                    ; Had error return value.
                    (begin
                      (if-verbose (if (not (eq? result 'want-read))
                                      (dbg "make-i/o-ports: The read did not go through on this try, as its result is " result ". Let's see what.")))
                      (let ((meaning result))
                        ; (if-verbose
                        ;  (if (not (eq? meaning 'want-read)) (dbg "make-i/o-ports: Read from TCP port error meaning was " meaning ".")))
                        (cond

                         ; 'zero-return :
                         ; The TLS/SSL connection has been closed. [..] does not necessarily indicate that the underlying transport has been closed.
                         ((eq? meaning 'zero-return)
                          ; Condition inspired by http://www.mail-archive.com/openssl-users@openssl.org/msg55567.html .
                          (if last-read-gave-zero-return?
                              (shutdown-ssl)
                              'zero-return))

                         ; 'want-read :
                         ; http://www.openssl.org/docs/ssl/SSL_get_error.html says:
                         ; "The operation did not complete; the same TLS/SSL I/O function should be called again later. If, by then, 
                         ; the underlying BIO has data available for reading (if the result code is SSL_ERROR_WANT_READ) [..]
                         ; then some TLS/SSL protocol progress will take place".
                         ;
                         ; According to http://groups.google.com/group/mailing.openssl.dev/msg/cd6fbb747111cb8c ,
                         ; want-read return from SSL_read ( = read) indicates that we should select() ( = port-wait-for-input).
                         ; libCurl's lib/ssluse.c takes an a bit different approach, it says:
                         ; "there's data pending, re-invoke SSL_read()" "basically EWOULDBLOCK".
                         ((eq? meaning 'want-read)
                          ;(set! max-read-recursions (- max-read-recursions 1))
                          ;(if (< 0 max-read-recursions)
                          ;    (read-again)
                          ;    (if-verbose (dbg "Max read recursions reached, returning.")))
                          ; (thread-yield!) ; Just to not hog too much CPU; if OpenSSL had actually wanted to read something,
                          ;                 ; it'd have done so already.
                          (if-verbose (dbg "SSL_read returned want-read i.e. as soon as more data is available do SSL_read again, "
                                           "so going to wait for data i.e. select() now."))
                          #t
                          ; (if-verbose (dbg "SSL_read returned want-read. Reading again. This will be reiteration "
                          ;                  (1+ at-reiteration) "."))
                          ; (ensure-from-beneficiary-thread) ; Let's ensure the writer is going, so any data to be sent is sent.
                          ; (read-again (1+ at-reiteration))
                          )

                         ; 'want-write :
                         ; libCurl's lib/ssluse.c says:
                         ; "there's data pending, re-invoke SSL_read()" (in ossl_recv row 2643)
                         ; "basically EWOULDBLOCK". (in ossl_send row 2591)
                         ((eq? meaning 'want-write)

                          (if (not called-by-write-thread?)
                              (thread-yield!)) ; The thought here is to hand over the CPU to the scheduler as for it to have time to run a write op.

                          ; (if-verbose
                          ;  (dbg "make-i/o-ports: What on earth, SSL_READ returned want-write. Uh. Let's just ignore it for now, we'll write when we want to write in any case."))

                          ; 2012-11-16: Since it says it wants to write, let's ensure that it can write before we reiterate it.
                          (if-verbose
                           (dbg "to-beneficiary-thread: got want-write. Waiting for output deblock."))
                          (port-wait-for-output tcp-socket-port)
                          (if-verbose
                           (dbg "to-beneficiary-thread: got output deblock. Trying to write again."))

                          #t)

                         ; Syscall error:
                         ((memq meaning '(syscall/eof-made-in-protocol-violating-way syscall/underlying-bio-reported-io-error syscall/other-io-error))
                          ; (dbg "Got syscall response from SSL_read, ignoring.")
                          (handle-syscall-response meaning "SSL_read")
                          ; (Always leads either to shutting down the connection or to an (error).)
                          )

                         (else
                          (error (dumps "Unknown return value to SSL_read " result " (error code is " (SSL_get_error ssl result) ")."))))))))))
          (lambda ()
            (mutex-unlock! read-mutex))))

       (set! to-beneficiary-thread
             (thread-start!
              (make-thread
               (lambda ()
                 (run-dump-exceptions
                  (lambda ()
                    
                    (let loop ((last-read-gave-zero-return? #f))
                      ; On entry: Go into read, to finalize connect phase.
                      ; If a loop: Invoke OpenSSL to process data
                      (let* ((r (do-read last-read-gave-zero-return?
                                         #f)) ; called-by-write-thread?
                             (last-read-gave-zero-return? (eq? r 'zero-return)))

                        (ensure-from-beneficiary-thread)

                        ; Wait for there to be data to feed OpenSSL with on the TCP socket side. 
                        (if-verbose
                         (dbg "make-i/o-ports: Waiting for input from TLS TCP port."))
                        (port-wait-for-input tcp-socket-port)
                        (if-verbose
                         (dbg "make-i/o-ports: Had input from TLS TCP port, reading."))

                        (loop last-read-gave-zero-return?))))))
               '(openssl reader-from-ssl-connection-thread))))

       beneficiary-port))))

; This procedure's code is basically analogous to that of |ssl-make-bidirectional-port| , though while that one delivers an
; application-level SSL port within Gambit's IO system, this procedure delivers 'barebones' routines for operating an SSL
; port in both directions, with as high performance as possible. (E.g. as little object allocations as possible.)
;
; These routines have been designed to be as close to Gambit's internal IO routines as possible, without compromising performance.
;
; => a procedure that takes an argument (operation) and in response returns a procedure that takes the following arguments:
;    'read-subu8vector (u8v start end #!optional (need 1))
;    'read-substring (str start end #!optional (need 1))
;    'read-u8 ()
;    'write-subu8vector (u8v start end)
;    'display (string) - Writes the attached string in ISO-8859-1 encoding. The string may get stuck in an internal buffer
;                        until 'force-output .
;    'force-output () - Makes the same as (force-port port 1) in Gambit's IO system. Flushes any buffers to the output.
; 
; threadsafe? = #t = The procedure returned may be invoked by several threads concurrently
;               #f = The procedure returned will be invoked by one single thread in total only, concurrently. (For example,
;                    the thread for a HTTP connection within a web server.)
;
; This current implementation is designed for half-duplex communication performance at the level of software i.e. access in
; both directions is never made exactly concurrently. Though on the other hand it's designed for singlecore Gambit solely too,
; so that's always half-duplex in this sense, effectively anyhow. This can be worked around by going through the logics
; carefully and restructuring the mutexes to not include most orginary read&write ops too.
;
; Performance for read-u8 and read-substring and which are buffered, and read-subu8vector which is unbuffered, is achieved
; through that read-subu8vector on invocation receives any remaining buffer content first and then is run in the unbuffered
; mode. (To keep the code for this simple, if there's a buffer left then it is given as results to read-subu8vector until
; consumed; buffer content and unbuffered reading are never combined into the same read-subu8vector result.)
;
; Equally, performance for write-subu8vector which is buffered and write-subu8vector which is unbuffered, is achived
; through 'force-output of any buffer when any write-subu8vector is done.
(define u8v-empty '#u8())
(define (ssl-channel ctx tcp-socket-port fd ssl #!optional (threadsafe? #t) timeout-seconds)
  (let* ((mutex (and threadsafe? (make-mutex))) ; (No need to allocate it if we won't use it)

         (read-buffer-length openssl-port-beneficiary-from-host-buffer-size)
         (read-buffer (make-u8vector read-buffer-length))
         (read-buffer-at 0)
         (read-buffer-end 0)
         (last-read-gave-zero-return? #f) ; This, when it happens, has a symbolic significance within OpenSSL use, for us in determining
                                          ; to shut down a connection.

         ; We always leave the write buffer with at least one byte of free space in it.
         (write-buffer-length openssl-port-beneficiary-to-host-buffer-size)
         (write-buffer (make-u8vector write-buffer-length))
         (write-buffer-at 0)

         (has-read? #f) (has-written? #f) ; We record these for use as a possible basis for differentiated error handling.

         (not-done-a-read-yet? #t) ; It's part of the OpenSSL API:s convention that you must start every round of channel I/O
                                   ; with a read operation.

         (closed? #f)
         
         (OpenSSL-error-vector* (alloc-OpenSSL-error-vector)) ; Currently when reads and writes never are done concurrently,
                                                              ; having one shared error vector is fine.
         )

    (define-macro (perform . code)
      `(begin
         (if threadsafe? (mutex-lock! mutex))
         (let ((r (begin ,@code)))
           (if threadsafe? (mutex-unlock! mutex))
           r)))

    (define (wait-for-input #!optional (timeout-seconds timeout-seconds))
      (let ((r (port-wait-for-input tcp-socket-port (if timeout-seconds
                                                        (+ (seconds-since-epoch) timeout-seconds)
                                                        +inf.0))))
        (if (not r) (if-verbose-dbg "(wait-for-input): Did not get any new data before the timeout. (r=" r ")"))
        r))

    (define (wait-for-outputability #!optional (timeout-seconds timeout-seconds))
      (let ((r (port-wait-for-output tcp-socket-port (if timeout-seconds
                                                         (+ (seconds-since-epoch) timeout-seconds)
                                                         +inf.0))))
        (if (not r) (if-verbose-dbg "(wait-for-outputability): Did not get opportunity to send any more data before the timeout."))
        r))

    (define (shutdown-ssl-unidirectionally)
      (error "Not implemented"))

    (define (shutdown-ssl-bidirectionally)
      (if-verbose-dbg "(shutdown-ssl): Invoked. Calling SSL shutdown.")

      ; Make the SSL connection shut down. This is done by calling shutdown infinitely until it doesn't return
      ; 'want-read nor 'want-write anymore.
      ;
      ; Theoretically shutdown() could reiterate forever, so we set a max limit here.
      ; Until now though I saw it always end on the first try, so therefore we can as well give it 0.1x3 seconds in total
      ; to play with, if in some extremely rare exception it would take three iterations to complete.
      (let loop ((iteration 0) (already-shut-down-unilaterally? #f))
        (if (eq? 3 iteration)
            (begin
              (if-verbose-dbg "Reached max number of iterations so skipping asking OpenSSL to shut down, and just forcing it down.")
              )
            (let ((result (shutdown ssl OpenSSL-error-vector*)))
              (case result
                ((want-read)
                 (if-verbose-dbg "(shutdown-ssl): SSL shutdown wanted us to read, so we do, waiting for more input data.")
                 (wait-for-input 0.1)
                 (thread-yield!) ; Just to safeguard against over-use of the CPU
                 (if-verbose-dbg "(shutdown-ssl): Got more input data, reiterating shutdown.")
                 (loop (fx1+ iteration) already-shut-down-unilaterally?))
                ((want-write)
                 (if-verbose-dbg "(shutdown-ssl): SSL shutdown wanted us to write, so we do, waiting for write-ability.")
                 (wait-for-outputability 0.1)
                 (thread-yield!) ; Just to safeguard against over-use of the CPU
                 (if-verbose-dbg "(shutdown-ssl): Socket writable, reiterating shutdown.")
                 (loop (fx1+ iteration) already-shut-down-unilaterally?))

                ((unidirectionally)
                 (if already-shut-down-unilaterally?
                     (begin
                       (if-verbose-dbg "(shutdown-ssl): SSL says it shut down the connection unidirectionally now, but we want "
                                       "it to be bidirectionally, so reiterating another round.")
                       (loop 0 #t))
                     (begin
                       (if-verbose-dbg "(shutdown-ssl): SSL says it shut down unilaterally, but that's what it said already in the past?! "
                                       "Stopping listening to what OpenSSL says and forcing shutdown.")
                       )))

                ((#t) (if-verbose-dbg "(shutdown-ssl): SSL says it's shut down, so freeing the SSL object, closing the port, and shutting down internally."))

                (else (if-verbose-dbg "(shutdown-ssl): SSL shutdown gave us unhandled result: " result ". Treating as that the connection is shut down."
                                      " OpenSSL error queue dump: " (OpenSSL-error-vector-dump-queue OpenSSL-error-vector*)))
                ))))


      ; When shut down, we ask the GC to free the ssl object, and close the |tcp-socket-port|.
      (ssl-free ssl)
      (close-port tcp-socket-port)

      (set! closed? #t)
      (if-verbose-dbg "(shutdown-ssl): Now marked the connection as closed.")
      #!eof
      )

    ; As long as run with actually-get-data-required? = #t, it waits until it has read at least one byte, and then returns,
    ; or in case of IO error/EOF before even one byte has been read, returns #!eof .
    ;
    ; => positive fixnum = Success, result value is how many bytes have been read.
    ;    #!eof = The SSL connection has been shut down prior to that even one byte was read. EOF.
    (define (ssl-read u8v start end #!optional (actually-get-data-required? #t))
      (set! not-done-a-read-yet? #f)
      (let read-again ((at-reiteration 0))
        (if-verbose-dbg "(ssl-read): Going into ssl-read-subu8vector-private for bytes " start " to " end ".")
        (if closed?
            (begin
              (if-verbose-dbg "(ssl-read): closed? has been set, aborting call.")
              #!eof)
            ; Please note that we do not precede the |ssl-read-subu8vector-private| call with a |wait-for-input| check,
            ; which I suppose would imply a roundtrip through select(). This means that theoretically we can read on
            ; forever without ever making a roundtrip through select(), which is excellent from the performance point
            ; of view.
            (let ((result (ssl-read-subu8vector-private u8v start end ssl OpenSSL-error-vector*)))
              (if-verbose-dbg "(ssl-read): ssl-read-subu8vector-private returned " result ", handling.")
              ; Got data?
              (if (fixnum? result)

                  ; Yes.
                  (begin
                    (if-verbose-dbg "(ssl-read): Read was successful, got " result " bytes, returning.")
                    (set! has-read? #t) ; We record this for use as a possible basis for differentiated error handling.
                    result)

                  ; No. Handle the return value and hopefully reiterate the reading in a sec.
                  (cond

                   ; 'zero-return :
                   ; The TLS/SSL connection has been closed. [..] does not necessarily indicate that the underlying transport has been closed. 
                   ((eq? result 'zero-return)
                    (if (not actually-get-data-required?)
                        ; Just return.
                        (begin
                          (if-verbose-dbg "(ssl-read): Made a read and did not get any data. This particular invocation was not required to "
                                          "actually get data though, so we just return now.")
                          #!eof)
                        ; Condition inspired by http://www.mail-archive.com/openssl-users@openssl.org/msg55567.html .
                        (if last-read-gave-zero-return?
                            (begin
                              (if-verbose-dbg "(ssl-read): This read gave zero return and the last one did too, considering the SSL channel closed - shutting down.")
                              (shutdown-ssl-bidirectionally)) ; (returns #!eof)
                            (begin
                              (if-verbose-dbg "(ssl-read): Got zero return. Waiting for more input, and continuing.")
                              (set! last-read-gave-zero-return? #t)
                              ; Partially just as not to deem a connection broken just because it tried to make another read and got no data,
                              ; while there was no data anyhow.
                              (if (not (wait-for-input))
                                  (begin
                                    (if-verbose-dbg "(ssl-read): (wait-for-input) timed out, shutting down.")
                                    (shutdown-ssl-bidirectionally)) ; (returns #!eof)
                                  (begin
                                    (if-verbose-dbg "(ssl-read): Got more input, reading again.")
                                    (read-again (1+ at-reiteration))))))))

                   ; 'want-read :
                   ; http://www.openssl.org/docs/ssl/SSL_get_error.html says:
                   ; "The operation did not complete; the same TLS/SSL I/O function should be called again later. If, by then, 
                   ; the underlying BIO has data available for reading (if the result code is SSL_ERROR_WANT_READ) [..]
                   ; then some TLS/SSL protocol progress will take place".
                   ;
                   ; According to http://groups.google.com/group/mailing.openssl.dev/msg/cd6fbb747111cb8c ,
                   ; want-read return from SSL_read ( = read) indicates that we should select() ( = port-wait-for-input).
                   ; libCurl's lib/ssluse.c takes an a bit different approach, it says:
                   ; "there's data pending, re-invoke SSL_read()" "basically EWOULDBLOCK".
                   ((eq? result 'want-read)

                    ; It's typical that we get into this routine when opening a connection for a client and before writing
                    ; we go into a read here with actually-get-data-required? set to #f just to drive the handshaking - and
                    ; the read op returns 'want-read as to signal that it needs to get some input data and be invoked again
                    ; in order to actually read data.
                    ;
                    ; This is exactly what we do not want - just returning is exactly the right behavior here, when in that
                    ; mode.
                    ;
                    ; *Though*, (though what?)
                    (if (not actually-get-data-required?)
                        ; Just return.
                        (begin
                          (if-verbose-dbg "(ssl-read): Made a read and did not get any data. This particular invocation was not required to "
                                          "actually get data though (but was made to drive the handshaking), so we just return now. "
                                          "Though first let's wait for there to come in data, because that's the point ssl-read "
                                          "typically has here: It wants us to get it more incoming data. Though, by not reiterating with "
                                          "ssl-read here we still save ourselves of getting stuck in a loop.")
                          (if (not (wait-for-input))
                              (begin
                                (if-verbose-dbg "(ssl-read): (wait-for-input) timed out, shutting down.")
                                (shutdown-ssl-bidirectionally)) ; (returns #!eof)
                              (begin
                                (if-verbose-dbg "(ssl-read): Got more input, now returning.")
                                0)))

                        (begin
                          (if-verbose-dbg "(ssl-read): OpenSSL wants us to do a read, so we wait for more input (tcp-socket-port=" tcp-socket-port ")")
                          (if (not (wait-for-input))
                              (begin
                                (if-verbose-dbg "(ssl-read): (wait-for-input) timed out, shutting down.")
                                (shutdown-ssl-bidirectionally)) ; (returns #!eof)
                              (begin
                                (if-verbose-dbg "(ssl-read): Got more input, reading again.")
                                (read-again (1+ at-reiteration)))))))

                   ; 'want-write :
                   ; libCurl's lib/ssluse.c says:
                   ; "there's data pending, re-invoke SSL_read()" (in ossl_recv row 2643)
                   ; "basically EWOULDBLOCK". (in ossl_send row 2591)
                   ;
                   ; (2013-03-04 thought: Just in case someday we'd actually get to conclude that we actually need to
                   ; do a SSL_write here, then do note that SSL_write requires an input buffer of more than zero bytes,
                   ; and thus it would be like a big work to 
                   ((eq? result 'want-write)
                    (if-verbose-dbg "(ssl-read): OpenSSL wants us to a write... Or not. Waiting for more input data.")
                    ; We treat this as that OpenSSL actually wants more *INPUT* data, so therefore we wait for that and
                    ; then go into |ssl-read-subu8vector-private| again.
                    (if (not (wait-for-input))
                        (begin
                          (if-verbose-dbg "(ssl-read): (wait-for-input) timed out, shutting down.")
                          (shutdown-ssl-bidirectionally)) ; (returns #!eof)
                        (begin
                          (if-verbose-dbg "(ssl-read): Got more input data, continuing.")
                          (read-again (1+ at-reiteration)))))

                   ; Syscall error:
                   ((memq result '(syscall/eof-made-in-protocol-violating-way syscall/underlying-bio-reported-io-error syscall/other-io-error))
                    ; All of these effectively mean EOF so let's handle it as that.
                    (if-verbose-dbg "(ssl-read): Got syscall result " result ", treating as EOF, shutting down SSL."
                                    " OpenSSL error queue dump: " (OpenSSL-error-vector-dump-queue OpenSSL-error-vector*))
                    (let ((r (shutdown-ssl-bidirectionally))) ; (returns #!eof)
                      ; (error "Syscall, yet to be implemented." ((c-lambda () unsigned-long "ERR_get_error")))
                      r))

                   ; 'ssl :
                   ; http://www.openssl.org/docs/ssl/SSL_get_error.html :
                   ; "A failure in the SSL library occurred, usually a protocol error."
                   ((eq? result 'ssl)
                    (if-verbose-dbg "(ssl-read): OpenSSL let us know that the protocol handshake or alike failed. Shutting down this SSL channel."
                                    " OpenSSL error queue dump: " (OpenSSL-error-vector-dump-queue OpenSSL-error-vector*))
                    (shutdown-ssl-bidirectionally)) ; (returns #!eof)

                   (else
                    (error "Handler for ssl-read-subu8vector-private result not implemented" result))))))))

    ; => #!eof = End of file
    ;    other = Success
    (define (ssl-write u8v start end)
      (if (not (eq? start end))

          (let write-again ((start start) (reiteration? #f))
            (if-verbose-dbg "(ssl-write): To write bytes " start " to " end ". "
                            (if reiteration?
                                "Waiting for output port to be ready for output."
                                "Coming in from the outside so skipping the wait for ready for output step as we can guess it is available."))
            ; I don't know Gambit's internals perfectly but |wait-for-outputability| might just involve a roundtrip through select()
            ; which is an expensive operation so we don't want to do that unless necessary. Therefore, when we enter this procedure
            ; from the outside world (i.e. not an internal reiteration), we presume the socket is writable, and generally it will
            ; be really. This is an enormous optimization for write performance - effectively it means we could write infinitely
            ; without ever making roundtrips through |wait-for-outputability|, which is excellent.
            (if (and reiteration? (not (wait-for-outputability)))
                (begin
                  (if-verbose-dbg "(ssl-write): (wait-for-outputability) timed out, shutting down.")
                  (shutdown-ssl-bidirectionally))
                (begin
                  (if-verbose-dbg "(ssl-write): " (if reiteration?
                                                      "Output port ready for output, invoking |ssl-write-subu8vector-private|."
                                                      "Ok going in for |ssl-write-subu8vector-private|."))
                  (if closed?
                      (begin
                        (if-verbose-dbg "(ssl-read): closed? has been set, aborting call.")
                        #!eof)
                      (let ((result (ssl-write-subu8vector-private u8v start end ssl OpenSSL-error-vector*)))
                        (if-verbose-dbg "(ssl-write): Doing a force-output for it to be really written in the OS")
                        (force-output tcp-socket-port 1)
                        (if-verbose-dbg "(ssl-write): ssl-write-subu8vector-private returned " result ", handling.")
                        (if (fixnum? result) ; (Note that if r is a fixnum here, it is positive.)

                            ; We're running with the SSL_MODE_ENABLE_PARTIAL_WRITE option turned on, so fewer bytes may have been written
                            ; than commanded - this we need to handle here.
                            (let ((start (fx+ start result)))
                              (set! has-written? #t) ; We record this for using as a possible basis in differentiated error handling.
                              (if (eq? start end)
                                  ; Write done.
                                  (begin
                                    (if-verbose-dbg "(ssl-write): All bytes to be written written, returning.")
                                    #!void)
                                  ; Bytes left to write.
                                  (begin
                                    (if-verbose-dbg "(ssl-write): More bytes left to write (namely bytes " start " to " end "), writing again.")
                                    (write-again start #t))))

                            (cond

                             ; 'zero-return :
                             ; The TLS/SSL connection has been closed. [..] does not necessarily indicate that the underlying transport has been closed. 
                             ((eq? result 'zero-return)
                              (if-verbose-dbg "(ssl-write): Had zero-return status, means connection is closed.")
                              (shutdown-ssl-bidirectionally))

                             ; 'want-read :
                             ; If I understand libCurl's lib/ssluse.c right here, it interprets both 'want-read and 'want-write
                             ; as blocking signals, only to lead to select().
                             ((eq? result 'want-read)
                              (if-verbose-dbg "(ssl-write): Got want-read.")
                              ; Well we'll meet this want:
                              (let ((result (ssl-read empty-u8v 0 0 #f))) ; actually-get-data-required? = #f

                                ; If the channel is still open, then continue writing.
                                ; (The "worstcase scenario" here would be that the channel has been completely shut down since we entered |ssl-read|,
                                ; and that all the SSL channel's resources have been freed, so that an attempt to use them again here by calling
                                ; |write-again| would lead to a segmentation fault.
                                (if (eq? #!eof result)
                                    #!eof
                                    (write-again start #t))))

                             ; 'want-write :
                             ; If I understand libCurl's lib/ssluse.c right here, it interprets both 'want-read and 'want-write
                             ; as blocking signals, only to lead to select().
                             ((eq? result 'want-write)
                              (if-verbose-dbg "(ssl-write): Got want-write. Waiting for output deblock.")
                              (if (not (wait-for-outputability))
                                  (begin
                                    (if-verbose-dbg "(ssl-write): (wait-for-outputability) timed out, shutting down.")
                                    (shutdown-ssl-bidirectionally))
                                  (begin
                                    (if-verbose-dbg "(ssl-write): Got output deblock. Trying to write again.")
                                    (write-again start #t))))

                             ; Syscall error:
                             ((memq result '(syscall/eof-made-in-protocol-violating-way syscall/underlying-bio-reported-io-error syscall/other-io-error))
                              ; All of these effectively mean EOF so let's handle it as that.
                              (if-verbose-dbg "(ssl-write): Got syscall result " result ", treating as EOF, shutting down SSL."
                                              " OpenSSL error queue dump: " (OpenSSL-error-vector-dump-queue OpenSSL-error-vector*))
                              (shutdown-ssl-bidirectionally)
                              ; (error "Syscall, yet to be implemented." ((c-lambda () unsigned-long "ERR_get_error")))
                              )

                             ; 'ssl , as in broken handshake:
                             ((eq? result 'ssl)
                              (if-verbose-dbg "(ssl-write): Got 'ssl result, means broken handshake, shutting down SSL."
                                              " OpenSSL error queue dump: " (OpenSSL-error-vector-dump-queue OpenSSL-error-vector*))
                              (shutdown-ssl-bidirectionally))

                             (else
                              (error "Handler for ssl-write-subu8vector-private result not implemented" result)))))))))))

    ; This procedure is invoked only when the buffer is already empty.
    ; It loads up to the buffer size of data into the buffer.
    ;
    ; We provide the actually-get-data-required? as to be set to #f on calls that are intended to drive the handshaking forward, only.
    (define (load-to-buffer #!optional (actually-get-data-required? #t))
      (if-verbose-dbg "(load-to-buffer): Going into |ssl-read|. actually-get-data-required? " actually-get-data-required?)
      (let ((result (ssl-read read-buffer 0 read-buffer-length actually-get-data-required?)))
        (if (fixnum? result)
            (begin
              (if-verbose-dbg "(load-to-buffer) loaded " result " bytes.")
              (set! read-buffer-at 0)
              (set! read-buffer-end result))
            (if-verbose-dbg "(load-to-buffer) got error " result ", so no data loaded to buffer."))
        result))

    ; It's OpenSSL convention that you must start every SSL connection by making a read.
    (define (ensure-ready-for-write!)
      (if not-done-a-read-yet?
          ; Probably this is the most convenient way to make a read: that we simply issue a
          ; load of the read buffer.
          (begin
            (if-verbose-dbg "(ensure-ready-for-write!) commands a load of the buffer.")
            (load-to-buffer #f))))

    (define (write-write-buffer!)
      ; First, ensure the SSL channel is ready for writing.
      (if (eq? #!eof (ensure-ready-for-write!))
          #!eof

          (if (nonzero? write-buffer-at)
              (let ((bytes write-buffer-at))
                ; Reset write buffer pointer
                (set! write-buffer-at 0)

                (if-verbose-dbg "(write-write-buffer!): Writing " bytes " bytes.")
                (let ((r (ssl-write write-buffer 0 bytes))) ; (may return #!eof)
                  (if-verbose-dbg "(write-write-buffer!): Write returned " r
                                  ; as of 2012-11-16 - we didn't quite define any return values for write-ssl except for #!eof .
                                  " (in the current implementation, #!void is fine)"
                                  )
                  r))
              (if-verbose-dbg "(write-write-buffer!): No write buffer data so doing nothing.")
              )))

    ; Making a procedure that returns lambdas for the individual operations makes perfect sense: This way, if performing an
    ; operation many times, the beneficiary can have a reference directly to that operation within the code, meaning minimum
    ; overhead for those invocations.
    (lambda (operation)
      (case operation

        ; ## Reading routines
        ((read-subu8vector)
         (lambda (u8v start end #!optional (need 1)) ; Note: start = 5 end = 10 means, "please read into u8v positions 5, 6, 7, 8 and 9".
           (define end-at (fx1- end))                ; Therefore, we calculate end-at which is the last byte position we may actually read.
           (if (fx> need (fx- end start)) (error "Need value inconsistent; the interval start .. end is not that big." start end need))
           (perform
            (let loop ((start start) (need (or need 0)) (bytes-read 0))
              (if-verbose-dbg "(read-subu8vector): Invoked to read to #" (object->serial-number u8v) " from " start " to " end " need " need "."
                              (if closed? " Closed so returning #!eof." ""))
              (if closed?
                  #!eof ; TODO?

                  ; Do we have read buffer data?
                  (if (eq? read-buffer-end read-buffer-at) ; Note that this comparison should be as fast as making a boolean comparison.

                      ; No. Read data straight from SSL and into u8v
                      (let ((result (ssl-read u8v start end)))

                        ; Reached EOF?
                        ; Please note that this means that |ssl-read| didn't read a thing, at least not as far as it knows from OpenSSL
                        ; (and that should be pretty correct info really). It never performs performs another read after having successfully 
                        ; performed a firsto ne.
                        (if (eq? #!eof result)

                            ; Yes. Handle:
                            ; If more data still needed,
                            (if (fx< 0 need)

                                ; Handle accordingly,
                                ; (error "Not implemented handling of EOF result value while there's still more SSL data needed" need)
                                ; Let's just report EOF as to make it clear we failed to fulfill the requirement
                                ; (The beneficiary specified the need with a reason, so now that the need is not
                                ; satisfied, we better just report it as #!eof as in, need failed to be satisfied.)
                                (begin
                                  (if-verbose-dbg "(read-subu8vector): Got less than needed so returning eof.")
                                  #!eof)

                                ; And otherwise return number of bytes read to caller.
                                (begin
                                  (if-verbose-dbg "(read-subu8vector): Success, returning " bytes-read ".")
                                  bytes-read))

                            ; No.
                            (let* ((bytes-transacted result)
                                   (got-data? (fx< 0 bytes-transacted))
                                   (start (fx+ start bytes-transacted))
                                   (need (fxmax 0 (fx- need bytes-transacted)))
                                   (bytes-read (fx+ bytes-read bytes-transacted))
                                   )
                              ; More left to read?
                              (if (or (and got-data? (fx< start end-at))
                                      (fx< 0 need))
                                  ; Yes. Go read it.
                                  (loop start need bytes-read)
                                  ; No. Return total number of bytes read.
                                  bytes-read))))

                      ; Yes. Use the read buffer data as material for the read operation.
                      (let* ((buffer-bytes-available (fx- read-buffer-end read-buffer-at))
                             (read-op-can-accept-bytes (fx- end start)) ; Note that here we go with end and not end-at -
                                                                        ; If we're at start 7 and end 10 (and end-at thus is 9), then it is
                                                                        ; 10 - 7 = 3 bytes we want to read, namely positions 7, 8 and 9.
                             (transact-bytes (fxmin buffer-bytes-available read-op-can-accept-bytes)))

                        ; So here we do a copying operation and copying is expensive.
                        ; Notice how rare it is though: The read buffer gets filled up on read-substring and read-u8 use *only*, when
                        ; a sequence of |read-subu8vector| invocations are done, there's max one copying operation like this in total.
                        ;
                        ; Note that we presume use of Gambit's internal |subu8vector-move!|. Some versions of the (std misc/u8v) module
                        ; has a |subu8vector-move!| implementation too, that has another order of arguments..
                        (subu8vector-move! read-buffer
                                           read-buffer-at ; src-start
                                           (fx+ read-buffer-at transact-bytes) ; src-end
                                           u8v ; dst
                                           start ; dst-start
                                           )

                        (fx+! read-buffer-at transact-bytes)

                        (if-verbose-dbg "(read-subu8vector): Took " transact-bytes " bytes from the read buffer. "
                                        "Now the read buffer has another " (- read-buffer-end read-buffer-at) " bytes in it, "
                                        "and this read operation has a want to read another " (- end (+ start transact-bytes)) " bytes.")

                        (let ((new-start (fx+ start transact-bytes)))

                          (if (eq? new-start end)
                              (begin
                                (if-verbose-dbg "(read-subu8vector): Turns out this was all the beneficiary wanted to read. So returning.")
                                transact-bytes)
                              (begin
                                (if-verbose-dbg "(read-subu8vector): More data remains to be read, so reiterating as to read it with |ssl-read|.")
                                (loop new-start (fxmax 0 (fx- need transact-bytes)) (fx+ bytes-read transact-bytes))))))))))))
        ((read-substring)
         (lambda (str start end #!optional (need 1)) ; Note, start = 5 end = 10 means to read positions 5, 6, 7, 8 and 9.

           ; This check is not needed for the algorithm to be robust, it's only to inform the caller if they made a mistake.
           (if (fx> need (fx- end start)) (error "Need value inconsistent; the interval start .. end is not that big." start end need))

           (perform
            (if-verbose-dbg "(read-substring): Invoked to read to #" (object->serial-number str) " from " start " to " end " need " need "."
                            (if closed? " Closed so returning #!eof." ""))
            (if closed?
                #!eof ; TODO?
                (let read-chars ((at start) (chars-read 0))
                  ; Reached end?
                  (if (fx<= end at)
                      ; Yes.
                      (begin
                        (if-verbose-dbg "(read-subu8vector): Success, returning " chars-read ", data read is "
                                        (string->list (substring str start (+ start chars-read))) ".")
                        chars-read)
                      ; No.
                      (let retry-buffer-read ()
                        ; Have at least one character available in the buffer?
                        (if (fx> read-buffer-end read-buffer-at)
                            ; Yes. Take it.
                            (begin
                              (string-set! str at (integer->char (u8vector-ref read-buffer read-buffer-at)))
                              (fx+1! read-buffer-at)
                              (read-chars (fx1+ at) (fx1+ chars-read)))
                            ; No. Load buffer and reiterate.
                            (if (eq? #!eof (load-to-buffer
                                            ; actually-get-data-required? :
                                            (fx< chars-read need)
                                            ))
                                (begin
                                  (if-verbose-dbg "(read-subu8vector): EOF so returning #!eof.")
                                  #!eof)
                                (retry-buffer-read))))))))))
        ((read-u8)
         (lambda ()
           (perform
            (if-verbose-dbg "(read-u8): Invoked."
                            (if closed? " Closed so returning #!eof." ""))
            (if closed?
                #!eof ; TODO?
                (let loop ()
                  ; Got any data in read buffer?
                  (if (fx> read-buffer-end read-buffer-at)
                      ; Yes. Take a byte from there.
                      (let ((b (u8vector-ref read-buffer read-buffer-at)))
                        (fx+1! read-buffer-at)
                        (if-verbose-dbg "(read-u8): Success, got byte " b " (char " (integer->char b) ").")
                        b)
                      ; No. Load more data and reiterate.
                      (if (eq? #!eof (load-to-buffer))
                          (begin
                            (if-verbose-dbg "(read-u8): EOF so returning #!eof.")
                            #!eof)
                          (loop))))))))

        ((make-read-line-until-crlf)
         (lambda (#!optional return-eof-on-eof?)
           (let* ((buffer-length 512) ; A row of HTTP header, for which this routine is very much primarily used to read,
                                      ; is essentially never longer than this. We have a performance optimum as long as
                                      ; each line read is not longer than this, so this is why we care.
                                      ; Cookie rows could be longer than this but kind of generally are not.
                  (buffer (make-string buffer-length)))
             (lambda ()
               (perform
                (if-verbose-dbg "(make-read-line-until-crlf): Invoked."
                                (if closed? (string-append " Closed so returning " (if return-eof-on-eof? "#!eof" "\"\"") ".") ""))
                (if closed?

                    ; EOF, just return
                    (if return-eof-on-eof? #!eof "")

                    ; At least til now not EOF, go read.
                    (let ((read-char (lambda ()
                                       ; Got data in buffer for the next character?
                                       (let loop ()
                                         (if (fx> read-buffer-end read-buffer-at)
                                             ; Yes. Take it from there.
                                             (let ((c (integer->char (u8vector-ref read-buffer read-buffer-at))))
                                               (fx+1! read-buffer-at)
                                               c)
                                             ; No. Load more data and reiterate.
                                             (begin
                                               (if-verbose-dbg "(make-read-line-until-crlf): Going into |load-to-buffer|.")
                                               (if (eq? #!eof (load-to-buffer))
                                                   #!eof
                                                   (loop))))))))

                    (let stack-loop ((stack '()) (last-char #f))
                      (let char-loop ((at 0) (last-char #f))

                        (define (string-read-this-far)
                          (let ((s (if (eq? 0 at) "" (substring buffer
                                                                0
                                                                (if (eq? last-char #\return) (fx- at 1) at))))) ; As not to include the ASCII 13 we just read
                            (if (null? stack)
                                s
                                (apply-string-append `(,@(reverse stack) ,s)))))

                        (let ((char (read-char)))
                          (cond

                           ; Reached newline?
                           ((and (eq? char      #\newline)  ; ASCII 10, \n
                                 (eq? last-char #\return )) ; ASCII 13, \r

                            ; Yes. Return the string that's been read up to now.
                            (let ((s (string-read-this-far)))
                              (if-verbose-dbg "(make-read-line-until-crlf): Reached newline so returning " (string->list s) ".")
                              s))

                           ; Reached EOF?
                           ((eq? char #!eof)
                            (if return-eof-on-eof?
                                (begin
                                  (if-verbose-dbg "(make-read-line-until-crlf): EOF so returning #!eof.")
                                  #!eof)
                                (string-read-this-far)))

                           ; If neither of the above, then we have read a new char that is to be inserted into the read string.
                           (else
                            (string-set! buffer at char)
                            (fx+1! at)
                            (if (eq? at buffer-length)
                                (stack-loop (cons (string-copy buffer) stack))
                                (char-loop at char))))))))))))))

        ; ## Writing routines
        ((write-subu8vector)
         (lambda (u8v start end)
           (perform
            (if-verbose-dbg "(write-subu8vector): Invoked to write #" (object->serial-number u8v) " from " start " to " end "."
                            " Namely " (map integer->char (u8vector->list (subu8vector u8v start end))) "."
                            (if closed? " Closed so returning #!eof." ""))
            (if closed?
                #!eof ; TODO?
                (begin
                  ; If we accumulated any write buffer, write it to the socket now using |write-write-buffer!|.
                  ; This way we flush the remaining write buffer accumulated from any previous display:s here.
                  (if (not (eq? #!eof (write-write-buffer!)))

                      (let ((r (ssl-write u8v start end)))
                        (if-verbose-dbg "(write-subu8vector): |ssl-write| returned with " r " so returning that.")
                        r)
                      
                      (begin
                        (if-verbose-dbg "(write-subu8vector): EOF already on |write-write-buffer!| so returning #!eof.")
                        #!eof)))))))
        ((display)
         (lambda (v)

           (define (get-string)
             (cond ((string? v) v)
                   ((number? v) (number->string v))
                   ((symbol? v) (symbol->string v))
                   (else (error "Unknown format of value to display"))))

           (perform
            (if-verbose-dbg "(display): Invoked to display " (string->list (get-string)) "."
                            (if closed? " Closed so returning #!eof." ""))
            (if closed?
                #!eof ; TODO?

                (let ((string (get-string)))
                  ; Do a (ensure-ready-for-write!) just to ensure any required handshaking and so on is done.
                  ; Doing it here might indeed be microseconds before it actually needed to be done, but,
                  ; it is a question of a small amount of time anyhow and for the case that the SSL handshake
                  ; would fail or there'd be another error or exception, it's more logical that an error is
                  ; thrown on the first SSL operation than on one a bit into the code.
                  (if (eq? #!eof (ensure-ready-for-write!))
                      (begin
                        (if-verbose-dbg "(display): |ensure-ready-for-write| returned #!eof so returning #!eof.")
                        #!eof)

                      ; Write string into write buffer. If write buffer gets full during the process, write it to
                      ; the socket using |write-write-buffer!|.
                      (let ((end (string-length string)))
                        (let loop ((at 0))
                          ; If we not reached the end yet,
                          (if (not (eq? at end))
                              ; Then go on committing the next character:
                              ; Pick up the next character.
                              (let* ((byte (char->integer (##string-ref string at))))
                                ; Write it into the write buffer
                                (##u8vector-set! write-buffer write-buffer-at byte)
                                ; Increase the write buffer pointer
                                (fx+1! write-buffer-at)
                                ; Any space left in write buffer?
                                (if (eq? #!eof (if (eq? write-buffer-at write-buffer-length)
                                                   ; Nope, flush it.
                                                   (write-write-buffer!)))
                                    ; Return EOF.
                                    (begin
                                      (if-verbose-dbg "(display): |write-write-buffer| returned #!eof so returning #!eof.")
                                      #!eof)
                                    ; Recur.
                                    (loop (fx1+ at))))
                              (if-verbose-dbg "(display): Done, wrote " at " bytes.")
                              )))))))))
        ((force-output)
         (lambda ()
           (perform
            (if-verbose-dbg "(force-output): Invoked."
                            (if closed? " Closed so returning #!eof." ""))
            (if closed?
                #!eof ; TODO?
                (begin
                  ; Write any data that's waiting to be written to the socket
                  (if (eq? #!eof (write-write-buffer!))
                      (begin
                        (if-verbose-dbg "(force-output): |write-write-buffer| returned #!eof so returning #!eof.")
                        #!eof)
                      ; Ask OS to send it to other party now
                      (begin
                        (force-output tcp-socket-port 1)
                        (if-verbose-dbg "(force-output): Invoked.")
                        )))))))

        ; ## Other
        ((close-output-port)
         (lambda ()
           ; (perform
           (if-verbose-dbg "(close-output-port): Invoked."
                           (if closed? " Closed so returning #!eof." ""))
            (if closed?
                #!eof ; TODO?
                (begin
                  (if-verbose-dbg "(close-output-port): Invoked. Calling (write-write-buffer).")
                  (if (eq? #!eof (write-write-buffer!))
                      (begin
                        ; Is this really the smartest thing to do? I mean, if it signals EOF by some reason that would
                        ; happen to be repairable, then perhaps it's fixed bynow and |shutdown-ssl-unidirectionally|
                        ; would work?
                        (if-verbose-dbg "(close-output-port): |write-write-buffer| returned #!eof so returning #!eof.")
                        #!eof)
                      (begin
                        (if-verbose-dbg "(ssl-write): Calling (shutdown-ssl-unidirectionally).")
                        (shutdown-ssl-unidirectionally)))))))

        ; (Have not given any attention to implementing a close-input-port , though at first glance looking at OpenSSL's
        ; documentation, I did not see indications that there would be the basis for doing such an operation.)

        ; Close port bidirectionally
        ((close-port)
         (lambda ()
           ; (perform
           (if-verbose-dbg "(close-port): Invoked."
                           (if closed? " Closed so returning #!eof." ""))
           (if closed?
               #!eof ; TODO?
               (begin
                 (if-verbose-dbg "(close-port): Invoked. Calling (write-write-buffer).")
                 (write-write-buffer!)
                 (if-verbose-dbg "(close-port): Calling (shutdown-ssl-bidirectionally).")
                 (shutdown-ssl-bidirectionally)))))

        (else (error "Unknown operation" operation))
        ))))

; Note that Nginx has an interesting way of not using SSL_CTX_use_certificate_chain_file at
; SSL_CTX_use_certificate_chain_file ngx_ssl_certificate row 231-304. Not sure why it does.
;
; => 1 = success, other = fail
; (No checking of the error queue done currently as it remains for us to see that queue elements may be added by this code.)
(define ssl-ctx-use-certificate-chain-file/internal
  ; SSL_CTX *ctx, const char *file
  (c-lambda (SSL-CTX* char-string OpenSSL-error-vector*) int #<<c-lambda-end

SSL_CTX* ssl_ctx = ___arg1;
const char* file = ___arg2;
OpenSSL_error_vector_t* error_vector = ___arg3;
OpenSSL_clear_error_queue("SSL_CTX_use_certificate_chain_file");
___result = SSL_CTX_use_certificate_chain_file(ssl_ctx,file);

// Result value is 1 for success, or otherwise check error queue. https://www.openssl.org/docs/ssl/SSL_CTX_use_certificate.html
if (___result != 1) OpenSSL_CTX_error_capture(error_vector);

c-lambda-end
            ))

; (See comment above ssl-ctx-use-certificate-chain-file/internal)
;
; => #t = success, other = failure, ssl-result return value returned, and if OpenSSL-error-vector*
; was passed then the error queue has been copied into there.
(define (ssl-ctx-use-certificate-chain-file ssl-ctx* filename #!optional OpenSSL-error-vector*)
  (let* ((OpenSSL-error-vector* (or OpenSSL-error-vector* (alloc-OpenSSL-error-vector)))
         (r (ssl-ctx-use-certificate-chain-file/internal ssl-ctx* filename OpenSSL-error-vector*)))
    (if (eq? 1 r) #t (OpenSSL-error-vector-dump-queue OpenSSL-error-vector*))))

; => 1 = success, other = failure, ssl-result return value returned, and if OpenSSL-error-vector*
; was passed then the error queue has been copied into there.
(define ssl-ctx-use-certificate-file/internal
  ; SSL_CTX *ctx, const char *file, int type
  (c-lambda (SSL-CTX* char-string int OpenSSL-error-vector*) int #<<c-lambda-end

SSL_CTX* ctx = ___arg1;
const char* file = ___arg2;
int type = ___arg3;
OpenSSL_error_vector_t* error_vector = ___arg4;
OpenSSL_clear_error_queue("SSL_CTX_use_certificate_file");
___result = SSL_CTX_use_certificate_file(ctx,file,type);

// Result value is 1 for success, or otherwise check error queue. https://www.openssl.org/docs/ssl/SSL_CTX_use_certificate.html
if (___result != 1) OpenSSL_CTX_error_capture(error_vector);

c-lambda-end
            ))

; => #t = success, #f = failure, ssl-result return value returned, and if OpenSSL-error-vector*
; was passed then the error queue has been copied into there.
(define (ssl-ctx-use-certificate-file ssl-ctx* filename type #!optional OpenSSL-error-vector*)
  (let* ((OpenSSL-error-vector* (or OpenSSL-error-vector* (alloc-OpenSSL-error-vector)))
         (r (ssl-ctx-use-certificate-file/internal ssl-ctx* filename type OpenSSL-error-vector*)))
    (if (eq? 1 r) #t (OpenSSL-error-vector-dump-queue OpenSSL-error-vector*))))

; => 1 = success, other = failure, ssl-result return value returned, and if OpenSSL-error-vector*
; was passed then the error queue has been copied into there.
(define ssl-ctx-use-private-key-file/internal
  ; SSL_CTX *ctx, const char *file, int type 
  (c-lambda (SSL-CTX* char-string int OpenSSL-error-vector*) int #<<c-lambda-end

SSL_CTX* ssl_ctx = ___arg1;
const char* file = ___arg2;
int type = ___arg3;
OpenSSL_error_vector_t* error_vector = ___arg4;
OpenSSL_clear_error_queue("SSL_CTX_use_PrivateKey_file");
___result = SSL_CTX_use_PrivateKey_file(ssl_ctx,file,type);

// Result value is 1 for success, or otherwise check error queue. https://www.openssl.org/docs/ssl/SSL_CTX_use_certificate.html
if (___result != 1) OpenSSL_CTX_error_capture(error_vector);

c-lambda-end
            ))

; => #t = success, #f = failure, ssl-result return value returned, and if OpenSSL-error-vector*
; was passed then the error queue has been copied into there.
;
; Note: Apparently OpenSSL supports two Private Key file types. We implement one, and the other one
; is RSA, which is implemented as SSL_CTX_use_RSAPrivateKey_file . In the OpenSSL FFI this module
; was originally inspired by, SSL_CTX_use_RSAPrivateKey_file was actually used by default.
(define (ssl-ctx-use-private-key-file ssl-ctx* filename type #!optional OpenSSL-error-vector*)
  (let* ((OpenSSL-error-vector* (or OpenSSL-error-vector* (alloc-OpenSSL-error-vector)))
         (r (ssl-ctx-use-private-key-file/internal ssl-ctx* filename type OpenSSL-error-vector*)))
    (if (eq? 1 r) #t (OpenSSL-error-vector-dump-queue OpenSSL-error-vector*))))

(c-declare #<<c-declare

// Of https://github.com/engineyard/loudmouth/blob/master/loudmouth/lm-ssl-openssl.c row 111:
int ssl_verify_cb(int preverify_ok, X509_STORE_CTX* x509_ctx) {
    /* As this callback doesn't get auxiliary pointer parameter we
     * cannot really use this. However, we can retrieve results later. */
    return 1;
}

c-declare
)

;; switch verification of peer on or off
;; TODO This doesn't work
; level = 0 = SSL_VERIFY_NONE
;         1 = SSL_VERIFY_PEER
;         2 = SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT
; (No checking of the error queue done currently as it remains for us to see that queue elements may be added by this code.)
(define (ssl-set-verify! ctx level)
  ((c-lambda (SSL-CTX* int) void #<<c-lambda-end
             SSL_CTX_set_verify(___arg1,
                                ((___arg2 > 0) ? SSL_VERIFY_PEER | (___arg2 == 2 ? SSL_VERIFY_FAIL_IF_NO_PEER_CERT : 0) : SSL_VERIFY_NONE),
                                ssl_verify_cb);
c-lambda-end
    )
   ctx level))

;; load trusted root certificates into SSL context
;;(define (ssl-load-verify-root-certificates! obj pathname #!optional (dirname #f))
;;  (if pathname (##sys#check-string pathname))
;;  (if dirname (##sys#check-string dirname))
;;  (unless (eq?
;;           ((c-lambda (SSL-CTX* char-string char-string)
;;                      int "SSL_CTX_load_verify_locations")
;;            (unwrap-context obj)
;;            (if pathname (##sys#expand-home-path pathname) #f)
;;            (if dirname (##sys#expand-home-path dirname) #f))
;;           1)
;;    (abort 'ssl-load-verify-root-certificates! #f pathname dirname)))

;; load suggested root certificates into SSL context
;(define (ssl-load-suggested-certificate-authorities! obj pathname)
;  (##sys#check-string pathname)
;  (cond
;   (((c-lambda (char-string) SSL* "SSL_load_client_CA_file")
;     (##sys#expand-home-path pathname))
;    => (cut
;        (c-lambda (SSL-CTX* STACK_OF_X509_NAME*) void "SSL_CTX_set_client_CA_list")
;        (unwrap-context obj) <>))
;   (else
;    abort 'ssl-load-suggested-certificate-authorities! #f pathname)))

(define (open-ssl-tcp-client port-settings)
  (let* ((tmp1 (if-verbose (dbg "Opening connection.")))
         (port (open-tcp-client port-settings))
         (tmp2 (if-verbose (dbg "Setting up SSL."))))
    (wrap-tcp-port-into-ssl-client-port port)))

; Sometimes you want to make a TLS connection out of a tcp-client port first after having done some
; initializing communication already, this is what this procedure is for. Such is the case for
; XMPP, for instance.
(define (wrap-tcp-port-into-ssl-port-or-channel/internal port ssl-ctx receiver-proc #!optional server?)
  (if-verbose (dbg "wrap-tcp-port-into-ssl-port-or-channel/internal invoked. port " port " ssl-ctx " ssl-ctx " server? " server?))

  (let ((ssl (ssl-new ssl-ctx)))
    ; http://h71000.www7.hp.com/doc/83final/ba554_90007/ch04s03.html section "Performing an SSL Handshake with SSL_read and SSL_write (Optional)"
    ; says that when these are set, calling ssl-accept/ssl-connect is not needed anymore, it's done during reads/writes.
    ;
    ; http://git.berlios.de/cgi-bin/gitweb.cgi?p=futhark;a=blob;f=src/encode/openssl-ports.scm;h=2651e358645a530354894a8f098fdcedbda8584f;hb=HEAD
    ; gives an example of not running either of these two.
    (if server?
        (SSL-set-accept-state ssl)
        (SSL-set-connect-state ssl))

    ; Inspired by Curl_socket_ready in curl's lib/ssluse.c:
    ; When Gambit creates a new TCP client, it does not wait for the connection to actually become established,
    ; so at least there this whould be required.
    (if-verbose (dbg "Waiting for socket to be ready for output. "
                     ; "(Don't know if this one is needed.)"
                     ))
    (port-wait-for-output port)

    (if-verbose (dbg "It's ready for output. Getting FD."))

    (let ((fd (tcp-socket-port->fd port)))

      (if-verbose (dbg "FD is " fd ". Setting."))
      (if (not (set-fd! ssl fd))
          (error "Could not set file descriptor on SSL structure."))

      (if ssl-perform-explicit-connect/accept-on-open?
          (begin
            (if-verbose (dbg (if server? "Accepting." "Connecting.")))
            (let* ((OpenSSL-error-vector* alloc-OpenSSL-error-vector)
                   (r (if server?
                          (ssl-accept  ssl OpenSSL-error-vector*)
                          (ssl-connect ssl OpenSSL-error-vector*))))
              (if (not (memq r '(#t want-read want-write)))
                  (error (dumps (if server? "SSL_connect" "SSL_accept") " returned " r
                                " rather than 1 or want-read or "
                                "want-write. Fatal error, for instance "
                                "with handshake."
                                
                                )))))
          (if-verbose (dbg "We do not make an " (if server? "SSL_accept" "SSL_connect") " now on open but let OpenSSL do this implicitly "
                           "through read/write ops.")))

      (if-verbose (dbg "Creating port wrapper."))
      (receiver-proc ssl-ctx port fd ssl))))
  
(define (wrap-tcp-port-into-ssl-client-port port)
  (wrap-tcp-port-into-ssl-port-or-channel/internal port client-sockets-ctx ssl-make-bidirectional-port #f))

(define (wrap-tcp-port-into-ssl-server-port port ssl-ctx)
  (wrap-tcp-port-into-ssl-port-or-channel/internal port ssl-ctx ssl-make-bidirectional-port #t))

; We allow a ssl-ctx parameter here, as the client very well may want custom security settings on the ssl-ctx.
(define (tcp-port->ssl-client-channel port #!optional (ssl-ctx client-sockets-ctx) (threadsafe? #t) timeout-seconds)
  (wrap-tcp-port-into-ssl-port-or-channel/internal port ssl-ctx
                                                   (lambda (ctx tcp-socket-port fd ssl) (ssl-channel ctx tcp-socket-port fd ssl threadsafe? timeout-seconds))
                                                   #f))

(define (tcp-port->ssl-server-channel port ssl-ctx #!optional (threadsafe? #t) timeout-seconds)
  (wrap-tcp-port-into-ssl-port-or-channel/internal port ssl-ctx
                                                   (lambda (ctx tcp-socket-port fd ssl) (ssl-channel ctx tcp-socket-port fd ssl threadsafe? timeout-seconds))
                                                   #t))

(define (call-with-ssl-tcp-client port-settings proc)
  (let* ((port (open-ssl-tcp-client port-settings))
         (result (proc port)))
    (close-port port)
    result))

(define openssl-version (c-lambda () nonnull-char-string "___result = (char*) SSLeay_version(SSLEAY_VERSION);"))

(define (openssl-dbg-test) (dbg "Verbosity " openssl-verbose))

(define (make-server-ssl-ctx #!optional (protocol default-server-ssl-version))
  (let ((ctx (ssl-ctx-new protocol #t)))
    
    ; ctx-set-options options documented at https://www.openssl.org/docs/ssl/SSL_CTX_set_options.html .

    ; Note that ssl-ctx-new obligatorily enables the SSL_MODE_ENABLE_PARTIAL_WRITE mode.

    ; Just to keep the deprecated SSLv2 out of the house. In some OpenSSL distros it's not even supported.
    ;
    ; From lib/ssluse.c of libCurl:
    ; disable SSLv2 in the default case (i.e. allow SSLv3 and TLSv1)
    (ctx-set-options ctx OpenSSL-OP-NO-SSLv2)
    
    ; From Nginx' https://github.com/git-mirror/nginx/blob/master/src/event/ngx_event_openssl.c ngx_ssl_create row 165-179:
    (ctx-set-options ctx OpenSSL-OP-SSLREF2-REUSE-CERT-TYPE-BUG)
    (ctx-set-options ctx OpenSSL-OP-MICROSOFT-BIG-SSLV3-BUFFER)

    ; Nginx comments: "this option allow a potential SSL 2.0 rollback (CAN-2005-2969)"
    (ctx-set-options ctx OpenSSL-OP-MSIE-SSLV2-RSA-PADDING)

    (ctx-set-options ctx OpenSSL-OP-SSLEAY-080-CLIENT-DH-BUG)
    (ctx-set-options ctx OpenSSL-OP-TLS-D5-BUG)
    (ctx-set-options ctx OpenSSL-OP-TLS-BLOCK-PADDING-BUG)

    (ctx-set-options ctx OpenSSL-OP-DONT-INSERT-EMPTY-FRAGMENTS)

    (ctx-set-options ctx OpenSSL-OP-SINGLE-DH-USE) ; Generate new key on each connection when using Diffie Hellman exchange.
                                                   ; Recommended by OpenSSL docs to do so, requires just a little bit of CPU.
                                                   ; For some additional note see http://www.lists.aldigital.co.uk/apache-ssl/msg02289.html ,
                                                   ; who for instance proposes that OpenSSL should really have this enabled by default.

    ; From lib/ssluse.c of libCurl:
    ; "Not sure it's needed to tell SSL_connect() that socket is
    ; non-blocking. It doesn't seem to care, but just return with
    ; SSL_ERROR_WANT_x."

    (ctx-ctrl ctx OpenSSL-BIO-SET-NBIO 1)
    
    ; Inspired by Nginx' https://github.com/git-mirror/nginx/blob/master/src/event/ngx_event_openssl.c ngx_ssl_create row 210:
    ; There is some comment on the web that some user experienced blocking of their socket when doing this - that should not
    ; happen for us as the sockets we use always are nonblocking.
    (ctx-set-read-ahead ctx 1)

    ctx))

(define (make-client-ssl-ctx #!optional (protocol default-client-ssl-version))
  (ssl-ctx-new protocol #f)) ; sslv2-or-v3 #f))   #f = client (not server).

(define (cert-files->server-ssl-ctx ssl-certificate-chain-file ssl-certificate-file ssl-private-key-file #!rest make-server-ssl-ctx-args)
  (let* ((OpenSSL-error-vector* (alloc-OpenSSL-error-vector))
         (ssl-ctx (apply make-server-ssl-ctx make-server-ssl-ctx-args)))

    (if ssl-certificate-chain-file

        ; For more info see comment above definition of |ssl-ctx-use-certificate-chain-file|.
        (let ((r (ssl-ctx-use-certificate-chain-file ssl-ctx ssl-certificate-chain-file OpenSSL-error-vector*)))

          (if (not (eq? #t r))
              (error "Failed to load SSL certificate chain file" ssl-certificate-chain-file r))))

    (if ssl-certificate-file
        (let ((r (ssl-ctx-use-certificate-file ssl-ctx ssl-certificate-file OpenSSL-FileType-PEM OpenSSL-error-vector*)))
          (if (not (eq? #t r))
              (error "Failed to load SSL certificate file"       ssl-certificate-file       r))))

    (if ssl-private-key-file
        (let ((r (ssl-ctx-use-private-key-file ssl-ctx ssl-private-key-file OpenSSL-FileType-PEM OpenSSL-error-vector*)))
          (if (not (eq? #t r))
              (error "Failed to load SSL private key file"       ssl-private-key-file       r))))

    ssl-ctx))

; ## GLOBAL INITIALIZATION ROUTINES 
(define client-sockets-ctx (make-client-ssl-ctx))

; (Note that ssl-ctx-new obligatorily enables the SSL_MODE_ENABLE_PARTIAL_WRITE mode.)

; ctx-set-options options documented at https://www.openssl.org/docs/ssl/SSL_CTX_set_options.html .

; From lib/ssluse.c of libCurl:
; OpenSSL contains code to work-around lots of bugs and flaws in various
; SSL-implementations. SSL_CTX_set_options() is used to enabled those
; work-arounds. The man page for this option states that SSL_OP_ALL enables
; all the work-arounds and that "It is usually safe to use SSL_OP_ALL to
; enable the bug workaround options if compatibility with somewhat broken
; implementations is desired."

; (ctx-set-options sockets-ctx OpenSSL-OP-ALL)

; Just to keep the deprecated SSLv2 out of the house. In some OpenSSL distros it's not even supported.
;
; From lib/ssluse.c of libCurl:
; disable SSLv2 in the default case (i.e. allow SSLv3 and TLSv1)
(ctx-set-options client-sockets-ctx OpenSSL-OP-NO-SSLv2)

; As per https://github.com/engineyard/loudmouth/blob/master/loudmouth/lm-ssl-openssl.c row 309,
; needed to get TLS connection with some broken Java SSL implementation going.
(ctx-set-options client-sockets-ctx OpenSSL-OP-NO-TICKET)

; As per Nginx' https://github.com/git-mirror/nginx/blob/master/src/event/ngx_event_openssl.c ngx_ssl_create row 160-164,
(ctx-set-options client-sockets-ctx OpenSSL-OP-MICROSOFT-SESS-ID-BUG)
(ctx-set-options client-sockets-ctx OpenSSL-OP-NETSCAPE-CHALLENGE-BUG)

(let ((r (ctx-set-default-verify-paths client-sockets-ctx))) ; Inspired by https://github.com/engineyard/loudmouth/blob/master/loudmouth/lm-ssl-openssl.c row 324
  ; (dbg "ctx-set-default-verify-paths returned " r)
  (if (>= 0 r) (error "ctx-set-default-verify-paths failed" r)))
(ssl-set-verify! client-sockets-ctx 1) ; "

; From lib/ssluse.c of libCurl:
; "Not sure it's needed to tell SSL_connect() that socket is
; non-blocking. It doesn't seem to care, but just return with
; SSL_ERROR_WANT_x."

(ctx-ctrl client-sockets-ctx OpenSSL-BIO-SET-NBIO 1)

