; Sack IO primitives module
; Copyright (C) 2012 Mikael More
;
; This module defines the IO primitives and their handling procedures, for the Sack web server.
; (May be extended for the http-client at a future time.)
;
; The purpose is to provide "standard" primitives for ordinary HTTP access, while opening up for
; special purpose primitives as for special delivery such as via HTTPS.

; A set of io-primitives is specific for a TCP connection. It's a set of closures for performing
; the different IO operations such as read line with ISO-8859-1 encoding, write u8vector, etc.
;
; The routines are:
;
;      Name                      Arguments & Result                  Behavior
;      read-subu8vector          (u8v start end #!optional (need 1)) Read u8vector data
;                                => bytes read or #!eof
;      read-substring            (str start end #!optional (need 1)) Read string data as ISO-8859-1
;                                => bytes read or #!eof
;      read-u8                   () => byte or #!eof                 Read a byte
;      write-subu8vector         (u8v start end)                     Write u8vector data
;                                => #t = succeeded as in written, #f = failed as attempt to write failed                                 ENSURE ACCORDANCE (passes #!eof now???)
;      display                   (v)                                 Write string, number or symbol as ISO-8859-1
;                                => #t = succeeded as in written, #f = failed as attempt to write failed                                 ENSURE ACCORDANCE (passes #!eof now???)
;      force-output              ()                                  Flush any write buffer to network
;                                => #t = succeeded as in done, #f = failed as in port broken                                             ENSURE ACCORDANCE (passes #!eof now???)
;      close-port                ()                                  Close port
;                                => #t = succeeded as in done, #f = failed as in was already closed (incl. by other party?)              ENSURE ACCORDANCE (passes #!eof now???)
;      make-read-line-until-crlf (#!optional return-eof-on-eof?)     Read line until reaching crlf. Permissive behavior.
;                                => procedure taking no args         The procedure returned may be used in a single thread only and thus
;                                => string or maybe #!eof, see desc  reuse the same buffers across invocations.
;                                                                    If return-eof-on-eof? is set then #!eof is returned in case of EOF.
;                                                                    If not, string reading is just ended and result is returned.
;
; The current format for the io-primitives structure is a vector. This is done as to provide the highest speed of access.
; (The vector access can be completely inlined in beneficiary code. Re |define-type| accessors I believe those remain
; non-inlined procedure calls always, leading to lower potential performance. Because of the paramount need of performance
; in this function, together with the very limited scope in which io-primitives structures are passed around, vectors are
; very suitable.)
; 

(declare (block) (standard-bindings) (extended-bindings) (mostly-fixnum))

(cond-expand
 
 ; Black Hole use
 (black-hole
  (import (only: (std srfi/13) reverse-list->string) gambit-noexception-io))

 ; Vanilla Gambit use
 (else
  (define (reverse-list->string clist)
    (let* ((len (length clist))
           (s (make-string len)))
      (do ((i (- len 1) (- i 1))   (clist clist (cdr clist)))
          ((not (pair? clist)))
          (string-set! s i (car clist)))
      s))
  (include "gambit-noexception-io.scm")
  ))

; ## Accessors
(define-macro (io-primitives-read-subu8vector     v) `(vector-ref ,v 0))
(define-macro (io-primitives-read-substring       v) `(vector-ref ,v 1))
(define-macro (io-primitives-read-u8              v) `(vector-ref ,v 2))
(define-macro (io-primitives-write-subu8vector    v) `(vector-ref ,v 3))
(define-macro (io-primitives-display              v) `(vector-ref ,v 4))
(define-macro (io-primitives-force-output         v) `(vector-ref ,v 5))
(define-macro (io-primitives-close-port           v) `(vector-ref ,v 6))
(define-macro (io-primitives-read-line-until-crlf v) `(vector-ref ,v 7))

(define-macro (with-io-primitives io-primitives . code)
  `(let ((read-subu8vector     (io-primitives-read-subu8vector     ,io-primitives))
         (read-substring       (io-primitives-read-substring       ,io-primitives))
         (read-u8              (io-primitives-read-u8              ,io-primitives))
         (write-subu8vector    (io-primitives-write-subu8vector    ,io-primitives))
         (display              (io-primitives-display              ,io-primitives))
         (force-output         (io-primitives-force-output         ,io-primitives))
         (close-port           (io-primitives-close-port           ,io-primitives))
         (read-line-until-crlf/s   ((io-primitives-read-line-until-crlf ,io-primitives) #f))  ; return-eof-on-eof? = #f.
         (read-line-until-crlf/eof ((io-primitives-read-line-until-crlf ,io-primitives) #t))) ; return-eof-on-eof? = #t.
     ,@code))

; ## Constructor (generic)
; Make an io-primitives structure for use by Sack, given thunks for the respective primitive
(define (make-io-primitives read-subu8vector read-substring read-u8 write-subu8vector display force-output close-port #!optional read-line-until-crlf)
  (vector read-subu8vector read-substring read-u8 write-subu8vector display force-output close-port read-line-until-crlf))

; ## Constructor (for using Gambit's IO system)
; Make an io-primitives structure that delivers Gambit's IO system's primitives, given a gambit-port
; on which to do the IO operations
(define (with-standard-primitives gambit-port timeout-seconds thunk #!optional (threadsafe? #t))
 (let* ((io-primitives (make-standard-primitives gambit-port timeout-seconds threadsafe?))
        (r (thunk io-primitives)))

   ((io-primitives-close-port io-primitives))

   r))

(define (make-standard-primitives gambit-port timeout-seconds #!optional (threadsafe? #t))

  (define not-closed? #t)
  (define eof-marker-if-closed #f)
  (define mutex (and threadsafe? (make-mutex)))
  
  (define display-buffer-size 1024)
  (define display-buffer #f)

  (define-macro (perform . code)
    `(begin
       (if threadsafe? (mutex-lock! mutex))
       (let ((r (begin ,@code)))
         (if threadsafe? (mutex-unlock! mutex))
         r)))

  (define (close)
    (if not-closed?
        (begin
          (close-port/noexception gambit-port)
          (set! not-closed? #f)
          (set! eof-marker-if-closed #!eof))))

  (input-port-timeout-set! gambit-port
                           timeout-seconds
                           (lambda ()
                             (close)
                             #t)) ; #t signals to Gambit that it should not cancel the IO operation that was ongoing on the
                                  ; timeout. We don't signal that because for output operations that would raise an exception.
                                  ; Instead, we hope Gambit identifies that the port is now closed and ends the respective
                                  ; operation without throwing an exception.
  (output-port-timeout-set! gambit-port
                            timeout-seconds
                            (lambda ()
                              (close)
                              #t)) ; #t signals to Gambit that it should not cancel the IO operation that was ongoing on the
                                   ; timeout. We don't signal that because for output operations that would raise an exception.
                                   ; Instead, we hope Gambit identifies that the port is now closed and ends the respective
                                   ; operation without throwing an exception.

  (let ((io-primitives
         (make-io-primitives
          ; read-subu8vector :
          (lambda (u8v start end #!optional (need 1))
            (perform
             (or eof-marker-if-closed
                 (read-subu8vector/noexception u8v start end gambit-port need)))) ; TODO: Guarantee #!eof return value on EOF
          ; read-substring :
          (lambda (str start end #!optional (need 1))
            (perform
             (or eof-marker-if-closed
                 (read-substring/noexception str start end gambit-port need)))) ; TODO: Guarantee #!eof return value on EOF
          ; read-u8 :
          (lambda ()
            (perform
             (or eof-marker-if-closed
                 (read-u8/noexception gambit-port))))
          ; write-subu8vector :
          (lambda (u8v start end)
            (perform
             (or eof-marker-if-closed
                 (write-subu8vector u8v start end gambit-port))))
          ; display
          (lambda (v)
            (perform
             (or eof-marker-if-closed
                 (begin
                   ; (display v gambit-port)
                   ;
                   ; There seems to be some issues with display:
                   ;      0  ##mutex-lock-out-of-line!
                   ;      1  ##make-device-port
                   ;      2  ##write-generic-to-character-port
                   ;      3  ##display
                   ;      4  io-primitives#with-standard-primitives         "lib/sack/src/io-primitives.scm"@50:1          (display 8#v 1#gambit-port)
                   ;
                   ; Now, we only need a very basic display implementation anyhow, namely one that works exactly like this:
                   ;
                   ; Implementation that has large memory turnover:
                   ; (let* ((v (cond ((symbol? v) (symbol->string v))
                   ;                 ((number? v) (number->string v))
                   ;                 (else        v                 )))
                   ;        (l (string-length v))
                   ;        (u8v (make-u8vector l)))
                   ;   (let loop ((at 0))
                   ;     (if (not (eq? at l))
                   ;         (begin
                   ;           (u8vector-set! u8v at (char->integer (string-ref v at)))
                   ;           (loop (fx+ at 1)))))
                   ;   (write-subu8vector u8v 0 l gambit-port)
                   ;   )
                   ;
                   ; Implementation with less memory turnover:
                   (if (not display-buffer) (set! display-buffer (make-u8vector display-buffer-size)))
                   (let* ((v (cond ((symbol? v) (symbol->string v))
                                   ((number? v) (number->string v))
                                   (else        v                 )))
                          (l (string-length v)))

                     (define (write-chars! string-start-idx count)
                       ; (dbg "(display): (write-chars!): Displaying " string-start-idx " and to " count ".")
                       (let loop ((at string-start-idx) (buffer-at 0))
                         (if (not (eq? buffer-at count))
                             (begin
                               (u8vector-set! display-buffer buffer-at (char->integer (string-ref v at)))
                               ; (dbg "(display): (write-chars!): Copying " at " to " buffer-at ".")
                               (loop (fx+ at 1) (fx+ buffer-at 1)))))
                       ; (dbg "(display): (write-chars!): Writing buffer 0 to " count ".")
                       (write-subu8vector display-buffer 0 count gambit-port))

                     (let loop ((start-at 0))
                       (let ((end-at (fx+ start-at display-buffer-size)))
                         (if (fx> end-at l)
                             (begin
                               ; (dbg "(display): Am at last segment, " start-at " and " (fx- l start-at) " onward.")
                               (write-chars! start-at (fx- l start-at)))
                             (if (eq? #!eof (write-chars! start-at display-buffer-size))
                                 #!eof
                                 (begin
                                   ; (dbg "(display): Am at a segment, " start-at " and " display-buffer-size " onward.")
                                   (loop (fx+ start-at display-buffer-size)))))))

                     )))))
          ; force-output :
          (lambda ()
            (perform
             (or eof-marker-if-closed
                 (force-output/noexception gambit-port 1))))
          ; close-port :
          (lambda ()
            (perform
             (or eof-marker-if-closed
                 (close)) ))
          ; read-line-until-crlf :
          (lambda (#!optional return-eof-on-eof?)
            ; Reads one line from port, using the byte reading functions and not
            ; read-line. Safe to US ASCII only.
            (lambda ()
              (or eof-marker-if-closed
                  (let loop ((lst '()))
                    (let ((x
                           (perform
                            (or eof-marker-if-closed
                                (read-u8/noexception gambit-port)))))
                      (cond
                       ((and (eq? x 10)
                             (pair? lst)
                             (eq? (car lst) #\return))
                        (reverse-list->string (cdr lst)))

                       ((eq? #!eof x)
                        (if return-eof-on-eof?
                            #!eof
                            (reverse-list->string lst)))

                       (else
                        (loop (cons (integer->char x) lst)))))))))
          )))

    io-primitives))

