
;
; HTTP FORM POST multipart/form-data format data decoder library
; For use with the Sack web middleware
; Copyright (C) 2010, 2013 Mikael More. MIT license.
;
; This format is the preferred one for HTTP POST-based file uploads, from HTML FORM:s and similar technologies.
;
; Overlaps a sack-app and provides it with a 'form:data environment parameter.
; Example:
;
; (sack-start! (multipart/form-data-form-post->
;                (lambda (env)
;                  (env 'form:data)
;                  ..
;                  )))
;
; For doing invocations manually, you can use (multipart/form-data-form-post-get env) => form-data .
;
; The decoding procedure used internally is
; (multipart/form-data-form-post-decode content-type-s get-post-data-s) => form-data .
; It invokes the get-post-data-s argument at first if the content-type-s has been found to be
; multipart/form-data , so we don't waste CPU resources on reading out the post-data without need.
;
; This module delivers the received form data values in the form of u8vectors, and can therefore be
; said to be pretty character encoding-agnostic. (Keys use to be in ascii anyhow.)
;
; Essentially what this library does is to step through the mime-encoded form data sent to this library,
; and pass each chunk through the mime library for decoding using the mime-decode-multipart-with-boundary
; procedure.
;
; ## TODO
;  * |multipart/form-data-form-post-decode| is written in a weird way so that the form data becomes available
;    at first during the sack-app's output procedure invocations. This is a completely unnecessary postponement
;    and the code can equally well be restructured so that it's available already on the initial sack-app-thunk
;    invocation. This is easy to implement, just do it as soon as of use.

(declare (block) (standard-bindings) (extended-bindings) (mostly-fixnum))

(import (std string/util srfi/13 misc/u8v) mime)

(export multipart/form-data-form-post->      ; Sack app wrapper adding 'form:data environment variable
        multipart/form-data-form-post-get    ; For manually invoking a decode
        multipart/form-data-form-post-decode ; Internal decoder procedure
        )

(define console-output-port (current-output-port))

(define (u8vector-ensure-ends-by-cr-lf u8v)
  (let ((l (u8vector-length u8v)))
    (if (or (< l 2)
            (not (eq? 13 (u8vector-ref u8v (- l 2))))
            (not (eq? 10 (u8vector-ref u8v (- l 1)))))
        (u8vector-append u8v '#u8(13 10))
        u8v)))

(define (multipart/form-data-form-post-decode content-type-s get-post-data-s)

  ; Confirm multipart/form-data content-type
  (let ((parts (string-split-char #\; content-type-s)))
    (and (>= (length parts) 2)
         (equal? "multipart/form-data" (car parts))

         ; Extract boundary string parameter (compulsory)
         (let* ((args (map
                       (lambda (v)
                         (or (let ((v (string-split-at-first-nice #\= v)))
                               (set-car! v (string-trim (car v) #\space))
                               v)
                             (cons #f #f)))
                       (cdr parts)))
                (boundary (assoc "boundary" args)))
           ; (print port: console-output-port "args=")
           ; (write args console-output-port)
           ; (print port: console-output-port" boundary=" boundary "\n")
           (and boundary
                (let* ((boundary (cdr boundary))

                       ; Perform MIME document decode on the post data string given the boundary
                       (request-body (get-post-data-s))
                       (request-body (u8vector-ensure-ends-by-cr-lf request-body))
                       (mime-data    (mime-decode-multipart-with-boundary
                                      request-body 0 (u8vector-length request-body) boundary #t)))
                  ; (display "form-post: Successfully parsed.\n" console-output-port)
                  ; (display "Input: request-body: " console-output-port)
                  ; (write request-body console-output-port)
                  ; (display " boundary: " console-output-port)
                  ; (write boundary console-output-port)
                  ; (print port: console-output-port "\n")
                  ; (display "Output: " console-output-port)
                  ; (write form-data console-output-port)
                  ; (print port: console-output-port "\n")
                  mime-data))))))

(define (multipart/form-data-form-post-get env)
  (let ((content-type ((env 'sack:headers) "content-type")))
    (and (not (null? content-type))
         (let* ((content-type (car content-type)))
           (multipart/form-data-form-post-decode
            content-type
            (lambda ()
              (let ((p (open-output-u8vector))
                    (buf (make-string 10240)))
                ((env 'sack:body) (lambda (chunk)
                                    ; (dbg "form-post-decode data handler received from client u8vector packet: " chunk)
                                    (write-subu8vector chunk 0 (u8vector-length chunk) p)))
                (let ((v (get-output-u8vector p)))
                  (close-port p)
                  (set! buf #f)
                  ; (print port: console-output-port "form-post-decode transfer data handler received from client: ")
                  ; (write v console-output-port)
                  ; (print port: console-output-port "\n")
                  v))))))))

(define (multipart/form-data-form-post-wrapper env thunk)
  ; (print port: console-output-port "form-post: Into. Thread=#" (object->serial-number (current-thread)) "\n")
  (let* ((goto-output-proc #f)
         (form-data #f)
         (got-form-data? #f)
         (get-form-data
          (lambda ()
            (if got-form-data?
                form-data
                (begin
                  (set! got-form-data? #t)
                  (set! form-data (multipart/form-data-form-post-get env))
                  form-data))))
         (env (lambda a
                (if (equal? a '(form:data))
                    (get-form-data)
                    (apply env a)))))

    (thunk env)))

(define (multipart/form-data-form-post-> thunk)
  (lambda (env)
    (multipart/form-data-form-post-wrapper env thunk)))

