
;
; Sack HTTP server middleware
;
; Copyright (C) 2008-2009 Per Eckerdal, 2010-2013 Mikael More, 2005-2007 Marc Feeley.
; All Rights Reserved.
; MIT license.
;
; Fundementally Sack delivers core HTTP server functionality only, without also providing any abstractions
; that the HTTP protocol does not imply. Thus for example, Sack does not provide functionality to read
; static files from a harddrive as to deliver in a path on the HTTP server.
;
; Sack is inspired by Ruby's Rack.
;
; ## General description of use
; The basic use goes like this:
;
;  * A Sack server is started using |sack-start!|. It takes the Sack app thunk as first argument, which
;    is the procedure that Sack invokes for each HTTP request received. Remaining arguments are
;    configuration options for what TCP port to use etc.
;
;  * The Sack app thunk takes one argument, being |env|. |env| is a procedure that contains the Sack app's
;    "environment", as in, information about URI, headers, HTTP request body contents etc.
;
;    For the default |env| provided by Sack, see |make-environment|.
;
;  * Sack provides extensibility through that "adapters" can be easily implemented, that form a layer
;    between Sack and the Sack app, and that overlaps the |env| procedure as to provide additional
;    functionality.
;
;    An example of such an extension is cookies support, provided by the cookies module.
;
;  * The Sack app procedure should return a structure (values http-code headers data-thunk) .
;
;    http-code  = HTTP response statatus code as an integer e.g. 200
;    headers    = HTTP response headers as an a-list e.g. '(("Content-Type" . "text/html; charset=utf-8"))
;    data-thunk = HTTP response body data provider thunk. Procedure taking no arguments and that returns
;                 either an u8vector containing the next chunk of body data to be sent to the client,
;                 or #f meaning that all data to be output has been provided and the provision of HTTP
;                 response thus is done now.
;
;  * On return of the sack-app, Sack sends to the client the HTTP response status and headers, and then
;    invokes data-thunk all until it returns #f.
;
; ## Simple use example
; Opens a HTTP server on port 8000, available to connect to from any peer. For each HTTP request, responds with
; HTTP status 200 (= HTTP OK), no headers, and with the text "Hello world!\n\nURI: " followed by a copy of the
; particular URI this visit is for, with domain name, path and query i.e. HTTP GET arguments.
;
; (sack-start! (lambda (env)
;                (define (r) (with-output-to-u8vector '() (lambda () (print "Hello world!\n\nURI: " (env 'sack:uri)))))
;                (values 200 '() (let ((f #t)) (lambda () (and f (begin (set! f #f) (r)))))))
;              server-address: "*" port-number: 8000)
;
; ## Note on keepalives and TCP protocol logics
; Note that the keep-alive logics are written in such a way that the client web browser is pushed to make
; some of the close:s. This is good as close:s are expensive (a structure must be kept in RAM for ~2mins
; and during this time not more than ~32000 connections including those closed ones can exist between
; the two IP:s).
;
; ## History
; 2011-02-01: Added level to selected force-output calls to ensure that data is actually
;             transmitted to the client immediately on chunk write.
; 2012-02-18: Updated the calling convention of sack:body to a more effective and functional one.
;             See the comments inside handle-sack-response for more info.
; 2013-03-06: Structured and commented code. Implemented the 'sack:client-ip/u8v and 'sack:client-ip
;             environment methods.
;
; ## TODO
;  * To specify to the server that the response is already chunk-encoded, you have to specify a
;    header like ("Transfer-Encoding" . "chunked"), that is, the value must not contain any
;    whitespace, commas or any other values (for instance "chunked" or "identity, chunked").
;    Otherwise it will not understand it. To specify more than one encoding, use several header
;    field definitions. (?)
;  * Implement maximal chunk size on chunked encoding. [how to do that?]
;  * Implement conditional gets (?)
;  * Implement partial gets (?)
;  * Implement IP serialization for IPv6 IP:s

(import (only: (std srfi/13)
               string-downcase!
               string-downcase
               reverse-list->string
               string-prefix?)
        (only: (std srfi/19)
               current-date)
        (only: (std srfi/1)
               filter)
        (only: (std string/util)
               string-split-char
               apply-dumps ; For |sack-dbg|
               apply-string-append join ; For IP string serialization
               )
        (only: uri
               uri-port-set
               uri-host-set
               uri-scheme-set
               uri-path
               uri-query
               parse-uri)
        (only: http-common
               date->rfc1123

               display-crlf
               display-header
               display-headers
               chunk-return-value->u8v&u8v-length
               http-write-with-chunked-encoding

               http-status-code
               read-headers
               chunked-coding-read-hex)
        io-primitives
        (std misc/token-table misc/exception))

(declare (standard-bindings)
         (extended-bindings)
         (block))

(export sack-start!)

(define sack-version '(0 2))

; How many HTTP requests it's the most effective thing for a web browser to pipeline in a single
; TCP connection, we let up to the web browser to decide. Due to the quite significant time it can 
; take to establish a TCP connection (including a SSL handshake, for example), it's not our
; job to impose anything that could be related to as limits in this regard.
(define default-http-connection-requests-roof 50)

;==============================================================================

; The code contains outputs for debug logging. Since it's extremely rare that Sack would need to be debugged,
; we provide the toggle for it by the user commenting our or not commenting out the actual code for it in here.  
(define sack-verbose #f)
(define-macro (sack-dbg . a)
  `(let () ; Incredibly using (begin here trigged BHBUG.
     (if sack-verbose
         (begin
           (print port: console-output-port "sack: " (apply-dumps (list ,@a)) "\n")
           (force-output console-output-port 1))))
  ; '#!void
  )

(define-macro (sack-dbg-warn . a)
  `(let ()
     (print port: console-output-port " XX sack warning: " (apply-dumps (list ,@a)) "\n")
     (force-output console-output-port 1)))

; Utilities.

;; For making debug output
(define console-output-port (current-output-port))

;; The parameter name is assumed to be lowercase.
(define (has-header? headers name)
  (let loop ((lst headers))
    (cond
     ((null? lst) #f)
     ((equal? name
              (string-downcase (caar lst))) #t)
     (else (loop (cdr lst))))))

;; The parameter name is assumed to be lowercase. Note that this
;; function does not understand comma-separated lists of in header
;; values.
(define (has-header-with-value? headers name value)
  (let loop ((lst headers))
    (cond
     ((null? lst)
      #f)
     
     ((and (equal? name
                   (string-downcase (caar lst)))
           (equal? value
                   (cdar lst)))
      #t)
     
     (else
      (loop (cdr lst))))))

;==============================================================================

; HTTP server.

(define (sack-start! sack-application             ; The Sack application thunk to be invoked for every received HTTP request.

                     #!key

                     (server-address          #f) ; #f  = Accept connections from localhost only (default set to this, for security).
                                                  ; "*" = Accept connections from anywhere
                                                  ; A local IP address can be specified e.g. "1.2.3.4" = Listen to connections made to that IP only.
                                                  ;                                                      This is of value on servers that have several IP:s
                                                  ;                                                      and you want to provide different servers on different IP:s.

                     (port-number             80) ; = TCP port to listen on, integer. Note that the ports < 1024 generally require
                                                  ;   administrator privileges to listen on.

                     (max-waiting-connections #f) ; #f = Use Gambit's default value (which is 128).

                     (threaded?               #t) ; #t = Create a Gambit thread for each HTTP connection, so that several HTTP requests can
                                                  ;      be handled concurrently.
                                                  ; #f = Process only one HTTP connection at a time, in the HTTP server's thread.
                                                  ;      Note that you need to disable keepalives completely for this to deliver.

                     (keep-alive-timeout      15) ; Integer, seconds that Sack waits for another HTTP request on a HTTP connection before considering
                                                  ; the HTTP connection to be inactive and closing it.

                     (keep-alive              default-http-connection-requests-roof) ; Upper limit of number of HTTP requests to process on one
                                                                                     ; HTTP connection.

                     (timeout                 300) ; IO timeout - if an IO operation takes longer than this, close HTTP connection.

                     (io-error-callback       (lambda (e error-specific-to-a-connection?) #f)) ; Procedure invoked on IO error.

                     (thread-name             'auto) ; Name passed to make-thread for HTTP server thread.
                                                     ; 'auto = Autogenerate a suitable name.

                     (uri-scheme              "http") ; "http" or "https". Use this scheme value in URI objects created.

                     with-io-primitives       ; We accept #f as meaning |with-standard-primitives|, just not
                                              ; to require any user of this lib that might want to pass this
                                              ; value, to import io-primitives too.
                     )
  (let* ((with-io-primitives (or with-io-primitives with-standard-primitives))
         (server-port
          (open-tcp-server
           `(server-address: ,(or server-address "")
             port-number: ,port-number
             ,@(if max-waiting-connections `(backlog: ,max-waiting-connections) '())
             ; reuse-address: #t - this is the default setting, no need mention.
             ;; server-address: '#u8(127 0 0 1) ; on localhost interface only
             char-encoding: ISO-8859-1 ; The reason we use this encoding is because all header data is communicated
                                       ; in this encoding anyhow. As for HTTP body content of the request or response,
                                       ; those are passed in u8vector format anyhow.
                                       ;
                                       ; We use ISO-8859-1 to guarantee symmetry of |with-standard-primitives| of io-primitives
                                       ; with any other io-primitives implementation such as that for HTTPS, which is to follow
                                       ; io-primitives' convention of usign ISO-8859-1 also.
             )))
         (serve/perform
          (lambda (gambit-port)

            (sack-dbg "Got a connection, creating IO primitives instance.")

            (with-io-primitives
             gambit-port keep-alive-timeout
             (lambda (io-primitives)

               ; (set-timeout! connection keep-alive-timeout io-primitives)

               (sack-dbg "IO primitives created, going into serve-connection .")
               (serve-connection sack-application
                                 timeout
                                 io-primitives
                                 threaded?
                                 port-number
                                 uri-scheme
                                 keep-alive
                                 gambit-port
                                 )
               'terminated-normally))))
         (serve
          (lambda (gambit-port)
            ; (with-exception/continuation-catcher
            ;  (lambda (e)
            ;    (io-error-callback e #t)
            ;    (list 'terminated-by-exception e)) ; Leaves thread termination message
            ;  (lambda () (serve/perform connection))
            (serve/perform gambit-port)
            ))

         ;; A mutex that is unlocked when the server should quit.
         (quit-mutex (make-mutex)))
    (mutex-lock! quit-mutex)

    (thread-start!
     (let ((thunk (lambda ()
                    (let loop ()
                      (let ((connection
                             ; If it was not for the possibility of ECONNABORTED Software caused connection abort ,
                             ; we could just read straight here. Now that that may happen though, we need a failsafe
                             ; method.
                             ; (read server-port)
                             (let loop ()
                               (with-exception-catcher
                                (lambda (e)
                                  (sack-dbg-warn "Reading from the HTTP server port gave exception, ignoring and resuming: " e)
                                  (thread-sleep! 0.1)
                                  (loop))
                                (lambda () (read server-port))))))
                        (if threaded?
                            ;; Multithreaded mode.
                            (let ((dummy-port (open-dummy)))
                              (parameterize
                               ((current-input-port  dummy-port)
                                (current-output-port dummy-port))
                               (thread-start!
                                (make-thread
                                 (lambda ()
                                   (sack-dbg "Got connection.")
                                   (serve connection))
                                 '(http-server connection-handler-thread)))))

                            ;; Single-threaded mode.
                            (begin
                              (sack-dbg "Got connection.")
                              (serve connection))))

                      ;; If the mutex is not locked, it means that we should quit.
                      (if (not (mutex-lock! quit-mutex 0))
                          (loop)))))
           (thread-name (if (eq? thread-name 'auto)
                            `(http-server connection-listener-thread port: ,port-number)
                            thread-name)))
       (if thread-name
           (make-thread thunk thread-name)
           (make-thread thunk))))

    (lambda ()
      (if (mutex-lock! quit-mutex 0)
          (error "Server has already quit (or is quitting)")
          (begin
            (mutex-unlock! quit-mutex)
            (close-port server-port))))))

;==============================================================================

; Error functions.

(define (show-error code str io-primitives)
  (display-crlf io-primitives "HTTP/1.1 " code " " (http-status-code code))
  (display-header io-primitives `("Content-Length" . ,(string-length str)))
  (display-header io-primitives `("Content-Type" . "text/html; char-encoding=UTF-8"))
  (display-crlf io-primitives)
  ((io-primitives-display io-primitives) str)
  ((io-primitives-close-port io-primitives)))

(define (method-not-implemented-error io-primitives)
  (show-error 501
              "<html><head><title>501 Method Not Implemented</title></head>\n<body><h1>Method Not Implemented</h1></body></html>"
              io-primitives))

(define (internal-server-error io-primitives)
  (show-error 500
              "<html><head><title>500 Internal Server Error</title></head>\n<body><h1>Internal Server Error</h1></body></html>"
              io-primitives))

(define (bad-request-error io-primitives)
  (show-error 400
              "<html><head><title>400 Bad Request</title></head>\n<body><h1>Bad Request</h1></body></html>"
              io-primitives))

;------------------------------------------------------------------------------

; Sack functions.

(define (make-environment threaded?
                          uri
                          request-method
                          ;; Attributes is an alist of lowercase
                          ;; header names and their values
                          attributes
                          gambit-port)
  (sack-dbg "(make-environment): Invoked. attributes " attributes)
  (let* ((client-ip/u8v (let ((d #f))
                          (lambda ()
                            (or d
                                (begin
                                  (set! d (socket-info-address (tcp-client-peer-socket-info gambit-port)))
                                  d)))))
         (client-ip     (let ((d #f))
                          (lambda ()
                            (or d
                                (let ((u (client-ip/u8v)))
                                  (set! d (case (u8vector-length u)
                                            ((4) (apply-string-append (join "." (map number->string (u8vector->list u)))))
                                            (else (error "IP string serialization not implemented for IP type" u))))
                                  d)))))
         (headers
          (lambda (name)
            (if (procedure? name)
                (for-each (lambda (pair)
                            (name (car pair) (cdr pair)))
                          attributes)
                (map cdr
                     (filter (lambda (pair)
                               (equal? (car pair) name))
                             attributes)))))
         (sack:body #f))
    (lambda (name)
      (case name

        ; Sack settings
        ((sack:version       ) sack-version)
        ((sack:single-thread?) (not threaded?))
        ((sack:root          ) "") ; ??? Remove?

        ; HTTP request information:
        ;  * HTTP connection peer information
        ((sack:client-ip/u8v ) (client-ip/u8v))
        ((sack:client-ip     ) (client-ip))
        ;
        ;  * Request and header contents
        ((sack:request-method) request-method)
        ((sack:uri           ) uri)
        ((sack:headers       ) headers)
        ;
        ;  * Body contents
        ((sack:body          ) ; The first call to sack:body is made by the internal |handle-sack-response| procedure.
                               ; It sets the sack:body variable to the right handler procedure, which is subsequently
                               ; returned on each call to 'sack:body . I.e., this code here is only like a trampoline
                               ; for the provision of that procedure.
                               (or sack:body
                                   (lambda (sack:body-to-set)
                                     (set! sack:body sack:body-to-set))))

        (else #f)))))

(define (handle-sack-response-headers io-primitives
                                      version
                                      code
                                      headers
                                      chunked-encoding?
                                      close-connection?
                                      has-close-header?)
  
  ;; Display headers
  (display-crlf io-primitives version " " code " " (http-status-code code))
  
  (if (not (has-header? headers "date"))
      (display-header io-primitives `("Date" . ,(date->rfc1123 (current-date 0)))))
  
  (if (and close-connection?
           (not has-close-header?))
      (display-header io-primitives `("Connection" . "close")))
  
  (display-headers io-primitives headers)

  ;; It's important that this header is sent after the other headers,
  ;; because they might contain other transfer encodings, and chunked
  ;; should always be the last one.
  (if chunked-encoding?
      (display-header io-primitives `("Transfer-Encoding" . "chunked")))

  (display-crlf io-primitives)
  ((io-primitives-force-output io-primitives)))

; HTTP request body reader procedure.
; Invoked by |handle-sack-response|.
;
; The reading work is done in units of the read buffer or the HTTP chunk size, whichever is
; smallest. The mechanism is structured in such a way that it can be invoked over and over. Here's
; the convention:
;
; (make-sack-request-body-handler environment connection)
; => ret
;
; ret = #f = nothing was read and there's nothing more to read
;       reader-thunk w args (data-thunk copying?), that when invoked, returns another ret.
;       (box reader-thunk) = same as the previos option, except, the data-thunk returned #f.
; Note that |reader-thunk|:s are single-use.
(define (make-sack-request-body-handler environment io-primitives)
  ; Read incoming HTTP request body
  ; (with-output-to-port console-output-port (lambda () (print "attributes=") (pp attributes)))
  (let* (; (data-thunk ((environment 'sack:body)))
         (sack:headers (environment 'sack:headers))
         (content-length (let ((v (sack:headers "content-length")))
                           ; (with-output-to-port console-output-port (lambda () (print "Sack: v=") (write v) (print "\n")))
                           (and (not (null? v)) (string->number (car v)))))
         (chunked-encoding (let ((v (sack:headers "transfer-encoding")))
                             ; (with-output-to-port console-output-port (lambda () (print "Sack: v=") (write v) (print "\n")))
                             (and (not (null? v)) (string-prefix? "chunked"
                                                                  (string-downcase (car v))))))
         (buf-size 4096)
         (buf #f))
    (define (init-buf!)
      (set! buf (make-u8vector buf-size)))
    (define (make-read-bytes bytes-left*)
      (lambda (data-thunk copying?)
        (let loop ((bytes-left bytes-left*) (data-thunk data-thunk) (copying? copying?))
          (if (< 0 bytes-left)
              (let* ((bytes-to-read (min buf-size bytes-left))
                     (_tmp (sack-dbg "(make-sack-request-body-handler): Going into read " bytes-to-read " bytes."))
                     (bytes-read ((io-primitives-read-subu8vector io-primitives) buf 0 bytes-to-read bytes-to-read)))
                (sack-dbg "(make-sack-request-body-handler): Read " bytes-read " bytes.")
                (if (and bytes-read (< 0 bytes-read))
                    (begin
                      (sack-dbg "(make-sack-request-body-handler): Read " bytes-read " bytes.")
                      (let* ((false-response? (if data-thunk
                                                 (not (if copying?
                                                          (data-thunk (subu8vector buf 0 bytes-read))
                                                          (data-thunk buf)))
                                                 #f))
                             (continue (lambda (data-thunk copying?)
                                         (loop (- bytes-left bytes-read) data-thunk copying?))))
                        (if false-response? (box continue) continue)))
                    #f))
              #f))))

    (sack-dbg "(make-sack-request-body-handler): Now into assigning handlers for reading request contents body."
              " chunked-encoding=" chunked-encoding ", content-length=" content-length)

    (cond
     ; HTTP request has chunked encoding
     ; (End of body is denoted by a zero-length chunk.)
     (chunked-encoding
      (sack-dbg "(make-sack-request-body-handler): Entered. Using chunked encoding so returning routine for reading that out.")
      (lambda (data-thunk copying?)
        (define read-line-until-crlf (io-primitives-read-line-until-crlf io-primitives))
        (sack-dbg "(make-sack-request-body-handler): Decoding chunked encoding.")
        (init-buf!)
        (let loop ((data-thunk data-thunk) (copying? copying?))
          (let* ((len-str (read-line-until-crlf))
                 (len (chunked-coding-read-hex len-str)))
            (if (not (zero? len))
                (let ((read-bytes (make-read-bytes len)))
                  (let read-bytes-loop ((data-thunk data-thunk) (copying? copying?))
                    (sack-dbg "(make-sack-request-body-handler): Reading " len " bytes http chunk")
                    (let ((k (read-bytes data-thunk copying?)))
                      (if k
                          (let ((boxed? (box? k)))
                            (set! read-bytes (if boxed? (unbox k) k))
                            (if boxed? (box read-bytes-loop) read-bytes-loop)) ; (We return read-bytes-loop)
                          (begin
                            (read-line-until-crlf io-primitives)
                            (loop data-thunk copying?))))))
                #f)))
        (sack-dbg "(make-sack-request-body-handler): Read all chunks, ending.")
        ))
     ; HTTP request has Content-Length set
     (content-length
      (sack-dbg "(make-sack-request-body-handler): Entered. We know the content length, so returning routine for reading that out.")
      (if (not (zero? content-length)) ; No need to allocate buffer and do read op if length is zero
          (lambda (data-thunk copying?)
            (init-buf!)
            (sack-dbg "(make-sack-request-body-handler): Content-Length set to " content-length ", processing.")
            ((make-read-bytes content-length) data-thunk copying?))
          #f))
     ; HTTP request has no Content-Length, but has Connection: Close
     ; HTTP request has neither Content-Length nor Connection: Close
     ; In this case we presume that there is no request body. (Presumably this is in accordance with the HTTP RFC?)
     (else
      (sack-dbg "(make-sack-request-body-handler): Entered. There's neither Content-Length nor Connection: Close nor "
                "Transfer-Encoding: Chunked, so presuming there's no request body, returning no handler.")
      #f)
     ; There could be some HTTP 1.0 or 0.9 client we'd be missing form data from, that just send it after the headers
     ; and then close the connection, without specifying Connection: Close or Content-Length.
     )

    ; Please note that the following code was written for this procedure when it had done all reading work already
    ; up to here. Now it's rather event-based. So the following code could not be taken in use straight off now.
    ; (let ((b1 (read-u8 connection)) (b2 (read-u8 connection)))
    ;   (if (not (and (memq b1 '(#\return #!eof)) (memq b2 '(#\newline #!eof))))
    ;       (begin
    ;         (with-output-to-port console-output-port (lambda () (print "Sack: Invalid end of request, expected chars 0D 0A. ")
    ;                                                    (write b1) (print " ") (write b2) (print "\n")))
    ;         (error "Invalid end of " b1 b2))))
    ; ^ I think this is commented out because the error never happened again. In case it would though
    ;   and it's because of erratic behavior from the client, 

    ; (with-output-to-port console-output-port (lambda () (print "Sack: Finished any request contents processing for connection.\n")))
    ))

; This procedure handles a HTTP request within a HTTP connection.
; It is invoked by |handle-request| of |serve-connection|.
;
; => keep-connection-alive?
(define (handle-sack-response keep-alive       ; The server's keep-alive setting, an integer saying how many more HTTP requests we allow to be
                                               ; processed over this HTTP connection.
                              sack-application ; Sack application thunk to invoke
                              environment      ; Environment closure for HTTP request (generally used under the name |env| in Sack app thunks).
                              io-primitives    ; io-primitives for doing IO on the HTTP connection
                              version          ; HTTP version symbol: 'HTTP/1.1 etc.
                              )

  ; We get here from handle-request of serve-connection . At this point, the HTTP request line and the request
  ; headers have been read in full, and none of the body has been read. The process now is to invoke the Sack
  ; application and get the response HTTP headers in full from it, and then output the headers and
  ; the HTTP body based on the |response-thunk|'s return values, and by that the handling of this HTTP
  ; request is done, with the exception for reading the HTTP request body.
  ;      This is done by invocation to the 'sack:body environment parameter. The sack application may
  ; do this as for it to get the HTTP request body's contents. After all execution of the sack app
  ; for this request, we also run 'sack:body from here to ensure complete drainage from the HTTP
  ; connection port of the body, so that the connection is reset correctly for handling of any
  ; subsequent HTTP request in the same connection.
  ;      At the end of the handling of this request i.e. at the bottom of this procedure, 'sack:body
  ; for this request is blocked from any subsequent reading, so that just in case it'd be called
  ; by the sack app after the processing of this request has finished (by another thread etc.) it
  ; won't interfere with the handling of subsequent HTTP requests using the same connection.

  (define http-request-body-read (make-sack-request-body-handler environment io-primitives))

  ; sack:body for the sack-application works as follows:
  ; ((environment 'sack:body) data-thunk #!key (copying? #t))
  ; If copying? is #t: data-thunk is a procedure that takes the arguments (u8v), where u8v is an
  ; u8vector containing the most recently read block/chunk of data. Sack makes no further use of u8v.
  ; If copying? is #f: data-thunk is a procedure that takes the arguments (u8v len), where u8v
  ; is an u8vector whose len first bytes contain the most recently read block/chunk of data. The
  ; data is guaranteed to be there up to and only up to that data-thunk returns.
  ;
  ; The read operation with sack:body per above returns when data-thunk returned #f or when the
  ; request body has been fully read.
  ; 
  ; sack:body returns #f if data-thunk returned #f or if the end was reached already on the previous
  ; sack:body call, otherwise #t. (Other than this, its return value is not affected by whether any
  ; data was actually read.)
  ;
  ; sack:body may be re-run at any time. If its previous return was because of end of data,
  ; the new run will just return #f . If the previous return was because data-thunk returned #f,
  ; the new sack:body run will continue reading right after the point the previous reading ended.
  (define sack:body (lambda (data-thunk #!key (copying? #t))
                      (let loop ()
                        (and http-request-body-read
                             (let* ((n (http-request-body-read data-thunk copying?))
                                    (boxed? (box? n)))
                               (set! http-request-body-read (if boxed? (unbox n) n))
                               (not boxed?))
                             (loop)))))
  ((environment 'sack:body) sack:body)

  (sack-dbg "Into handle-sack-response. (version=" version ")")
  ;; Invoke handler (i.e. page generator)
  (call-with-values
   (lambda () (sack-application environment))
   (lambda (code headers response-thunk)
     (let* ((write-subu8vector (io-primitives-write-subu8vector io-primitives))
            (write-response-body-verbatim
             (lambda ()
               (let loop ()
                 (let ((chunk (response-thunk)))
                   (if chunk
                       (chunk-return-value->u8v&u8v-length
                        chunk
                        (write-subu8vector u8v 0 u8v-length)
                        (loop)))))))

            (chunked-encoding?
             (and (not (has-header? headers "content-length"))
                  (not (has-header-with-value? headers
                                               "transfer-encoding"
                                               "chunked"))

                  ; It seems that for web browsers to properly understand 304 Not Modified, we must not tell HTTP
                  ; response has chunked encoding because this is received as that new content is coming in.
                  ;
                  ; Due to the greate similarity with 303 See Other and probably all 3xx status codes,
                  ; we treat all of them with this rule.
                  (not (fx<= 300 code 399))
                  ))

            (has-close-header?
             (has-header-with-value? headers
                                     "connection"
                                     "close"))

            (close-connection?
             (or (not (eq? 'HTTP/1.1 version))
                 (member "close"
                         ((environment 'sack:headers) "connection"))
                 has-close-header?
                 (<= keep-alive 1))))

       (case version
         ((HTTP/1.1 HTTP/1.0)
          ;; Write HTTP response headers
          (handle-sack-response-headers io-primitives
                                        version
                                        code
                                        headers
                                        chunked-encoding?
                                        close-connection?
                                        has-close-header?)

          ;; Write HTTP response body
          (if (and (not (equal? "head"
                                (environment 'sack:request-method)))
                   (not (eq? 304 code)))
              (if chunked-encoding?
                  (http-write-with-chunked-encoding io-primitives
                                                    (response-thunk) ; first-chunk
                                                    response-thunk   ; get-chunk
                                                    )
                  (write-response-body-verbatim)))

          ; If this is a keepalive connection then ensure that all HTTP request body contents available
          ; have been read out. This is to ensure that on the start of processing of the next HTTP request
          ; on the HTTP connection, we will start reading the HTTP request line at the correct place.
          (if (not close-connection?)
              (sack:body (lambda (u8v len) (void)) copying?: #f))
          )

         ; HTTP 0.9, just dump the HTTP response body as one data block.
         (else
          (write-response-body-verbatim)))

       ; Was:
       ; (if close-connection?
       ;     (close-port conn)
       ;     (force-output conn))
       ;
       ; Just to ensure completely that the data is sent (I believe this is superfluous but let's keep it for now):
       ((io-primitives-force-output io-primitives))
       (if close-connection? ((io-primitives-close-port io-primitives)))

       (not close-connection?)))))

;------------------------------------------------------------------------------

; Low-level serving functions.

(define version-table (make-token-table
                       ("HTTP/1.0" 'HTTP/1.0)
                       ("HTTP/1.1" 'HTTP/1.1)))

; This procedure sets IO timeouts at the level of Gambit TCP port.
; It's invoked by the HTTP TCP connection bootstrap code in |sack-start!|.
(define (set-timeout! connection timeout io-primitives)
  ;; Configure the connection with the client so that if we can't read
  ;; the request after [timeout] seconds, the read/write operation
  ;; will fail (and the thread will terminate).
  (input-port-timeout-set! connection
                           timeout
                           (lambda ()
                             (sack-dbg "Port timed out as there was no new data within the timeout period of " timeout
                                       " seconds, so closing connection now.")
                             ; (close-port connection)
                             ((io-primitives-close-port io-primitives))
                             #f)) ; Signals to Gambit that the operation that timed out should be cancelled.
  (output-port-timeout-set! connection
                            timeout
                            (lambda ()
                              (sack-dbg "Port timed out as no new data could be written within the timeout period of " timeout
                                        " seconds, so closing connection now.")
                              ; (close-port connection)
                              ((io-primitives-close-port io-primitives))
                              #f))) ; Signals to Gambit that the operation that timed out should be cancelled.

(define (serve-connection sack-application
                          timeout
                          io-primitives
                          threaded?
                          port-number
                          uri-scheme
                          keep-alive
                          gambit-port ; In here we pass it to |make-environment| only, for resolving client IP.
                          )

  (let ((req (((io-primitives-read-line-until-crlf io-primitives)
               #t ; = return-eof-on-eof?
               ))))

    (sack-dbg "Got request: " req)

    (if (eq? req #!eof)
        ; If HTTP connection closed before sending us any data, then just close the connection.
        ; This happens for instance when a Keepalive connection is closed.
        #!void

        (begin
          ; (with-output-to-port console-output-port (lambda () (print "Got HTTP req line: ") (write req) (print ".\n")))

          ; (set-timeout! connection timeout) - done by the connection bootstrack code in |sack-start!| now.
          ; |set-timeout!| is specific to Gambit's IO system, that's the reason for keeping it out of |serve-connection|.

          (let* ((end
                  (let loop ((i 0))
                    (cond ((= i (string-length req))
                           #f)
                          ((char=? (string-ref req i) #\space)
                           i)
                          (else
                           (loop (+ i 1))))))
                 (method-name
                  (let ((m (and end (substring req 0 end))))
                    (if m (string-downcase! m)) ; (If there's no space in the method, m is #f.)
                    m)))

            ; Invoked by |handle-version| below only.
            ; On success, invokes |handle-sack-response| defined in the global namespace above for proceeding with handling the HTTP request.
            ;
            ; => keep-connection-alive?
            (define (handle-request version attributes uri)
              (with-exception/continuation-catcher
               (lambda (e)
                 (let ((exception (exception/continuation-exception e)))
                   (if (and (os-exception? exception)
                            (let ((v (os-exception-arguments exception)))
                              (and (list? v) (>= (length v) 1)
                                   (port? (car v)) ; (eq? connection (car v)) - Now this new check is a bit arbitrary. Best would be if we
                                                   ;                            could make the Gambit IO routines not throw any exceptions.
                                   )))

                       ; Typically this is a "Broken pipe" exception. Don't know exactly how
                       ; to typecheck for it though.
                       (print port: console-output-port
                              " #### Sack application crashed, most probably IO error from connection failure. " e "\n")

                       (print port: console-output-port
                              "\n\n #### Sack application crashed with exception:\n"
                              (exception/continuation->string e #f #t 100 100))) ;; TODO

                   ; Send internal server error to connection.
                   ; If we get an exception while doing that, ignore it.
                   (with-exception-catcher
                    (lambda (e) #t)
                    (lambda () (internal-server-error io-primitives)))
                   
                   (list 'exception e (exception/continuation->string e))))
               (lambda ()
                 ;; Add some more info to the uri object. This is
                 ;; useful for the sack environment object.
                 (let* ((host/port (let ((ret (assoc "host" attributes)))
                                     (and ret (cdr ret))))
                        (host+port (string-split-char #\: (or host/port
                                                              "0.0.0.0"))))
                   (set! uri (uri-port-set uri
                                           (or (and (pair? (cdr host+port))
                                                    (string->number (cadr host+port)))
                                               port-number)))
                   (set! uri (uri-host-set uri (car host+port)))
                   (set! uri (uri-scheme-set uri uri-scheme)))

                 ; (with-output-to-port console-output-port (lambda () (print "Sack: Handles request " uri ".\n")))

                 ; Create an environment closure. It is the core handling block for each HTTP request.
                 ;
                 ; It is passed on verbatim onto the Sack app thunk.
                 ;
                 ; Normally within a Sack app thunk, the environment goes under the variable name |env|.
                 (let* ((environment (make-environment threaded?
                                                       uri
                                                       method-name
                                                       attributes
                                                       gambit-port)))

                   ; Proceed with handling the HTTP request.
                   (handle-sack-response keep-alive
                                         sack-application
                                         environment
                                         io-primitives
                                         version))
                 )))

            ; Invoked by request handling code below this define only.
            ; On success, invokes |handle-request| defined above for proceeding with handling the HTTP request.
            ;
            ; Procedure to proceed with the handling of a HTTP request, given knowledge of the HTTP protocol
            ; version number specified in the HTTP request.
            ;
            ; => keep-connection-alive?
            (define (handle-version version uri)
              (sack-dbg "Request uses HTTP version " version)
              (case version
                
                ; HTTP 1.0 or 1.1 request
                ((HTTP/1.0 HTTP/1.1)
                 (sack-dbg "Request is HTTP 1.0 or 1.1.")
                 (let ((attributes (read-headers io-primitives))) ; Read headers.
                   (sack-dbg "Loaded headers " attributes)
                   (cond

                    ; If failed to read headers, fail.
                    ((not attributes)
                     (sack-dbg "Request contained no attributes, aborting.")
                     (bad-request-error io-primitives)
                     #f)

                    ; If is HTTP 1.1 request and has no Host: header, fail.
                    ; (It is essential for any HTTP 1.1 request to have a Host: header.)
                    ((and (eq? 'HTTP/1.1 version)
                          (not (has-header? attributes "host")))
                     (sack-dbg "Request is for HTTP/1.1 but has no \"host\" attribute, aborting.")
                     (bad-request-error io-primitives)
                     #f)

                    ; Proceed with handling request.
                    (else
                     (sack-dbg "Going into |handle-request|.")
                     (handle-request version attributes uri)))))

                ; HTTP 0.9 request
                ((#f)
                 (sack-dbg "Treating HTTP request as being for HTTP/0.9, now going into |handle-request|.")
                 (handle-request 'HTTP/0.9
                                 '() ; headers - HTTP 0.9 does not support nor headers nor keepalives.
                                     ;           Here thusly we specify an empty list of headers.
                                 uri))

                ; Invalid HTTP version specified, fail.
                (else
                 (sack-dbg "HTTP version is " version ", unknown, aborting.")
                 (bad-request-error io-primitives)
                 #f)))

            (if method-name

                ; So we got past parsing out the method field. Now, end + 1 is the string position at which the URL should start.
                ; We ask |parse-uri| of the uri module to parse it out, and call us back with the URI object loaded and with the
                ; index in the req string that the URI ended, as for us to know at what index + 1 the HTTP version description starts.
                (parse-uri
                 req                 ; = str
                 (+ end 1)           ; = start
                 (string-length req) ; = end
                 #t                  ; = decode?
                 (lambda (uri uri-end-at-string-index)
                   (sack-dbg "Got URI object " uri ", uri-end-at-string-index=" uri-end-at-string-index)

                   ; Handle connection i.e. process the HTTP request fully and provide HTTP response fully.
                   (let ((keep-connection-alive?
                          (cond

                           ; If no URI was provided in HTTP request (or was so extremely misformatted that parse-uri
                           ; wouldn't accept it), fail.
                           ((not uri)
                            (sack-dbg "parse-uri gave us #f for uri, must have been very broken.")
                            (bad-request-error io-primitives)
                            #f)

                           ; If the HTTP request lacks HTTP version, then it's a HTTP 0.9 request.
                           ; (For more info see http://stackoverflow.com/questions/6686261/what-at-the-bare-minimum-is-required-for-an-http-request
                           ; and the HTTP 1.0 RFC, 1945.
                           ;
                           ; HTTP 0.9 connections do not support nor headers nor keepalives.
                           ((not (< uri-end-at-string-index (string-length req)))
                            (handle-version #f uri))

                           ; If the URI is not followed by a whitespace, fail.
                           ((not (char=? (string-ref req uri-end-at-string-index) #\space))
                            (sack-dbg "No space after GET/POST+space+PATH part of HTTP request.")
                            (bad-request-error io-primitives)
                            #f)

                           ; A HTTP version was specified. This is the general usecase.
                           (else
                            (let ((version-index (token-table-lookup-substring version-table
                                                                               req
                                                                               (+ uri-end-at-string-index 1)
                                                                               (string-length req))))
                              (if version-index

                                  ; Valid HTTP version specified, proceed with handling request.
                                  (let ((http-protocol-version-sy (vector-ref version-table (+ version-index 1)))) ; 'HTTP/1.1 etc.
                                    (handle-version http-protocol-version-sy uri))

                                  ; Invalid HTTP version specified, fail.
                                  (begin
                                    (sack-dbg "Bad HTTP version, tried to parse it out from " (string->list (substring req (+ uri-end-at-string-index 1) (string-length req)))
                                              ". (=index " (+ uri-end-at-string-index 1) " up to the end of the string, whose length is " (string-length req))
                                    (bad-request-error io-primitives)
                                    #f)))))))

                     ; If this connection is to be kept alive, then simply reiterate this HTTP request handling procedure.
                     ;
                     ; (Subroutines determine if this is the case - essentially if the request is with HTTP 1.1 and does
                     ; not have a Connection: Close header set, then it will be kept alive.)
                     (if keep-connection-alive?
                         (serve-connection sack-application
                                           timeout
                                           io-primitives
                                           threaded?
                                           port-number
                                           uri-scheme
                                           (- keep-alive 1)
                                           gambit-port)))))

                ; Request line contained no method. This means it was a bad HTTP request.
                ; At least for now, we always send a full HTTP response here, if not else then as to make clear
                ; to the requestor that this is a web server.
                ; (method-not-implemented-error io-primitives)
                (bad-request-error io-primitives)
                ))))))

;==============================================================================
