;
; TCP/IP utilities
;
; Copyright (c) 2008, 2012 Mikael More, 1994-2009 Marc Feeley,
; All Rights Reserved.
;
; History:
; 2012-11-22: Brought 64bit architecture support to |tcp-socket-port->fd|.

(declare (block) (mostly-fixnum) (standard-bindings) (extended-bindings)
         (not optimize-dead-local-variables) ; This is of relevance as certain objects that are loaded as arguments or local variables,
                                             ; need to not be garbage collected until the end of the procedure call as their dependency
                                             ; C structures are used of our code, and those would be deallocated if GC happened before
                                             ; that, for example the |port| argument to |tcp-client-port-sockaddr|.
         )

(import ../string/util)

; TODO: Move out the C unsigned integer array handling to a separate C FFI library
(c-define-type int-array (nonnull-pointer "int" int-array))
(define (make-int-array size #!optional fill)
  (let ((a ((c-lambda (unsigned-int) int-array "___result_voidstar = malloc(___arg1 * sizeof(int));") size)))
    (if fill (int-array-fill! a 0 size fill))
    a))
(define int-array-fill! (c-lambda (int-array unsigned-int unsigned-int int) void #<<c-lambda
int* a = ___arg1;
int from = ___arg2;
int to = ___arg3;
int fill_with = ___arg4;
for (int i = from; i < to; i++) a[i] = fill_with;
c-lambda
))
(define int-array-ref (c-lambda (int-array unsigned-int) int "___result = ___arg1[___arg2];"))
(define int-array-ref-unsigned (c-lambda (int-array unsigned-int) unsigned-int "___result = ___arg1[___arg2];"))

(c-declare "

// For sockaddr_un, msghdr etc.
#include <sys/socket.h>
// Needed???
#include <sys/un.h>

// stdlib.h replaces malloc.h , at least says OBSD 5.2's gcc.
// #include <malloc.h>
#include <stdlib.h>

")

(define (call-with-tcp-client hostname port-num proc)
  (let ((port #f))
    (dynamic-wind
        (lambda ()
          (if port
              (error "Cannot re-enter call-with-tcp-client"))
          (set! port
                (open-tcp-client
                 (list server-address: hostname
                       port-number: port-num))))
        (lambda ()
          (proc port))
        (lambda ()
          (close-port port)))))

(define (ipv4->u8vector ip)
  (let ((v (string-split #\. ip))
        (o (make-u8vector 4)))
    
    (if (not (= (length v) 4))
        (error "Invalid IPv4 IP number!" ip))
    
    (list->u8vector (map string->number v))))

;;; (More or less extracted from ~~/lib/_io#.scm)

(define (device-port-rdevice-condvar port)
  (##vector-ref port 44))
(define (device-port-wdevice-condvar port)
  (##vector-ref port 45))
(define (port-rtimeout port)
  (##vector-ref port 11))
(define (port-wtimeout port)
  (##vector-ref port 15))

; #t = Got new data to read, or EOF is reached
; #f = Got no new data
(define (port-wait-for-input port #!optional timeout-epoch-time)
  (##wait-for-io! (device-port-rdevice-condvar port)
                  (or timeout-epoch-time (port-rtimeout port))))

; #t = Port can take more output
; #f = Port can still not take more output
(define (port-wait-for-output port #!optional timeout-epoch-time)
  (##wait-for-io! (device-port-wdevice-condvar port)
                  (or timeout-epoch-time (port-wtimeout port))))

(c-declare "

// For offsetof()
#include <stddef.h>

// For the definition of iovec
#include <sys/uio.h> 

// fprintf
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#ifdef WIN32
#include <c:\\cygwin\\usr\\local\\Gambit-C\\lib\\os_io.h>
#else
#include </usr/local/Gambit-C/lib/os_io.h> // I'm well aware this is not the default installation directory for this file.
                                           // Though let's use it for convention. Requires os.h and os_time.h .
#endif

// Definition of htonll .
// Thanks to https://svn.apache.org/repos/asf/hive/branches/branch-0.2/service/include/thrift/protocol/TProtocol.h
#if __BYTE_ORDER == __BIG_ENDIAN
# define ntohll(n) (n)
# define htonll(n) (n)
#elif __BYTE_ORDER == __LITTLE_ENDIAN
# if defined(__GNUC__) && defined(__GLIBC__)
#  include <byteswap.h>
#  define ntohll(n) bswap_64(n)
#  define htonll(n) bswap_64(n)
# else /* GNUC & GLIBC */
#  define ntohll(n) ( (((unsigned long long)ntohl(n)) << 32) + ntohl(n >> 32) )
#  define htonll(n) ( (((unsigned long long)htonl(n)) << 32) + htonl(n >> 32) )
# endif /* GNUC & GLIBC */
#else /* __BYTE_ORDER */
# error Cant_ define_htonll_or_ntohll
#endif

// For sockaddr_in and sockaddr_in6
#include <netinet/in.h>

// For inet_ntop
#include <arpa/inet.h>

// This should accurately reflect the same define in os_io.c, I think:
#define SOCKET_TYPE int
#define SOCKET_LEN_TYPE socklen_t

// For our purposes here this should accurately reflect the same define in os_io.c, I think:
typedef struct ___device_tcp_client_struct {
    ___device_stream base;
    SOCKET_TYPE s;
    struct sockaddr server_addr;
    SOCKET_LEN_TYPE server_addrlen;
} ___device_tcp_client;


int create_u8vector_from_buffer(const void* buf, size_t bytes) {
     ___SCMOBJ r = ___EXT(___alloc_scmobj)(___sU8VECTOR,bytes,___MOVABLE0);
     ___EXT(___release_scmobj) (r);

     // Determine where in RAM the u8vector is (0 is the offset into u8vector)
     void *u8vectorptr = ___CAST(void*,&___FETCH_U8(___BODY(r),___INT(0)));

     // Copy gotten data to it
     memcpy(u8vectorptr,buf,(int) bytes);

     return r;
}

")

; How to get file descriptor of TCP socket port:
; ##make-device-port on _io.scm 1202 puts its rdevice and wdevice parameters
; as condition-variable-name of ##vector-ref 44 and 45 of the port.
; ##make-device-port is invoked by ##make-device-port-from-single-device of row 1560,
; which is in turn invoked by ##make-tcp-client-port 5767, which is in turn invoked
; by ##open-tcp-client on row 5816. Please note that rdevice and wdevice are the
; same variable. It is generated by ##os-device-tcp-client-open invoked on
; row 5807, that's declared in _kernel.scm row 3966, which calls ___os_device_tcp_client_open
; that's declared in os_io.c row 7272. So the return value contains this.
; The return value is generated by ___NONNULLPOINTER_to_SCMOBJ (defined in c_intf.c 4617) that convers an arbitrary pointer
; to a SCMOBJ. In this case it's the ___device_tcp_client structure dev.
;
; The socket (i.e. the file descriptor) is generated by a call to the OS function socket
; on row 3407 in the function create_tcp_socket. The result is written to the SOCKET_TYPE
; pointer argument that's the first parameter of the function. The function is called by
; ___device_tcp_client_setup_from_sockaddr declared on row 3559. After generation it passes
; the socket into ___device_tcp_client_setup_from_socket declared on row 3472, as third parameter.
; As first parameter, the function takes the ___device_tcp_client. It copies the socket into the ->s
; property of the structure.
;
; ___device_tcp_client is declared on row 2960 in os_io.c . The structure is designed like that first
; in it, there's a ___device_stream, then the socket comes. ___device_stream is defined in os_io.h row 414.
; SOCKET_TYPE is declared in os_io.c and is defined to int on USE_POSIX systems, and to SOCKET on USE_WIN32
; systems.
;
; Thus, to get the socket file descriptor of an open TCP socket port, take the int sizeof(___device_stream)
; bytes into (condition-variable-name (##vector-ref port 44)) .
;
; ** In case this function stops working with a new version of Gambit,
; ** First check if the ___device_tcp_client structure has been changed, that is, if the first property is not
; ** a ___device_stream and the second the socket anymore.
; ** Second, check if the name parameter of vector-index no. 44 of TCP connection ports aren't
; ** (see macro-device-port-rdevice-condvar's definition)
; ** foreign-objects in general, and ___device_tcp_client values in particular, anymore. Use (##vector->list port). 

(define (tcp-socket-port->fd port)
  (let* ((___device_tcp_client (condition-variable-name (device-port-rdevice-condvar port)))
         (ptr (foreign-address ___device_tcp_client)))
    ; (print "The ___device_tcp_client we got is ") (write ___device_tcp_client) (print "and its foreign-address is " ptr ".\n")
    ; We use unsigned-int64 in order to get compatibility with both 32- and 64bit architectures.
    ; The extra expense this incurs on a 32bit machine ought to be almost nil compared with unsigned-int32, and considering
    ; invocation to this procedure is extremely rare it's more than justified.
    ((c-lambda (unsigned-int64) int #<<c-lambda
       size_t ptrv = ___arg1;
       // fprintf(stderr,"In C now, the pointer value is %L. sizeof(___device_stream) is %i.\n\n",ptrv,sizeof(___device_stream));
       // fprintf(stderr,"Let's try to access it! Here's the int at that address: %i\n",*((int*)ptrv));
       // fprintf(stderr,"Worked.\n");
       ___device_stream* v = ___CAST(___device_stream*,(int*) ptrv);
       // fprintf(stderr,"v is %L.\n",(long long) v);
       ptrv += sizeof(___device_stream); // By some reason (int) offsetof(___device_tcp_client_struct,s) doesn't work.
       int* ptr = (int*) ptrv;
       // fprintf(stderr,"In C now, after having added to the pointer, it's %l (%l). The value found at its location is %i.\n\n",(long) ptr,ptrv,*ptr);
       ___result = *ptr;
c-lambda
               ) ptr)))

(define (tcp-socket->tcp-server-port socket)
  .....
  )

(c-declare #<<c-declare-end

sockaddr* gambit____device_tcp_client_to_sockaddr(size_t ptr, bool clone) {
     // Deduce the sockaddr* value
     sockaddr* server_addr = (sockaddr*) (ptr + offsetof(___device_tcp_client_struct,server_addr));
     SOCKET_LEN_TYPE* server_addr_len_ptr = (SOCKET_LEN_TYPE*) (ptr + offsetof(___device_tcp_client_struct,server_addrlen));
     size_t server_addr_len = *server_addr_len_ptr;
     // fprintf(stderr,"In C now, after having added to the pointer, it's %i (%i). The value found at its location is %i. It's %i long\n\n",
     //         (int) server_addr,ptr,*((int*)server_addr_len_ptr),server_addr_len);

     if (clone) {
          sockaddr* n = (sockaddr*) malloc(server_addr_len);
          memcpy(server_addr,n,server_addr_len);
          return n;
     }

     return server_addr;
}

___U64 sockaddr_to_int(sockaddr* sockaddr) {
     // Get the IP
     int sa_family = sockaddr->sa_family;
     if (sa_family == AF_INET) {
          sockaddr_in* s = (sockaddr_in*) sockaddr;
          // ___result = create_u8vector_from_buffer(&(s->sin_addr),sizeof(s->sin_addr));
          ___U32 ip_networkbyteorder = *((___U32*) &(s->sin_addr));
          return ntohl(ip_networkbyteorder);
     } else if (sa_family == AF_INET6) {
          sockaddr_in6* s = (sockaddr_in6*) sockaddr;
          // ___result = create_u8vector_from_buffer(&(s->sin6_addr),sizeof(s->sin6_addr));
          ___U64 ip_networkbyteorder = *((___U64*) &(s->sin6_addr));
          return ntohll(ip_networkbyteorder);
     } else
          return 0;
}

// Since it's the ordinary C free() that frees this object, really no special deallocator procedure like this would have been needed.
___SCMOBJ release_sockaddr_star(void* p) {
     if (p == 0) fprintf(stderr,"tcpip: release_sockaddr_star: Called with null parameter! Internal inconsistency !! Ignoring.\n");
     else { free(p); // fprintf(stderr,"tcpip: release_sockaddr_star: socket.h-sockaddr* %i garbage collected.\n",(int) p);
     return ___FIX(___NO_ERR);
}}

c-declare-end
)

(c-define-type socket.h-sockaddr* (nonnull-pointer "sockaddr" socket.h-sockaddr* "release_sockaddr_star"))


(define (gambit-tcp-client-port-___device_tcp_client_ptr gambit-tcp-client-port)
  ; (print port: (repl-output-port) "gambit-tcp-client-port-___device_tcp_client_ptr entered. gambit-tcp-client-port is " gambit-tcp-client-port ".\n") (force-output (repl-output-port) 1)
  (let ((rdevice-condvar (device-port-rdevice-condvar gambit-tcp-client-port)))
    ; (print port: (repl-output-port) "The rdevice-condvar is " rdevice-condvar ".\n") (force-output (repl-output-port) 1)
    (let ((___device_tcp_client (condition-variable-name rdevice-condvar)))
      ; (print port: (repl-output-port) "The ___device_tcp_client is #" ___device_tcp_client ".\n") (force-output (repl-output-port) 1)
      (let ((ptr (foreign-address ___device_tcp_client)))
        ; (print port: (repl-output-port) "The ___device_tcp_client is #" ___device_tcp_client " and the pointer value is " ptr ".\n") (force-output (repl-output-port) 1)
        ptr))))

; Get the IP of the remote host of a TCP client port. (Note: The TCP connection ports that are returned by a TCP server ports,
; are TCP client ports too.)
;
; We choose integer as return format here as at least on 64bit architectures, they should generally be passed around in the
; system by value, which is good from a performance point of view. (The alterantive would be as a string with the IP number.)
(define (tcp-client-port-sockaddr-int gambit-tcp-client-port)
  
  ; We use unsigned-int64 in order to get compatibility with both 32- and 64bit architectures.
  ; The extra expense this incurs on a 32bit machine ought to be almost nil compared with unsigned-int32, and considering
  ; the invocastions of this procedure are not extremely many in number, it's more than justified.
  ;
  ; This choice is also mirrored in the two procedures below.
  ((c-lambda (unsigned-int64) unsigned-int64 #<<c-lambda
     size_t ptr = ___arg1;
     sockaddr* sockaddr = gambit____device_tcp_client_to_sockaddr(ptr,false);
     ___result = sockaddr_to_int(sockaddr);
c-lambda
               ) (gambit-tcp-client-port-___device_tcp_client_ptr gambit-tcp-client-port)))

(define (tcp-client-port-sockaddr gambit-tcp-client-port)
  ; For motivation for the choice of unsigned-int64 here, see the comment in the corresponding place
  ; in |tcp-client-port-sockaddr-int| above. 
  ((c-lambda (unsigned-int64) socket.h-sockaddr* #<<c-lambda
     size_t ptr = ___arg1;
     sockaddr* s = gambit____device_tcp_client_to_sockaddr(ptr,true);
     ___result = s;
c-lambda
               ) (gambit-tcp-client-port-___device_tcp_client_ptr gambit-tcp-client-port)))

(define (tcp-client-port-ip-string gambit-tcp-client-port)
  ; For motivation for the choice of unsigned-int64 here, see the comment in the corresponding place
  ; in |tcp-client-port-sockaddr-int| above.
  ((c-lambda (unsigned-int64) nonnull-char-string #<<c-lambda
     size_t ptr = ___arg1;
     // fprintf(stderr,"tcp-client-port-ip-string: Got ptr %i.\n",ptr);
     sockaddr* sockaddr = gambit____device_tcp_client_to_sockaddr(ptr,false);

     char ipv4_r [INET_ADDRSTRLEN];
     char ipv6_r [INET6_ADDRSTRLEN];
     
     int sa_family = sockaddr->sa_family;     
     if (sa_family == AF_INET) {
          sockaddr_in* s = (sockaddr_in*) sockaddr;
          ___result = (char*) inet_ntop(sockaddr->sa_family,&(s->sin_addr),ipv4_r,INET_ADDRSTRLEN);
     } else if (sa_family == AF_INET6) {
          sockaddr_in6* s = (sockaddr_in6*) sockaddr;
          ___result = (char*) inet_ntop(sockaddr->sa_family,&(s->sin6_addr),ipv6_r,INET6_ADDRSTRLEN);
     } else
          ___result = 0; // = #f = failure, will generate exception as the result type is nonnull-char-string.

c-lambda
               ) (gambit-tcp-client-port-___device_tcp_client_ptr gambit-tcp-client-port)))

; This variant does not return info about the IP of the remote.. maybe wanted, hm.
; => socket (unsigned-int) = received connection
;    #f                    = failure
(define (tcp-socket-unix-accept server-socket)
  (let ((r ((c-lambda (unsigned-int) unsigned-int #<<c-lambda
{
// Changed from sockaddr_un to sockaddr .
sockaddr remote;
int s = ___arg1;
// TODO. ___result = accept(s,&remote,(socklen_t) sizeof(struct sockaddr));
}
c-lambda
             ) server-socket)))
    (and (not
          (zero? r) ; = false. CORRECT? TODO
          )
         r)))

(define (tcp-socket-unix-close socket)
  (c-lambda (unsigned-int) bool "close")) ; Or shutdown? TODO.



; Unix bidirectional pipe
(define (unix-create-bidirectional-pipe)
  (let ((a (make-int-array 2 0)))
    ((c-lambda (int-array) void "socketpair(AF_UNIX,SOCK_STREAM,0,___arg1);") a) ; Code examples seem to presume it works
    (values (int-array-ref-unsigned a 0)
            (int-array-ref-unsigned a 1))))

; Unix unidirectional pipe
(define (unix-create-pipe)
  (let ((a (make-int-array 2 0)))
    ((c-lambda (int-array) void "pipe(___arg1);") a) ; Always works?
    (values (int-array-ref-unsigned a 0)
            (int-array-ref-unsigned a 1))))

(define (unix-socket-create&bind filename)
  ((c-lambda (nonnull-char-string) unsigned-int #<<c-lambda

char* filename = ___arg1;
unsigned int s;
struct sockaddr_un local, remote;
int len;

/* TODO.
// Allocate socket
s = socket(AF_UNIX,SOCK_STREAM,0);
// Bind socket
local.sun_family = AF_UNIX;  / * local is declared before socket() ^ * /
local.sun_path = *filename;
unlink(local.sun_path);
// Error handling for unlink result?
len = strlen(local.sun_path) + sizeof(local.sun_family);
bind(s,(struct sockaddr*) &local,len);
// Error handling for bind result?
*/
___result = s;

c-lambda
    )
   filename))

(define (unix-socket-listen socket #!optional max-waiting-connections)
  ((c-lambda (unsigned-int int) bool #<<c-lambda

unsigned int s = ___arg1;
int backlog = ___arg2;

listen(s,backlog);
// Error handling for listen result?
___result = true;

c-lambda
    )
   (or max-waiting-connections 128)))

(define (unix-socket-port-open/open filename)
  (let ((socket (unix-socket-create&bind filename)))
    (values socket (tcp-socket->tcp-server-port socket))))

(define (unix-socket-port-open/listen socket #!optional max-waiting-connections)
  (unix-socket-listen socket max-waiting-connections))

(define (unix-socket-port-open filename #!optional max-waiting-connections)
  (let ((socket (unix-socket-create&bind filename)))
    (unix-socket-listen socket max-waiting-connections)
    (values socket (tcp-socket->tcp-server-port socket))))

(c-declare "

// http://www.masterraghu.com/np/introduction/unix_network_programming_v1.3/app04lev1sec2.html#app04fig02 :
#define HAVE_MSGHDR_MSG_CONTROL 1

ssize_t write_fd(int fd, void *ptr, size_t nbytes, int sendfd) {
     struct msghdr msg;
     struct iovec iov[1];

#ifdef HAVE_MSGHDR_MSG_CONTROL
     union {
          struct cmsghdr cm;
          char    control[CMSG_SPACE(sizeof(int))];
     } control_un;
     struct cmsghdr *cmptr;

     msg.msg_control = control_un.control;
     msg.msg_controllen = sizeof(control_un.control);

     cmptr = CMSG_FIRSTHDR(&msg);
     cmptr->cmsg_len = CMSG_LEN(sizeof(int));
     cmptr->cmsg_level = SOL_SOCKET;
     cmptr->cmsg_type = SCM_RIGHTS;
     *((int *) CMSG_DATA(cmptr)) = sendfd;
#else
     msg.msg_accrights = (caddr_t) & sendfd;
     msg.msg_accrightslen = sizeof(int);
#endif

     msg.msg_name = NULL;
     msg.msg_namelen = 0;

     iov[0].iov_base = ptr;
     iov[0].iov_len = nbytes;
     msg.msg_iov = iov;
     msg.msg_iovlen = 1;

     return (sendmsg(fd,&msg,0));
}

")

(define (tcp-socket-unix-send-descriptor target-socket text socket-to-send)
  ((c-lambda (unsigned-int nonnull-UTF-8-string unsigned-int) bool #<<c-lambda

int target_socket = ___arg1;
char* text_to_send = ___arg2;
int socket_to_send = ___arg3;


c-lambda
    ) target-socket text socket-to-send))
